# Prism-Agents

## Purpose
Prism is an artificial exchange in which participants, or agents, can submit orders via a web service API endpoint.

Prism-Agents offers a set of tools to facilitate the development and interaction of 
python-built agents with the Prism server. 



## Tutorials
Further below,  we describe the high-level components that make up Prism-Agents. For concrete examples of their
use in creating an end-to-end agent, see the [tutorials](tutorials/README.md) folder. 
In that folder, we have a series of notebook tutorials
([A](tutorials/tutorial_a_server_interface.ipynb), [A.2](tutorials/tutorial_a_2_barebones_api.ipynb),
[B](tutorials/tutorial_b_agents.ipynb), [C](tutorials/tutorial_c_trading_rules_and_forecasters.ipynb), 
[D](tutorials/tutorial_d_experiments_tutorial.ipynb))
which gradually introduce layers of abstraction.

We encourage the reader to quickly go through these examples to understand how their own agents 
may use the `prism-agents` toolkit to interact with Prism. Tutorials [A](tutorials/tutorial_a_server_interface.ipynb) and
[B](tutorials/tutorial_b_agents.ipynb) gives the minimal introduction that will be suitable for most
participants.

*** Note that `prism-agents` contains a lot 
of features. We do not expect participants to go through all tutorials or use all available tools.
The more involved the features, the more abstractions and opinions are baked into the design. 
Feel free to deviate from these during your development process. ***


## The Prism Server and Dashboard
The Prism server offers a dashboard to visualize recent trades and the state of its order book.
Participants can access it via the url provided by the Prism team.

![Prism Dashboard](prism_dashboard.png)

The dashboard has 3 sections. 
### Left: State of the Order Book

This column shows the bids/asks with their price, quantity and total quantity. For instance, the
first row show orders for 5 shares at a price of 9.72 USD, the third column contains the number 68
which represents the total shares that bidders are willing to buy for a price of 9.72 USD or higher.
This total number of available shares for a given price is plotted on the right and give a visual insight into the
amount of orders, and their depth, that exists in the orderbook.

There is also a vertical split to separate bids at the top (in green), and asks at the bottom (in red).
The mid point price is shown in the middle and is the average between the best ask and best bid, in this 
state we see a mid point of 9.98 USD.

### Center
The center top element of the dashboard contain the trade flows in the last 60 seconds. 
Participants are represented as edges on the Chord diagram and one can however over these to see the corresponding alias.
The ribbons size indicate the number of shares that were exchanged between two participants.


The bottom half shows the various participant metrics. Rows are denoted by the participant's alias and 
show the position amounts
(e.g. number of shares owned where a negative number implies having a short position), the cost basis of the position,
the unrealized (U) and realized (R) P&L. 

### Right
At the top we have a Volume Weighted Average Price (VWAP) curve with every
trade of a given price being shown as a dot whose size denotes the volume of the trade.

At the bottom we see the details of these trades between the participants. An interesting piece of 
information in this box is the realized P&L that each participant obtained from making the trade.


## Environment setup
To run any of the code mentioned here, you will need to set up your environment as described in the Environment Setup section of the [repository readme](../README.md).

## Components

### AgentRunner
All prism-agents are children of the base class
[`AgentRunner`](runner/agent_runner.py)
which implements a `run` method which starts up an Agent interaction with a Prism server identified by 
a URL via [`RemoteServerInterface`](api/remote_request_funcs.py).

The `RemoteServerInterface` wraps the Prism API endpoints to avoid interacting with Prism directly with the
`requests` library and offer typing for the expected fields.

A more concrete implementation of the [`AgentRunner`](runner/agent_runner.py)
is the [`MarketReplay`](agents/zero_intelligence/market_replay.py)
agent whose orders are based off of a dataset of past market movements. Its implementation of the `run` method
shows one concrete example of interaction with the prism server via `prism-agents` SDK.

Nevertheless, the [`MarketReplay`](agents/zero_intelligence/market_replay.py) agent
is not very intelligent. More sophisticated order logic will exist for agents wanting to optimize 
their Profits and Loss (PnL). The [tutorials](tutorials) offer a gentle introduction to the tools offered by
`prism-agents` to build these smarter agent.

### Forecasting based Agents
One way to generate positive PnL is to create an agent that can forecast the future price of a symbol in order to buy low and sell high. 

There are two components required for this type of agents: a 
[`Forecaster`](agents/forecast/forecast_agent.py)
and a 
[`TradingRule`](agents/forecast/forecast_agent.py).

For example, my forecast may suggest that the future price of symbol $S_t$ at time $t>t_0$ is higher than the current price $S_{t_0}$,
the trading rule must use this information to decide if it should submit a buy (or sell?!) order, and the details such as the quantity/limit price.

Once a Forecaster and TradingRule are implemented, they can be combined into a 
[`ForecastRunner`](agents/forecast/forecast_agent.py)
that handles the boilerplate code of interacting with the Prism server. Note that this runner automatically cancels outstanding limit orders
if new trades are to be placed.

#### Forecasters
The [`Forecaster`](agents/forecast/forecast_agent.py)
abstract class includes 3 abstract methods. We will look at them using an ARIMA model. Statistical forecasting models tend to be strong baselines.

The [`AutoArimaForecaster`](agents/forecast/forecasters/refitting/auto_arima.py)
demonstrates each method. `init_state` is called when the agent is started. 

Subsequent state evolution is performed via the `observe` method which queries the market at predetermined times set by the forecast runner's period.
For the ARIMA model, this `observe` method updates the internal parameters in an online fashion. It is worth seeing how a 
[fixed LSTM based forecaster](agents/forecast/forecasters/lstm/lstm.py)
or an
[adaptive LSTM based forecaster](agents/forecast/forecasters/lstm/adaptive_lstm.py)
may implement these methods to keep track of the current input and/or to perform some online gradient updates.

#### Trading Rules
In `prism-agents` we offer simple trading rules to quickly turn forecasts into actions. One such simple trading rule is
the [`SimpleTradingRule`](agents/forecast/trading_rule/simple_trading_rule.py)
which buys or sell a fixed volume depending on whether the forecast clears a multiple of the observed spread.

### Experiments Management
To evaluate the performance of the agents, we offer some tooling to help set up personal experiments involving multiple interacting agents.
#### Experiment Runner
The [`run_experiment`](runner/experiment_runner.py)
function allows us to quickly run a list of agents in parallel.
