from __future__ import annotations

from abc import ABC, abstractmethod

import attr
import methodtools
import numpy as np


class AnnealingState(ABC):
    """
    Represents the current state of an annealing schedule
    """

    @abstractmethod
    def advance(self) -> AnnealingState:
        pass

    @abstractmethod
    def get_multipliers(self, n: int) -> np.ndarray:
        """
        Return a length n vector of multipliers with which to scale
        an initial vector of poisson rates of length n
        to produce an annealed vector of poisson rates of length n
        """

    def apply(self, rates: np.ndarray) -> np.ndarray:
        """
        Scale a vector of initial poisson rates according to the current
        annealing state
        """
        return rates * self.get_multipliers(len(rates))

    def apply_float(self, rate: float) -> float:
        """
        Scale a floating point number according to the current annealing state
        """
        return float(self.apply(np.array([rate])))


class ConstantAnnealingState(AnnealingState):
    def get_multipliers(self, n: int) -> np.ndarray:
        return np.ones(n)

    def advance(self) -> ConstantAnnealingState:
        return self


@attr.s(frozen=True, auto_attribs=True)
class ExponentialAnnealingState(AnnealingState):
    """
    Annealer that multiplies every poisson rate by the same factor in every step
    """

    factor: float
    val: float = 1.0

    def advance(self) -> AnnealingState:
        return attr.evolve(self, val=self.val * self.factor)

    def get_multipliers(self, n: int) -> np.ndarray:
        return np.full(n, fill_value=self.val)

    @classmethod
    def create_with_half_life(cls, half_life: float) -> ExponentialAnnealingState:
        return cls(factor=0.5 ** (1 / half_life))


@attr.s(frozen=True, auto_attribs=True)
class RestrictedExponentialAnnealingState(AnnealingState):
    """
    Annealer that exponentially anneals all poisson
    rates up to max_anneal_depth, while
    leaving poisson rates beyond max_anneal depth unchanged.
    """

    factor: float
    max_anneal_depth: int
    val: float = 1.0

    def advance(self) -> AnnealingState:
        return attr.evolve(self, val=self.val * self.factor)

    def get_multipliers(self, n: int) -> np.ndarray:
        if n <= self.max_anneal_depth:
            return np.full(n, fill_value=self.val)
        return np.concatenate(
            [
                np.full(self.max_anneal_depth, fill_value=self.val),
                np.ones(n - self.max_anneal_depth),
            ]
        )

    @classmethod
    def create_with_half_life(
        cls, half_life: float, max_anneal_depth: int
    ) -> RestrictedExponentialAnnealingState:
        return cls(factor=0.5 ** (1 / half_life), max_anneal_depth=max_anneal_depth)


def assign_with_region_boundaries(values: np.ndarray, boundaries: np.ndarray, n: int) -> np.ndarray:
    """
    Create a piecewise constant array of length n.
    Indexes j between satisfying boundaries[i-1]<=j<boundaries[i]
    are assigned value[i].
    """
    assert len(values) == len(boundaries) + 1
    range_array = np.arange(n, dtype=values.dtype)
    conditions = (
        [range_array < boundaries[0]]
        + [
            (boundaries[i] <= range_array) & (range_array < boundaries[i + 1])
            for i in range(len(boundaries) - 1)
        ]
        + [boundaries[-1] <= range_array]
    )
    result = np.piecewise(range_array, conditions, values)  # type: ignore
    return result


@attr.s(auto_attribs=True)
class MultiPartExponentialAnnealingState(AnnealingState):
    """
    Annealer that exponentially anneals
    Poisson rates in different parts of the order book
    according to different annealing factors.

    factors: Array of annealing factors, one for each region.
    region_boundaries.  Array giving the boundaries of the regions.
            Should have length one less than the number of regions.
    val: current value of the annealing process. One value for each region.


    """

    factors: np.ndarray
    region_boundaries: np.ndarray
    val: np.ndarray

    def __attrs_post_init__(self):
        assert len(self.factors) == len(self.val) == len(self.region_boundaries) + 1
        assert self.num_regions() > 0

    def num_regions(self) -> int:
        return len(self.val)

    def advance(self) -> AnnealingState:
        return attr.evolve(self, val=self.val * self.factors)

    @methodtools.lru_cache()
    def get_multipliers(self, n: int) -> np.ndarray:

        val = assign_with_region_boundaries(values=self.val, boundaries=self.region_boundaries, n=n)
        return val

    @classmethod
    def create(
        cls, factors: np.ndarray, region_boundaries: np.ndarray
    ) -> MultiPartExponentialAnnealingState:
        return cls(factors=factors, region_boundaries=region_boundaries, val=np.ones_like(factors))

    @classmethod
    def create_with_half_life(
        cls, half_lives: np.ndarray, region_boundaries: np.ndarray
    ) -> MultiPartExponentialAnnealingState:
        return cls(
            factors=0.5 ** (1 / half_lives),
            region_boundaries=region_boundaries,
            val=np.ones_like(half_lives),
        )
