from decimal import ROUND_HALF_DOWN, Decimal
from random import choice, choices
from typing import AbstractSet, Callable, Iterator, Optional, Sequence, Set, Union
from uuid import UUID

import attr
import numpy as np

from prism_agents.agents.background.annealing_schedule import (
    AnnealingState,
    ConstantAnnealingState,
    MultiPartExponentialAnnealingState,
    RestrictedExponentialAnnealingState,
    assign_with_region_boundaries,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import FundamentalState
from prism_agents.agents.forecast.trading_rule.volume_profile_trading_rule import (
    peaked_poisson_rates,
    two_part_peaked_poisson_rates,
)
from prism_agents.api.agent_schemas import Order, OrderRequest, OrderType
from prism_agents.api.server_interface import ServerInterface
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner

buy_market_label = "buy_market"
sell_market_label = "sell_market"
ask_limit_label = "ask_limit"
bid_limit_label = "bid_limit"
cancel_label = "cancel"


@attr.s(frozen=True, auto_attribs=True)
class _CurRates:
    """
    Current Poisson annealing rates
    """

    cur_bid_rates: np.ndarray
    cur_ask_rates: np.ndarray
    cur_buy_mkt_rate: float
    cur_sell_mkt_rate: float
    cur_total_bid_rate: float
    cur_total_ask_rate: float


def _compute_cur_rates(
    anneal_state: AnnealingState,
    base_bid_rates: np.ndarray,
    base_ask_rates: np.ndarray,
    base_buy_mkt_rate: float,
    base_sell_mkt_rate: float,
) -> _CurRates:
    bid_rates = anneal_state.apply(base_bid_rates)
    ask_rates = anneal_state.apply(base_ask_rates)
    return _CurRates(
        cur_bid_rates=bid_rates,
        cur_ask_rates=ask_rates,
        cur_buy_mkt_rate=anneal_state.apply_float(base_buy_mkt_rate),
        cur_sell_mkt_rate=anneal_state.apply_float(base_sell_mkt_rate),
        cur_total_bid_rate=sum(bid_rates),
        cur_total_ask_rate=sum(ask_rates),
    )


@attr.s(frozen=True, auto_attribs=True)
class PoissonFundamentalTrader(StepwiseAgentRunner):
    """
    Trader that keeps track of a fundamental state, and stochastically
    submits orders around this fundamental state according
    to a poisson process to create order book depth

    Summary of algorithm:
    A)  Compute the time to next event.  This is calculating by computing the total
        poisson process rate combining together the individual processes for all
        event types (limit orders, market orders, and cancellations)
    B)  Determine the type of the next event, by sampling from the available poisson
        processes.  Each process is weighted by its rate.
    C)  Advance the fundamental process according to the fundamental period
    D)  Implement the chosen event type.  The value of the fundamental process
        is used as a reference.  For instance, if the chosen event is limit order
        The limit order will be submitted a given number of steps  from the current
        value of the fundamental process.


    initial_state: the starting state of the fundamental process
    ask_rates: vector of poisson rates for ask limit orders.  ask_rates[0] is the rate
                for limit orders closest to the current fundamental process value
    bid_rates: vector of poisson rates for bid limit orders.  bid_rates[0] is the rate
              for limit orders closest to the current fundamental process value
    buy_market_rate:  Poisson rate for buy market orders
    sell_market_rate: Poisson rate for sell market orders
    cancel_factor: the poisson rate of cancellation is cancel_factor*number of orders
    fundamental_period: time between updates to fundamental process
    volume_function: function giving the volume of position. So
                volume_function(i) gives the volume of orders submitted i steps from
                the lowest possible

    Here are some facts that can used to help choose rates:
    -- In steady state, we can expect sum(ask_rates)/cancel_factor orders on ask side,
    and sum(bid_rates)/cancel_factor orders on the bid side.
    -- If another agent places a large bid order of size Z very far above the current
       fundamental process value, it will take approximately Z/sum(ask_rates) seconds
       for this bid order to be filled and removed by the orders of the
       fundamental agent.
    """

    initial_state: FundamentalState
    ask_rates: np.ndarray
    bid_rates: np.ndarray
    buy_market_rate: float
    sell_market_rate: float
    cancel_factor: float
    fundamental_period: float
    alias: str = "fundamentalist"
    seed: int = 42
    symbol: str = "PRSM"
    horizon: int = 1
    volume_function: Callable[[int], int] = lambda x: 1
    tick_size: Decimal = Decimal("0.01")
    client_id: Optional[UUID] = None
    initial_annealing_state: AnnealingState = ConstantAnnealingState()

    def run_step(
        self, server_interface: ServerInterface, client_id: Optional[UUID] = None
    ) -> Iterator[float]:
        state = self.initial_state
        client_id = client_id or self.client_id
        if client_id is None:
            client_id = server_interface.register(self.alias)
        generator = np.random.default_rng(seed=self.seed)
        time = 0.0
        anneal_state = self.initial_annealing_state
        while True:
            cur_rates = _compute_cur_rates(
                anneal_state,
                base_bid_rates=self.bid_rates,
                base_ask_rates=self.ask_rates,
                base_buy_mkt_rate=self.buy_market_rate,
                base_sell_mkt_rate=self.sell_market_rate,
            )
            open_orders = get_open_orders(server_interface=server_interface, client_id=client_id)
            wait_time = self.time_to_next_event(
                orders=open_orders,
                cur_buy_mkt_rate=cur_rates.cur_buy_mkt_rate,
                cur_sell_mkt_rate=cur_rates.cur_sell_mkt_rate,
                cur_total_bid_rate=cur_rates.cur_total_bid_rate,
                cur_total_ask_rate=cur_rates.cur_total_ask_rate,
                generator=generator,
            )
            event_type = self.type_of_next_event(
                orders=open_orders,
                cur_buy_mkt_rate=cur_rates.cur_buy_mkt_rate,
                cur_sell_mkt_rate=cur_rates.cur_sell_mkt_rate,
                cur_total_bid_rate=cur_rates.cur_total_bid_rate,
                cur_total_ask_rate=cur_rates.cur_total_ask_rate,
            )
            yield wait_time
            new_time = time + wait_time
            state_advances = int(new_time / self.fundamental_period) - int(
                time / self.fundamental_period
            )
            for i in range(state_advances):
                state = state.advance()
                anneal_state = anneal_state.advance()
            fundamental_forecast = state.conditional_expectation(horizon=self.horizon)
            open_orders = get_open_orders(server_interface=server_interface, client_id=client_id)
            self.apply_event(
                event_type,
                current_orders=open_orders,
                fundamental_forecast=fundamental_forecast,
                server_interface=server_interface,
                client_id=client_id,
                generator=generator,
                cur_ask_rates=cur_rates.cur_ask_rates,
                cur_bid_rates=cur_rates.cur_bid_rates,
            )
            time = new_time

    def quantize(self, x: Union[Decimal, float]) -> Decimal:
        return max(Decimal(x).quantize(self.tick_size, rounding=ROUND_HALF_DOWN), Decimal(0))

    def apply_event(
        self,
        kind: str,
        current_orders: AbstractSet[Order],
        fundamental_forecast: float,
        server_interface: ServerInterface,
        cur_bid_rates: np.ndarray,
        cur_ask_rates: np.ndarray,
        client_id: UUID,
        generator: np.random.Generator,
    ):

        total_ask_rate = sum(cur_ask_rates)
        total_bid_rate = sum(cur_bid_rates)
        normalized_bid_rates = list(cur_bid_rates / total_bid_rate)
        normalized_ask_rates = list(cur_ask_rates / total_ask_rate)

        d_forecast = self.quantize(fundamental_forecast)
        market_simulate_ticks = (
            500  # simulate a market order with a limit order this many ticks away from the forecast
        )

        ask_base = self.quantize(d_forecast + self.tick_size / 2)
        bid_base = self.quantize(d_forecast - self.tick_size / 2)
        if kind == buy_market_label:
            server_interface.submit_order(
                client_id=client_id,
                order=OrderRequest(
                    symbol=self.symbol,
                    quantity=self.volume_function(0),
                    type=OrderType.LIMIT,
                    limit_price=float(d_forecast + self.tick_size * market_simulate_ticks),
                ),
            )
        elif kind == sell_market_label:
            server_interface.submit_order(
                client_id=client_id,
                order=OrderRequest(
                    symbol=self.symbol,
                    quantity=-self.volume_function(0),
                    type=OrderType.LIMIT,
                    limit_price=float(d_forecast - self.tick_size * market_simulate_ticks),
                ),
            )
        elif kind == ask_limit_label:
            location = int(generator.choice(len(normalized_ask_rates), p=normalized_ask_rates))
            server_interface.submit_order(
                client_id=client_id,
                order=OrderRequest(
                    symbol=self.symbol,
                    quantity=-self.volume_function(location),
                    type=OrderType.LIMIT,
                    limit_price=float(ask_base + location * self.tick_size),
                ),
            )

        elif kind == bid_limit_label:
            location = int(generator.choice(len(normalized_bid_rates), p=normalized_bid_rates))
            server_interface.submit_order(
                client_id=client_id,
                order=OrderRequest(
                    symbol=self.symbol,
                    quantity=self.volume_function(location),
                    type=OrderType.LIMIT,
                    limit_price=float(bid_base - location * self.tick_size),
                ),
            )
        elif kind == cancel_label:
            order_list = list(current_orders)
            if len(order_list) == 0:
                return
            chosen = choice(order_list)
            server_interface.cancel_order(order_id=chosen.id)
        else:
            raise ValueError("Unknown event type")

    def time_to_next_event(
        self,
        generator: np.random.Generator,
        orders: AbstractSet[Order],
        cur_buy_mkt_rate: float,
        cur_sell_mkt_rate: float,
        cur_total_bid_rate: float,
        cur_total_ask_rate: float,
    ) -> float:
        total_rate = (
            cur_total_bid_rate
            + cur_total_ask_rate
            + len(orders) * self.cancel_factor
            + cur_buy_mkt_rate
            + cur_sell_mkt_rate
        )
        return float(generator.exponential(1 / total_rate))

    def type_of_next_event(
        self,
        orders: AbstractSet[Order],
        cur_buy_mkt_rate: float,
        cur_sell_mkt_rate: float,
        cur_total_bid_rate: float,
        cur_total_ask_rate: float,
    ) -> str:
        total_cancelation_rate = len(orders) * self.cancel_factor
        return choices(
            [
                buy_market_label,
                sell_market_label,
                bid_limit_label,
                ask_limit_label,
                cancel_label,
            ],
            [
                cur_buy_mkt_rate,
                cur_sell_mkt_rate,
                cur_total_bid_rate,
                cur_total_ask_rate,
                total_cancelation_rate,
            ],
        )[0]


def get_open_orders(client_id: UUID, server_interface: ServerInterface) -> Set[Order]:
    return {
        order
        for order in server_interface.get_client_orders(client_id=client_id).orders
        if order.status in {"open", "partial"}
    }


def create_poisson_fundamentalist(
    initial_state: FundamentalState,
    limit_arrival_rate: int,
    fundamental_period: float,
    seed: int,
    alias: str = "fundamental",
    market_frac: float = 0.1,
    cancel_factor: float = 0.5,
    annealer: AnnealingState = ConstantAnnealingState(),
    cliend_id: Optional[UUID] = None,
    symbol: str = "PRSM",
    sz: int = 25,
    decay_exponent: float = 0.5,
    order_vol: int = 1,
) -> PoissonFundamentalTrader:
    """
    function for ergonomics
    limit_arrival_rate: poisson rate for arrival of limit orders
                       on each side of order book
    fundamental_period: period for updates to fundamental process
    market_frac: rate of market orders as a fractions of limit order arrival rate
    cancel_factor:  rate of cancellation is number of orders * cancel factor
    """
    return PoissonFundamentalTrader(
        initial_state=initial_state,
        ask_rates=peaked_poisson_rates(
            expected_vol=limit_arrival_rate, sz=sz, decay_exponent=decay_exponent
        ),
        bid_rates=peaked_poisson_rates(
            expected_vol=limit_arrival_rate, sz=sz, decay_exponent=decay_exponent
        ),
        buy_market_rate=limit_arrival_rate * market_frac,
        sell_market_rate=limit_arrival_rate * market_frac,
        cancel_factor=cancel_factor,
        fundamental_period=fundamental_period,
        alias=alias,
        seed=seed,
        initial_annealing_state=annealer,
        client_id=cliend_id,
        symbol=symbol,
        volume_function=lambda x: order_vol,
    )


def create_two_part_poisson_fundamentalist(
    initial_state: FundamentalState,
    seed: int,
    fundamental_period: float,
    inner_anneal_half_life: float,
    inner_limit_arrival_rate: int = 10,
    outer_limit_arrival_rate: int = 3,
    alias: str = "fundamental",
    market_frac: float = 0.1,
    cancel_factor: float = 0.5,
    cliend_id: Optional[UUID] = None,
    symbol: str = "PRSM",
    inner_sz: int = 15,
    outer_sz: int = 10,
    decay_exponent_inner: float = 0.5,
    decay_exponent_outer: float = 0.5,
    vol_inner: int = 1,
    vol_outer: int = 10,
) -> PoissonFundamentalTrader:
    """
    As above, but we divide each side of the order book into two regions:
    one that we anneal and one that we do not.
    The size and limit arrival rates of the region can be set separately.

    The main motivating use case for this function is
    create a fundamental trader with ``safety bumpers''
    that prevents participants from diving the price
     too far from the fundamental process.

    Since there is no annealing in the outer region,
    we can set a high vol_outer to create
    ``safety bumpers'' that will persist even after
    the volume in the inner region has been annealed away.

    The average number of orders in the inner region will be

    vol_inner*inner_limit_arrival_rate/cancel_factor
    * current_anneal factor

    The average number of orders in the  outer region will be

    vol_outer*outer_limit_arrival_rate/cancel_factor


    """
    rates = two_part_peaked_poisson_rates(
        expected_vol_inner=inner_limit_arrival_rate,
        expected_vol_outer=outer_limit_arrival_rate,
        sz_inner=inner_sz,
        sz_outer=outer_sz,
        decay_exponent_inner=decay_exponent_inner,
        decay_exponent_outer=decay_exponent_outer,
    )
    annealer = RestrictedExponentialAnnealingState.create_with_half_life(
        inner_anneal_half_life, max_anneal_depth=inner_sz
    )
    return PoissonFundamentalTrader(
        initial_state=initial_state,
        ask_rates=rates,
        bid_rates=rates,
        buy_market_rate=inner_limit_arrival_rate * market_frac,
        sell_market_rate=inner_limit_arrival_rate * market_frac,
        cancel_factor=cancel_factor,
        fundamental_period=fundamental_period,
        alias=alias,
        seed=seed,
        initial_annealing_state=annealer,
        client_id=cliend_id,
        symbol=symbol,
        volume_function=lambda x: vol_inner if x < inner_sz else vol_outer,
    )


@attr.s(frozen=True, auto_attribs=True)
class PoissonRegion:
    """
     Specifies a region of the order book
     and the corresponding behaviour of the Poisson agent
     in this region

     arrival rate:  Total expected number of orders arriving
                  per second in this region.

     sz: The size of the region measured in ticks

     volume: The order size (in shares) of orders submitted
            in this region.

     peak:  The relative location (measured in ticks)
            at which the order arrival rate peaks.
            Peak=0 indicates that order
            rate is highest closest to the mid-price.

    decay exponent:  Order rate will decay away from the peak according
                     to the power law 1/ticks**decay exponent.  Choose
                     decay exponent=0 for uniform rate

    annealing_factor:  Factor according to which to anneal the arrival rate
                       in this region at every tick of the fundamental process
    """

    arrival_rate: int
    sz: int
    volume: int = 1
    peak: int = 0
    decay_exponent: float = 0.5
    annealing_factor: float = 1.0

    @classmethod
    def create_with_half_life(
        cls,
        arrival_rate: int,
        sz: int,
        half_life: float,
        volume: int = 1,
        peak: int = 1,
        decay_exponent: float = 0.5,
    ):
        return PoissonRegion(
            arrival_rate=arrival_rate,
            sz=sz,
            volume=volume,
            peak=peak,
            decay_exponent=decay_exponent,
            annealing_factor=0.5 ** (1 / half_life),
        )


def create_multi_part_poisson_fundamentalist(
    initial_state: FundamentalState,
    seed: int,
    fundamental_period: float,
    regions: Sequence[PoissonRegion],
    alias: str = "fundamental",
    market_frac: float = 0.1,
    cancel_factor: float = 0.5,
    symbol: str = "PRSM",
    client_id: Optional[UUID] = None,
) -> PoissonFundamentalTrader:
    """
    Create a Poisson fundamental agent that
    behaves differently in different regions
    of the order book.  These regions are described
    as PoissonRegions.
    """
    rate_arrays = []
    for region in regions:
        rate_arrays.append(
            peaked_poisson_rates(
                expected_vol=region.arrival_rate,
                peak=region.peak,
                decay_exponent=region.decay_exponent,
                sz=region.sz,
            )
        )
    rates = np.concatenate(rate_arrays)
    sizes = [region.sz for region in regions]
    cumulative_sizes = np.cumsum(sizes)
    region_boundaries = cumulative_sizes[:-1]
    annealing_factors = np.array([region.annealing_factor for region in regions])
    annealer = MultiPartExponentialAnnealingState.create(
        factors=annealing_factors, region_boundaries=region_boundaries
    )
    volume_vect = assign_with_region_boundaries(
        values=np.array(
            [region.volume for region in regions],
        ),
        boundaries=region_boundaries,
        n=cumulative_sizes[-1],
    )

    def volume_function(loc: int) -> int:
        return volume_vect[loc]

    return PoissonFundamentalTrader(
        initial_state=initial_state,
        ask_rates=rates,
        bid_rates=rates,
        buy_market_rate=rates[0] * market_frac,
        sell_market_rate=rates[0] * market_frac,
        cancel_factor=cancel_factor,
        fundamental_period=fundamental_period,
        alias=alias,
        seed=seed,
        initial_annealing_state=annealer,
        client_id=client_id,
        symbol=symbol,
        volume_function=volume_function,
    )
