from __future__ import annotations

from typing import List, Sequence, Tuple, TypeVar

import attr
import torch
from backports.cached_property import cached_property

from prism_agents.agents.cst_agent_torch.cst_event_types import (
    BuyMarketOrder,
    CancelAskLimitOrder,
    CancelBidLimitOrder,
    CSTEvent,
    PlaceAskLimitOrder,
    PlaceBidLimitOrder,
    SellMarketOrder,
)

"""
See Cont, Rama, Sasha Stoikov, and Rishi Talreja. "A stochastic model for order book dynamics."
Operations research 58.3 (2010): 549-563.
"""


@attr.s(slots=False, auto_attribs=True, frozen=True)
class CSTState:
    """
    The current state of the order book. Corresponds to the vector X in the paper.
    negative values represent bid orders, while positive values
    represent ask orders.

    The "quantities" vector should be combined with a "prices", indicating the the price level
    at each tick.

    The vector "quantities" is integer-valued. It represents order volume for some choice
    of volume unit.  See pg 552.
    """

    quantities: torch.Tensor = attr.ib()

    @quantities.validator
    def check(self, attribute, value: torch.Tensor):
        """
        See equation 3 in the paper. Check that we have a valid state.
        """
        q = value
        best_bid = self.best_bid_index
        best_ask = self.best_ask_index
        if not torch.all(q[: best_bid + 1] <= 0):
            raise ValueError("All bids must have non-positive quantities")
        if not torch.all(q[best_bid + 1 : best_ask] == 0):
            raise ValueError("Quantities in spread must be zero")
        if not torch.all(q[best_ask:] >= 0):
            raise ValueError("all asks must have non-negative quantities")

    @cached_property
    def best_bid_index(self) -> int:
        """
        Given a state, compute the index of the best bid.
        If there are no bids in the order book, return -1
        Adapted from page 550 in the paper
        """
        q = self.quantities
        n = len(q)
        indexes = torch.arange(n)
        bid_index = indexes[q < 0]
        if len(bid_index) == 0:
            return -1
        return int(bid_index[-1])

    @cached_property
    def best_ask_index(self) -> int:
        """
        Given a state, compute the index of the best ask.
        If there are no asks in the order book, return n
        Adapted from page 550 in the paper
        """
        q = self.quantities
        n = len(q)
        indexes = torch.arange(n)
        ask_indexes = indexes[q > 0]
        if len(ask_indexes) == 0:
            return n
        return int(ask_indexes[0])

    @cached_property
    def has_best_bid(self) -> bool:
        return self.best_bid_index >= 0

    @cached_property
    def has_best_ask(self) -> bool:
        return self.best_ask_index < len(self.quantities)

    @cached_property
    def mid_price_index(self) -> float:
        return (self.best_ask_index + self.best_bid_index) / 2

    @cached_property
    def order_count(self) -> int:
        return int(torch.sum(abs(self.quantities)))

    @cached_property
    def bid_order_count(self) -> int:
        return int(abs(torch.sum(self.quantities[self.quantities < 0])))

    @cached_property
    def ask_order_count(self) -> int:
        return int(abs(torch.sum(self.quantities[self.quantities > 0])))

    @cached_property
    def cumulative_asks(self) -> torch.Tensor:
        """
        Returns a tensor whose ith entry is the number of ask orders of whose price level
        is less than equal to the ith price level
        """
        return torch.cumsum(torch.maximum(self.quantities, torch.tensor([0])), dim=0)

    @cached_property
    def cumulative_bids(self) -> torch.Tensor:
        """
        Returns a tensor whose ith entry is the number of bid orders of whose price level
        is greater than equal to the ith price level
        """
        bids = torch.minimum(self.quantities, torch.tensor([0]))
        return abs(torch.sum(bids) + bids - torch.cumsum(bids, 0))


@attr.s(auto_attribs=True, frozen=True)
class CSTParams:
    """
    The defining parameters of the CST model
    See Table 2 in the paper
    lamb: lamb[i] is the tensor indicating the Poisson rate of arrival for
          limit orders i+1 steps from the opposite
          best price.
    theta: theta[i] is the tensor indicating the (relative)
           Poisson rate of arrival for cancellations  i+1 steps from
           the opposite best price.
    mu: Poisson rate of arrival for market orders on both sides.
    k, alpha: used to extrapolate lamb beyond len(lamb).
    max_width:  max distance from opposite best that events can happen
    """

    lamb: torch.Tensor
    theta: torch.Tensor
    mu: float
    k: float
    alpha: float
    max_width: int

    def __attrs_post_init__(self):
        assert len(self.lamb) == len(self.theta)

    @property
    def param_size(self):
        return len(self.lamb)


def get_bid_extra_steps_and_param_size(state: CSTState, base_param_size: int) -> Tuple[int, int]:
    """
    As per page 552 of the paper, compute:
    a) the number of steps we need to extrapolate lambda and theta on the bid side,
    b) the number of values from state.lam and state.theta we will need on the bid side.
    """
    best_ask_index = state.best_ask_index
    return max(best_ask_index - base_param_size, 0), min(base_param_size, best_ask_index)


def get_ask_extra_steps_and_param_size(
    state: CSTState, sz: int, base_param_size: int
) -> Tuple[int, int]:
    """
    As per page 552 of the paper, compute:
    a) the number of steps we need to extrapolate lambda and theta on the ask side,
    b) the number of values from state.lam and state.theta we will need on the ask side.
    """
    best_bid_index = state.best_bid_index
    return max(sz - best_bid_index - base_param_size - 1, 0), min(
        base_param_size, sz - best_bid_index - 1
    )


def get_lambda_vects(state: CSTState, params: CSTParams) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    As per page 552 in the paper, compute the vectors containing the poisson rate of arrival of
    bid and ask limit order.
    """
    sz = len(state.quantities)
    bid_extra_step, bid_param_size = get_bid_extra_steps_and_param_size(
        state, base_param_size=params.param_size
    )
    ask_extra_step, ask_param_size = get_ask_extra_steps_and_param_size(
        state, sz, base_param_size=params.param_size
    )

    vect_bid = torch.cat(
        [
            params.k
            / (
                torch.flip(torch.arange(bid_extra_step) + 1 + bid_param_size, dims=[0])
                ** params.alpha
            ),
            torch.flip(params.lamb[:bid_param_size], [0]),
            torch.zeros(max(sz - bid_param_size - bid_extra_step, 0)),
        ]
    )
    vect_ask = torch.cat(
        [
            torch.zeros(max(sz - ask_param_size - ask_extra_step, 0)),
            params.lamb[:ask_param_size],
            params.k / ((torch.arange(ask_extra_step) + 1 + ask_param_size) ** params.alpha),
        ]
    )
    return vect_bid, vect_ask


def get_theta_vects(state: CSTState, params: CSTParams) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute the (relative) poisson rates of order cancellation on the bid and ask sides.
    """
    sz = len(state.quantities)
    bid_extra_step, bid_param_size = get_bid_extra_steps_and_param_size(
        state, base_param_size=params.param_size
    )
    ask_extra_step, ask_param_size = get_ask_extra_steps_and_param_size(
        state, sz, base_param_size=params.param_size
    )
    vect_bid = torch.cat(
        [
            torch.full([bid_extra_step], fill_value=float(params.theta[-1])),
            torch.flip(params.theta[:bid_param_size], dims=[0]),
            torch.zeros(max(sz - bid_param_size - bid_extra_step, 0)),
        ]
    )
    vect_ask = torch.cat(
        [
            torch.zeros(max(sz - ask_param_size - ask_extra_step, 0)),
            params.theta[:ask_param_size],
            torch.full([ask_extra_step], fill_value=float(params.theta[-1])),
        ]
    )
    return vect_bid, vect_ask


def get_rate_till_next_event(state: CSTState, params: CSTParams) -> float:
    """
    Combine all Poisson processes to get aggregate poisson rate.
    """
    weights = compute_event_weights(state, params)
    total_rate = sum(item[1] for item in weights)
    return total_rate


def get_time_to_next_event(state: CSTState, params: CSTParams, generator: torch.Generator) -> float:
    """
    Sample an exponential to determine time till next event.
    """
    rate = get_rate_till_next_event(state=state, params=params)
    draw = torch.zeros(1).exponential_(lambd=rate, generator=generator)
    return float(draw)


def compute_event_weights(state: CSTState, params: CSTParams) -> Sequence[Tuple[CSTEvent, float]]:
    """
    Return a sequence associating each possible event with a non-normalized probability weight.
    See pg 551 in the paper.
    """
    return compute_event_weights_generic(state=state, params=params, state_for_cancelation=state)


def compute_event_weights_generic(
    state: CSTState, state_for_cancelation: CSTState, params: CSTParams
) -> Sequence[Tuple[CSTEvent, float]]:
    """
    As above, but allow for the possibility that a different state is used to weight cancellation
    probabilities

    An example use case would be when multiple agents are trading in the market, but we only
    want to consider our own orders as potential candidates for cancellation
    """
    assert torch.all(abs(state_for_cancelation.quantities) <= abs(state.quantities))

    sz = len(state.quantities)
    bid_lamb, ask_lamb = get_lambda_vects(state, params)
    bid_theta, ask_theta = get_theta_vects(state, params)

    bid_theta_weighted = bid_theta * abs(state_for_cancelation.quantities)
    ask_theta_weighted = ask_theta * abs(state_for_cancelation.quantities)

    result: List[Tuple[CSTEvent, float]] = []
    result.append((BuyMarketOrder(volume=1), params.mu if state.has_best_ask else 0))
    result.append((SellMarketOrder(volume=1), params.mu if state.has_best_bid else 0))
    result.extend(
        [
            (PlaceBidLimitOrder(index=i, volume=1), float(bid_lamb[i]))
            for i in range(sz)
            if (state.best_ask_index - i) < params.max_width
        ]
    )
    result.extend(
        [
            (PlaceAskLimitOrder(index=i, volume=1), float(ask_lamb[i]))
            for i in range(sz)
            if (i - state.best_bid_index) < params.max_width
        ]
    )
    result.extend(
        [
            (CancelBidLimitOrder(index=i), float(bid_theta_weighted[i]))
            for i in range(sz)
            if (state.best_ask_index - i) < params.max_width
        ]
    )
    result.extend(
        [
            (CancelAskLimitOrder(index=i), float(ask_theta_weighted[i]))
            for i in range(sz)
            if (i - state.best_bid_index) < params.max_width
        ]
    )
    result = list(filter(lambda x: x[1] > 0, result))
    return result


T = TypeVar("T")


def select_event(events_and_weights: Sequence[Tuple[T, float]], generator: torch.Generator) -> T:
    """
    Use weights to select an item from a sequence according to a multinomial distribution.
    """
    choice = int(
        torch.multinomial(
            torch.tensor([item[1] for item in events_and_weights]), 1, generator=generator
        )
    )
    return events_and_weights[choice][0]
