import attr
import torch

from prism_agents.agents.cst_agent_torch.cst_core import (
    CSTParams,
    CSTState,
    compute_event_weights,
    select_event,
)
from prism_agents.agents.cst_agent_torch.cst_event_types import (
    BuyMarketOrder,
    CancelAskLimitOrder,
    CancelBidLimitOrder,
    CSTEvent,
    PlaceAskLimitOrder,
    PlaceBidLimitOrder,
    SellMarketOrder,
)


def apply_event_local(event: CSTEvent, state: CSTState) -> CSTState:
    """
    Locally apply a market event to a state, to get an updated state.
    """
    if isinstance(event, BuyMarketOrder):
        return apply_buy_market_order(state, event.volume)
    if isinstance(event, SellMarketOrder):
        return apply_sell_market_order(state, event.volume)
    if isinstance(event, PlaceBidLimitOrder):
        return apply_bid_limit_order(state, event.index, event.volume)
    if isinstance(event, CancelBidLimitOrder):
        return _modify_limit_orders(state, event.index, 1)  # bids are represented as negative
    if isinstance(event, PlaceAskLimitOrder):
        return apply_ask_limit_order(state, event.index, event.volume)
    if isinstance(event, CancelAskLimitOrder):
        return _modify_limit_orders(state, event.index, -1)


def local_simulate_one_step(
    state: CSTState,
    params: CSTParams,
    generator: torch.Generator,
) -> CSTState:
    weights = compute_event_weights(state, params=params)
    event = select_event(weights, generator=generator)
    new_state = apply_event_local(event=event, state=state)
    return new_state


def apply_buy_market_order(state: CSTState, volume: int) -> CSTState:
    if volume > state.ask_order_count:
        raise MarketOrderTooLarge()
    new_quantities = torch.clone(state.quantities)
    i = state.best_ask_index
    while volume > 0:
        filled = min(abs(int(new_quantities[i])), volume)
        new_quantities[i] -= filled
        volume -= filled
        i += 1
    return attr.evolve(state, quantities=new_quantities)


def apply_sell_market_order(state: CSTState, volume: int) -> CSTState:
    if volume > state.bid_order_count:
        raise MarketOrderTooLarge()
    new_quantities = torch.clone(state.quantities)
    i = state.best_bid_index
    while volume > 0:
        filled = min(abs(int(new_quantities[i])), volume)
        new_quantities[i] += filled  # since bids are negative numbers
        volume -= filled
        i -= 1
    return attr.evolve(state, quantities=new_quantities)


def apply_bid_limit_order(state: CSTState, index: int, volume: int) -> CSTState:
    if index < 0 or index >= len(state.quantities):
        raise InvalidOrderIndex()
    if state.best_ask_index > index:
        return _modify_limit_orders(state, index, -volume)
    i = state.best_ask_index
    new_quantities = torch.clone(state.quantities)
    while volume > 0 and i <= index:
        filled = min(abs(int(new_quantities[i])), volume)
        new_quantities[i] -= filled
        volume -= filled
        i += 1
    new_quantities[index] -= volume
    return attr.evolve(state, quantities=new_quantities)


def apply_ask_limit_order(state: CSTState, index: int, volume: int) -> CSTState:
    if index < 0 or index >= len(state.quantities):
        raise InvalidOrderIndex()
    if state.best_bid_index < index:
        return _modify_limit_orders(state, index, volume)
    i = state.best_bid_index
    new_quantities = torch.clone(state.quantities)
    while volume > 0 and i >= index:
        filled = min(abs(int(new_quantities[i])), volume)
        new_quantities[i] += filled  # since bids are negative numbers
        volume -= filled
        i -= 1
    new_quantities[index] += volume
    return attr.evolve(state, quantities=new_quantities)


def _modify_limit_orders(state: CSTState, loc: int, amount: int) -> CSTState:
    new_quantities = torch.clone(state.quantities)
    new_quantities[loc] += amount
    return attr.evolve(state, quantities=new_quantities)


class MarketOrderTooLarge(Exception):
    pass


class InvalidOrderIndex(Exception):
    pass
