from typing import Sequence

import numpy as np
import torch
import xarray as xr
from tqdm import tqdm

from prism_agents.agents.cst_agent_torch.cst_core import CSTParams, CSTState, get_time_to_next_event
from prism_agents.agents.cst_agent_torch.cst_local import local_simulate_one_step

EVENT_DIMENSION = "event"
TICK_DIMENSION = "price"
QUANTITY = "quantity"
TIME = "time"
MID_TICK = "mid_price"


def state_and_times_to_xr(states: Sequence[CSTState], times: Sequence[float]) -> xr.Dataset:
    """
    Make an xarray dataset describing a sequence of market states produced by a CST model
    useful for plotting
    """
    return xr.Dataset(
        {
            QUANTITY: xr.DataArray(
                np.concatenate(
                    [state.quantities.numpy().reshape(-1, 1) for state in states], axis=1
                ),
                dims=(TICK_DIMENSION, EVENT_DIMENSION),
            ),
            TIME: xr.DataArray(times, dims=(EVENT_DIMENSION,)),
            MID_TICK: xr.DataArray(
                [state.mid_price_index for state in states], dims=(EVENT_DIMENSION,)
            ),
        }
    )


def n_steps_dataset(
    params: CSTParams, initial_state: CSTState, num_steps: int, generator: torch.Generator
) -> xr.Dataset:
    """
    Simulate a CST model for n steps, and produce a
    xarray dataset describing this evolution
    """
    state = initial_state
    current_time = 0.0
    state_list = []
    time_list = []
    state_list.append(state)
    time_list.append(current_time)
    for i in tqdm(range(num_steps)):
        current_time += get_time_to_next_event(state=state, params=params, generator=generator)
        state = local_simulate_one_step(state=state, params=params, generator=generator)
        state_list.append(state)
        time_list.append(current_time)
    return state_and_times_to_xr(states=state_list, times=time_list)
