import math
from collections import defaultdict
from datetime import datetime
from decimal import ROUND_HALF_DOWN, Decimal
from typing import Dict, Mapping, Sequence, Set, Tuple
from uuid import UUID

import attr
import torch

from prism_agents.agents.cst_agent_torch.cst_core import CSTState
from prism_agents.api.agent_schemas import AggregateOrders, ClientOrders, MarketData, Ticker
from prism_agents.api.schema_utility_functions import get_mid_price

"""
This module is concerned with converting order book data between two representations:
1)  The CST order book model's internal representation.
2)  The representation used by the Prism API.
"""


@attr.s(frozen=True, auto_attribs=True)
class CSTMarketParseParams:
    """
    Parameters describing how to interpret market data received from Prism as a
    CSTState.
    tick_size: minimal difference between adjacent limit orders.
    volume_unit: Number of shares to interpret as a single order in the CST model.
    num_ticks: the range of prices to keep track of, measured in tick units.
    """

    tick_size: Decimal = Decimal("0.01")
    volume_unit: int = 1
    num_ticks: int = attr.ib(default=10)

    @num_ticks.validator
    def check(self, attribute, value):
        assert self.num_ticks >= 3, "Too few ticks"


def _price_to_vol_defaultdict(
    bids: Sequence[AggregateOrders],
    asks: Sequence[AggregateOrders],
    parse_params: CSTMarketParseParams,
) -> Mapping[Decimal, int]:
    """
    Given aggregated bids and asks received from prism, get a
    dictionary mapping from price to number of orders
    (where orders are measured in multiples of the volume unit)
    """
    result: Dict[Decimal, int] = defaultdict(lambda: 0)
    for item in bids:
        price = Decimal(item.price).quantize(parse_params.tick_size, ROUND_HALF_DOWN)
        result[price] -= abs(item.volume // parse_params.volume_unit)  # bids are negative
    for item in asks:
        price = Decimal(item.price).quantize(parse_params.tick_size, ROUND_HALF_DOWN)
        result[price] += abs(item.volume // parse_params.volume_unit)
    return result


def market_data_to_state_and_prices(
    data: MarketData,
    parse_params: CSTMarketParseParams,
) -> Tuple[CSTState, Sequence[Decimal]]:
    """
    Given MarketData received from prism, create a CSTState
    and a list indicating the price corresponding to every tick.

    Attempt to center the list of prices around the mid price.  However, if the mid-price
    is close to zero, instead use an asymmetric list of prices to avoid invalid (negative) prices


    Converting MarketData to the CST representation allows us to apply the
    CST order book model as described in Cont et al. (2010)
    """
    mid = get_mid_price(data, tick_size=parse_params.tick_size).quantize(
        parse_params.tick_size, rounding=ROUND_HALF_DOWN
    )
    before = math.floor((parse_params.num_ticks - 1) / 2)
    left_price = max(mid - parse_params.tick_size * before, Decimal(0))
    prices = [left_price + parse_params.tick_size * i for i in range(parse_params.num_ticks)]
    return _market_data_and_prices_to_state(data, prices=prices, parse_params=parse_params), prices


def _market_data_and_prices_to_state(
    data: MarketData, prices: Sequence[Decimal], parse_params: CSTMarketParseParams
) -> CSTState:
    price_to_vol = _price_to_vol_defaultdict(
        bids=data.bids, asks=data.asks, parse_params=parse_params
    )
    quantities = [price_to_vol[price] for price in prices]

    return CSTState(quantities=torch.tensor(quantities))


def state_to_market_data(
    state: CSTState,
    prices: Sequence[Decimal],
    parse_params: CSTMarketParseParams,
    last_trade_price: float,
    last_trade_volume: int,
    timestamp: datetime,
    symbol: str,
) -> MarketData:
    """
    Convert a CSTState back to prism MarketData
    Main purpose is for round-trip verification.
    """
    bids = []
    asks = []
    for p, q in zip(prices, state.quantities.tolist()):
        if q < 0:
            bids.append(AggregateOrders(price=float(p), volume=abs(q) * parse_params.volume_unit))
        elif q > 0:
            asks.append(AggregateOrders(price=float(p), volume=abs(q) * parse_params.volume_unit))
    return MarketData(
        bids=bids,
        asks=asks,
        last_trade_price=last_trade_price,
        last_trade_volume=last_trade_volume,
        timestamp=timestamp,
        symbol=symbol,
    )


def client_orders_to_bid_ask_aggregated_orders(
    orders: ClientOrders, ticker: str
) -> Tuple[Set[AggregateOrders], Set[AggregateOrders]]:
    """
    Convert a ClientOrders struct from prism to the AggregateOrders
    structs corresponding to bids and asks

    This can be used to apply functions that usually work on MarketData to ClientOrders
    """
    price_vol_bid: Dict[float, int] = defaultdict(lambda: 0)
    price_vol_ask: Dict[float, int] = defaultdict(lambda: 0)
    for order in orders.orders:
        if order.status == "open" and str(order.symbol.value) == str(ticker):
            if order.volume > 0:
                price_vol_bid[order.price] += abs(order.volume)
            elif order.volume < 0:
                price_vol_ask[order.price] += abs(order.volume)
    return {AggregateOrders(price=key, volume=value) for key, value in price_vol_bid.items()}, {
        AggregateOrders(price=key, volume=value) for key, value in price_vol_ask.items()
    }


def _client_orders_order_id_defaultdict(
    orders: ClientOrders,
    ticker: str,
    tick_size: Decimal,
) -> Mapping[Decimal, Set[UUID]]:
    result: Dict[Decimal, Set[UUID]] = defaultdict(lambda: set())
    for order in orders.orders:
        if order.status == "open" and order.symbol == Ticker(ticker):
            result[Decimal(order.price).quantize(tick_size)].add(order.id)
    return result


def client_orders_to_state(
    orders: ClientOrders,
    ticker: str,
    prices: Sequence[Decimal],
    parse_params: CSTMarketParseParams,
) -> CSTState:
    """
    By using this function in combination with
    market_data_to_state_and_prices
    we can create two CSTStates: one representing the overall state of the order book,
    and one representing the orders placed by the CST agent.
    """
    bids, asks = client_orders_to_bid_ask_aggregated_orders(orders=orders, ticker=ticker)
    mapping = _price_to_vol_defaultdict(bids=list(bids), asks=list(asks), parse_params=parse_params)
    quantities = [mapping[price] for price in prices]
    return CSTState(quantities=torch.tensor(quantities))
