import logging
from decimal import Decimal
from typing import Sequence, Tuple
from uuid import UUID

from prism_agents.agents.cst_agent_torch.cst_event_types import (
    BuyMarketOrder,
    CancelAskLimitOrder,
    CancelBidLimitOrder,
    CSTEvent,
    PlaceAskLimitOrder,
    PlaceBidLimitOrder,
    SellMarketOrder,
)
from prism_agents.agents.cst_agent_torch.cst_remote.cst_remote_core import (
    CSTMarketParseParams,
    _client_orders_order_id_defaultdict,
)
from prism_agents.api.server_interface import ServerInterface

logger = logging.getLogger(__name__)


def apply_event_via_interface(
    event: CSTEvent,
    symbol: str,
    prices: Sequence[Decimal],
    limit_order_client_id: UUID,
    market_order_client_id: UUID,
    parse_param: CSTMarketParseParams,
    server_interface: ServerInterface,
) -> None:
    """
    Given an order book event represented using the CST model's
    internal representation, send an appropriate request
    to the server represented by server_interface
    """
    if isinstance(event, BuyMarketOrder):
        server_interface.market_order(
            symbol=symbol,
            client_id=market_order_client_id,
            volume=parse_param.volume_unit * event.volume,
        )
    elif isinstance(event, SellMarketOrder):
        server_interface.market_order(
            symbol=symbol,
            client_id=market_order_client_id,
            volume=-parse_param.volume_unit * event.volume,
        )
    elif isinstance(event, PlaceBidLimitOrder):
        server_interface.limit_order(
            symbol=symbol,
            client_id=limit_order_client_id,
            volume=parse_param.volume_unit * event.volume,
            price=prices[event.index],
        )
    elif isinstance(event, CancelBidLimitOrder):
        _remote_cancel_limit_order(
            server_interface=server_interface,
            limit_order_client_id=limit_order_client_id,
            price=prices[event.index],
            ticker=symbol,
            parse_params=parse_param,
        )
    elif isinstance(event, PlaceAskLimitOrder):
        server_interface.limit_order(
            symbol=symbol,
            client_id=limit_order_client_id,
            volume=-parse_param.volume_unit * event.volume,
            price=prices[event.index],
        )
    elif isinstance(event, CancelAskLimitOrder):
        _remote_cancel_limit_order(
            server_interface=server_interface,
            limit_order_client_id=limit_order_client_id,
            price=prices[event.index],
            ticker=symbol,
            parse_params=parse_param,
        )


def _remote_cancel_limit_order(
    server_interface: ServerInterface,
    limit_order_client_id: UUID,
    price: Decimal,
    ticker: str,
    parse_params: CSTMarketParseParams,
):
    orders = server_interface.get_client_orders(limit_order_client_id)
    order_dict = _client_orders_order_id_defaultdict(
        orders=orders, ticker=ticker, tick_size=parse_params.tick_size
    )
    id_set = order_dict[price]
    if len(id_set) == 0:
        # assume some other agent matched with this order
        # before we could cancel it
        return
    order_id = list(id_set)[0]
    server_interface.cancel_order(order_id=order_id)


def get_market_limit_client_ids(
    base_alias: str, server_interface: ServerInterface
) -> Tuple[UUID, UUID]:
    """
    Get two client IDs: one for use for submitting market orders,
    one for use for submitting limit orders

    This is necessary since an client's market orders cannot match with their own
    limit orders
    """
    market_id = server_interface.register(alias=f"{base_alias}_market")
    limit_id = server_interface.register(alias=f"{base_alias}_limit")
    return market_id, limit_id
