from decimal import Decimal
from typing import Iterator, Optional
from uuid import UUID

import attr
import torch

from prism_agents.agents.cst_agent_torch.cst_consts import PAPER_CST_PARAMS
from prism_agents.agents.cst_agent_torch.cst_core import (
    CSTParams,
    compute_event_weights_generic,
    get_time_to_next_event,
    select_event,
)
from prism_agents.agents.cst_agent_torch.cst_remote.cst_remote_core import (
    CSTMarketParseParams,
    client_orders_to_state,
    market_data_to_state_and_prices,
)
from prism_agents.agents.cst_agent_torch.cst_remote.cst_remote_interface import (
    apply_event_via_interface,
    get_market_limit_client_ids,
)
from prism_agents.api.server_interface import ServerInterface
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner

_prsm = "PRSM"


@attr.s(frozen=True, auto_attribs=True)
class CSTRunner(StepwiseAgentRunner):
    """
    Encapsulates everything required to run a CST agent against a given server
    """

    params: CSTParams = PAPER_CST_PARAMS
    parse_params: CSTMarketParseParams = CSTMarketParseParams(
        num_ticks=25, tick_size=Decimal("0.01")
    )
    seed: int = 42
    alias: str = "cst"

    def run_step(
        self, server_interface: ServerInterface, client_id: Optional[UUID] = None
    ) -> Iterator[float]:

        generator = torch.Generator().manual_seed(self.seed)
        if client_id is None:
            market_id, limit_id = get_market_limit_client_ids(
                base_alias=self.alias, server_interface=server_interface
            )
        else:
            market_id, limit_id = client_id, client_id

        while True:
            data = server_interface.get_market_data(symbol=_prsm)
            state, prices = market_data_to_state_and_prices(
                data=data, parse_params=self.parse_params
            )
            orders = server_interface.get_client_orders(client_id=limit_id)
            state_for_cancellation = client_orders_to_state(
                orders, ticker=_prsm, prices=prices, parse_params=self.parse_params
            )
            weights = compute_event_weights_generic(
                state, params=self.params, state_for_cancelation=state_for_cancellation
            )
            event = select_event(weights, generator=generator)
            apply_event_via_interface(
                server_interface=server_interface,
                event=event,
                symbol=_prsm,
                prices=prices,
                limit_order_client_id=limit_id,
                market_order_client_id=market_id,
                parse_param=self.parse_params,
            )
            timestep = get_time_to_next_event(state=state, params=self.params, generator=generator)
            yield timestep
