import logging
from abc import ABC, abstractmethod
from typing import Generic, Iterator, List, Optional, Tuple, TypeVar
from uuid import UUID

import attr

from prism_agents.api.agent_schemas import ClientData, MarketData, OrderRequest
from prism_agents.api.schema_utility_functions import cancel_all_client_orders
from prism_agents.api.server_interface import ServerInterface
from prism_agents.configuration.prism_agents_logging import prism_agents_configure_logging
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner

T = TypeVar("T")
U = TypeVar("U")
F = TypeVar("F")


logger = logging.getLogger(__name__)
logger = prism_agents_configure_logging(logger)


class Forecaster(Generic[T, F], ABC):
    """
    Responsible for building a time series model and extrapolating from it.

    Parameterized by type T, which should be the type of the state of the time
    series model.

    Also parameterized by the type F of the forecast.

    Forecaster objects should be immutable.  They should store all their
    state externally in objects of type T.
    """

    @abstractmethod
    def init_state(self, data: MarketData) -> T:
        """
        Initialize the state of the time series model.
        """

    @abstractmethod
    def forecast(self, state: T, horizon: int) -> F:
        """
        Given a time series model with state T,
        forecast the mid price horizon steps ahead.
        """

    @abstractmethod
    def observe(self, state: T, data: MarketData) -> T:
        """
        Given the market data for this time step, get the time series model
        state for the next time step.
        """


class TradingRule(ABC, Generic[U, F]):
    """
    A rule for converting a forecast into a
    limit order volume and price.

    parameterized by U, the type of the trading rule state
    and by F the type of the forecast.


    """

    @abstractmethod
    def generate_trades(
        self, forecast: F, state: U, horizon: int, data: MarketData, client_data: ClientData
    ) -> Tuple[List[OrderRequest], U]:
        pass

    @abstractmethod
    def init_state(self) -> U:
        pass


@attr.s(frozen=True, auto_attribs=True)
class ForecastRunner(StepwiseAgentRunner, Generic[T, U, F]):
    """
    Executes the trading strategy implied by a forecaster and a trading rule.
    """

    forecaster: Forecaster[T, F]
    trading_rule: TradingRule[U, F]
    alias: str
    horizon: int = 1
    symbol: str = "PRSM"
    period: float = 1.0
    offset: float = 0.0
    client_id: Optional[UUID] = None

    def run_step(
        self, server_interface: ServerInterface, client_id: Optional[UUID] = None
    ) -> Iterator[float]:
        client_id = client_id or self.client_id
        if client_id is None:
            client_id = server_interface.register(self.alias)
        data = server_interface.get_market_data(symbol=self.symbol)
        state = self.forecaster.init_state(data)
        trade_state = self.trading_rule.init_state()
        yield self.offset
        while True:
            yield self.period
            data = server_interface.get_market_data(symbol=self.symbol)
            client_data = server_interface.get_client_data(client_id=client_id)
            state = self.forecaster.observe(state=state, data=data)
            pred = self.forecaster.forecast(state=state, horizon=self.horizon)
            trades, trade_state = self.trading_rule.generate_trades(
                forecast=pred,
                horizon=self.horizon,
                data=data,
                state=trade_state,
                client_data=client_data,
            )
            if len(trades) > 0:
                cancel_all_client_orders(client_id=client_id, server=server_interface)
            for trade in trades:
                logger.debug(
                    f"Forecast agent {self.alias} predicted a future price of {pred}"
                    f"and is therefore submitting an order {trade}"
                )
                server_interface.submit_order(client_id=client_id, order=trade)
