from __future__ import annotations

from abc import abstractmethod
from typing import Generic, Optional, TypeVar

import attr

from prism_agents.agents.forecast.forecast_agent import Forecaster, ForecastRunner
from prism_agents.agents.forecast.trading_rule.spread_trading_rule import SpreadTradingRule
from prism_agents.agents.forecast.trading_rule.volume_profile_trading_rule import (
    VolumeProfileTradingRule,
    poisson_peaked_profile,
)
from prism_agents.api.agent_schemas import MarketData

F = TypeVar("F")

G = TypeVar("G")


class FundamentalState(Generic[F]):
    """
    Represents a process driving the fundamental value of the asset

    Parameterized by F, the type of the forecast.  This may be a scalar, a vector,
    or a distribution.
    """

    @abstractmethod
    def advance(self: G) -> G:
        """
        Advance the process to the next state
        """

    @abstractmethod
    def conditional_expectation(self, horizon: int) -> F:
        """
        Forecast horizon steps ahead
        """


@attr.s(frozen=True, auto_attribs=True)
class SimpleFundamentalState(FundamentalState[Optional[float]]):
    """
    A simple, scalar-valued
    fundamental processes that can be evaluated at an arbitrary time step
    without reference to history.
    """

    step: int

    @abstractmethod
    def _compute(self) -> float:
        """
        Evaluate the process at the current step.
        """

    def advance(self) -> SimpleFundamentalState:
        return attr.evolve(self, step=self.step + 1)

    def conditional_expectation(self, horizon: int) -> float:
        """
        evaluate the process at step n
        """
        state = self
        for _ in range(horizon):
            state = state.advance()
        return state._compute()


@attr.s(frozen=True, auto_attribs=True)
class Fundamentalist(Generic[F], Forecaster[FundamentalState[F], F]):
    """
    A forecaster that ignores market data a forecasts based on some
    (possibly stochastic) process
    """

    start_state: FundamentalState[F]

    def init_state(self, data: MarketData) -> FundamentalState[F]:
        return self.start_state

    def forecast(self, state: FundamentalState[F], horizon: int) -> F:
        return state.conditional_expectation(horizon=horizon)

    def observe(self, state: FundamentalState, data: MarketData) -> FundamentalState:
        return state.advance()


def create_spread_fundamentalist(
    initial_state: FundamentalState,
    volume: int = 10,
    alias: str = "fundamental",
    period: float = 1.0,
) -> ForecastRunner:
    return ForecastRunner(
        forecaster=Fundamentalist(start_state=initial_state),
        trading_rule=SpreadTradingRule(
            volume=volume,
        ),
        horizon=1,
        symbol="PRSM",
        alias=alias,
        period=period,
    )


def create_profile_fundamentalist(
    initial_state: FundamentalState,
    volume: int = 10,
    alias: str = "fundamental",
    bid_seed: int = 420,
    ask_seed: int = 1337,
    period: float = 1.0,
) -> ForecastRunner:
    """
    Create a fundamental trader which trades according to a VolumeProfileTradingRule
    """
    bid_profile_dist = poisson_peaked_profile(expected_vol=volume, seed=bid_seed)
    ask_profile_dist = poisson_peaked_profile(expected_vol=volume, seed=ask_seed)
    return ForecastRunner(
        forecaster=Fundamentalist(start_state=initial_state),
        trading_rule=VolumeProfileTradingRule(
            ask_profile_dist=bid_profile_dist,
            bid_profile_dist=ask_profile_dist,
        ),
        horizon=1,
        symbol="PRSM",
        alias=alias,
        period=period,
    )
