from __future__ import annotations

from typing import Optional

import attr
import numpy as np

from prism_agents.agents.forecast.forecasters.fundamentalist.process.regime_switching import (
    transition_func,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState
from prism_agents.agents.utils.vector_simulation.brownian import vector_arithmetic_brownian_motion


@attr.s(frozen=True, auto_attribs=True)
class BrownianState(ValueState):
    """
    State object representing simple arithmetic Brownian motion
    """

    value: float
    sigma: float
    mu: float
    generator: np.random.Generator = attr.ib(factory=np.random.default_rng)

    def advance(self) -> BrownianState:
        new_value = vector_arithmetic_brownian_motion(
            num_paths=1,
            mu=np.array([self.mu]),
            sigma=np.array(self.sigma),
            C=np.array([[1]]),
            step_sizes=np.array([1]),
            start_value=np.array(self.value),
            random_generator=self.generator,
        )
        return attr.evolve(self, value=float(new_value))

    def conditional_expectation(self, horizon: int) -> float:
        return self.value + horizon * self.mu

    def get_value(self) -> Optional[float]:
        return self.value


def brownian_transition_func(sigma: float, mu: float) -> transition_func:
    """
    Returns a function that can be used to initialize a Brownian motion of
    fixed parameters at a given value with a given random generator.  Useful for regime switching

    """

    def transition(value: float, generator: np.random.Generator) -> BrownianState:
        return BrownianState(value=value, sigma=sigma, mu=mu, generator=generator)

    return transition
