from __future__ import annotations

from decimal import Decimal
from typing import Optional

import attr
import numpy as np
import pandas as pd
from typing_extensions import Literal

from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState

process_mode = Literal["repeat", "terminate"]


@attr.s(frozen=True, auto_attribs=True)
class DatasetProcessState(ValueState):
    """
    State for a process that replays a dataset represented as a series
    of mid-price changes (deltas).

    mode specifies how to handle the end of the series of deltas.
    """

    value: float
    delta_array: np.ndarray
    current_time: int
    mode: process_mode = "terminate"

    def next_value(self) -> Optional[float]:
        if self.mode == "terminate":
            return (
                self.value + self.delta_array[self.current_time]
                if self.current_time < len(self.delta_array)
                else None
            )
        elif self.mode == "repeat":
            return self.value + self.delta_array[self.current_time % len(self.delta_array)]

    def advance(self) -> DatasetProcessState:
        return attr.evolve(
            self,
            current_time=self.current_time + 1,
            value=self.next_value(),
        )

    def conditional_expectation(self, horizon: int) -> Optional[float]:
        state = self
        assert horizon >= 0, "horizon must be non-negative"
        for i in range(horizon):
            state = state.advance()
        return state.value

    def get_value(self) -> Optional[float]:
        return self.value


def create_dataset_process_from_series_scaled(
    ser: pd.Series, median: float, inter_quantile_range: float
) -> DatasetProcessState:
    iqr = ser.quantile(0.75) - ser.quantile(0.25)
    assert iqr > 0, "Got zero interquantile range"
    ser = (ser - ser.median()) / iqr
    scaled: pd.Series = ser * inter_quantile_range + median
    return create_dataset_process_from_series(scaled)


def create_dataset_process_from_series(
    series: pd.Series, mode: process_mode = "terminate"
) -> DatasetProcessState:
    delta = (series - series.shift(1)).dropna().values
    return DatasetProcessState(value=series.values[0], delta_array=delta, current_time=0, mode=mode)


def get_delta_series_normalized_median_move(
    series: pd.Series,
    median_move_ticks: int = 1,
    tick_size: Decimal = Decimal("0.01"),
) -> pd.Series:
    delta: pd.Series = (series - series.shift(1)).dropna()
    nonzero = delta[delta != 0]
    abs_median = abs(nonzero).median()
    delta = delta / abs_median * float(tick_size) * median_move_ticks
    return delta


def create_dataset_process_from_series_delta_scaled(
    series: pd.Series,
    mode: process_mode = "terminate",
    median_move_ticks: int = 1,
    tick_size: Decimal = Decimal("0.01"),
    initial_value: Optional[float] = None,
) -> DatasetProcessState:
    """
    Create a dataset process from a series.
    Useful for building fundamental traders.
    Scale the series so that changes between adjacent observations are on average
    a given number of ticks
    """
    delta = get_delta_series_normalized_median_move(
        series, median_move_ticks=median_move_ticks, tick_size=tick_size
    )
    if initial_value is None:
        initial_value = series.values[0]
    return DatasetProcessState(
        value=initial_value, delta_array=delta.values, current_time=0, mode=mode
    )
