from __future__ import annotations

from typing import Optional

import attr
import numpy as np

from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState
from prism_agents.agents.utils.vector_simulation.kim_omberg import vector_arithmetic_kim_omberg


@attr.s(frozen=True, auto_attribs=True)
class KOState(ValueState):
    """
    State object representing arithmetic kim omberg process
    """

    value: float
    current_drift: float
    sigma: float
    drift_alpha: float
    drift_b: float
    drift_sigma: float

    generator: np.random.Generator = attr.ib(factory=np.random.default_rng)

    def advance(self) -> KOState:
        new_value, new_drift = vector_arithmetic_kim_omberg(
            num_paths=1,
            start_value=np.array([self.value]),
            start_drift=np.array([self.current_drift]),
            sigma=np.array([self.sigma]),
            drift_alpha=np.array([self.drift_alpha]),
            drift_sigma=np.array([self.drift_sigma]),
            drift_b=np.array([self.drift_b]),
            step_sizes=np.array([1]),
            random_generator=self.generator,
        )
        return attr.evolve(self, value=float(new_value), current_drift=float(new_drift))

    def conditional_expectation(self, horizon: int) -> float:
        return self.value + horizon * self.current_drift

    def get_value(self) -> Optional[float]:
        return self.value
