from __future__ import annotations

from typing import Optional

import attr
import numpy as np

from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState
from prism_agents.agents.utils.vector_simulation.jump import vector_geometric_merton_jump


@attr.s(auto_attribs=True, frozen=True)
class MertonGeometricState(ValueState):
    """
    Represents the state of a Merton Jump-Diffusion model
    This is:

    dS = mu S dt + sigma S dW +dJ S

    where J is the the jump process whose jumps arrive according to a Poisson process
    and have  lognormal jump size
    """

    value: float
    mu: float
    sigma: float
    poisson_rate: float
    jump_size_mean: float
    jump_size_std: float
    generator: np.random.Generator

    def advance(self) -> MertonGeometricState:
        advanced = vector_geometric_merton_jump(
            num_paths=1,
            mu=np.array([self.mu]),
            sigma=np.array([self.sigma]),
            poisson_rates=np.array([self.poisson_rate]),
            jump_size_mean=np.array([self.jump_size_mean]),
            jump_size_std=np.array([self.jump_size_std]),
            start_value=np.array([self.value]),
            random_generator=self.generator,
            C=np.array([[1]]),
            step_sizes=np.array([1]),
        )
        return attr.evolve(
            self,
            value=float(advanced),
        )

    def conditional_expectation(self, horizon: int) -> Optional[float]:
        return self.get_value() * np.exp(
            horizon * self.mu
            + horizon * self.poisson_rate * (self.jump_size_mean - self.jump_size_std**2 / 2)
        )

    def get_value(self) -> Optional[float]:
        return self.value
