from __future__ import annotations

from typing import Optional

import attr
import numpy as np

from prism_agents.agents.forecast.forecasters.fundamentalist.process.regime_switching import (
    transition_func,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState


@attr.s(auto_attribs=True, frozen=True)
class OrnsteinUhlenbeckState(ValueState):
    """
    Describes the state of an Ornstein Uhlenbeck process

    See Section 3.3 in Glasserman "Monte Carlo Methods in Financial Engineering"

    Process evolves according to the SDE


    dr(t) = alpha*(b-r(t))dt + sigma*dW(t)
    where W is Brownian motion
    """

    value: float
    alpha: float
    b: float
    sigma: float
    step: int = 0
    step_size: int = 1
    generator: np.random.Generator = attr.ib(factory=np.random.default_rng)

    @classmethod
    def create_with_half_life(
        cls,
        half_life: float,
        b: float,
        sigma: float,
        step: int,
        value: float,
        step_size: int = 1,
        generator: Optional[np.random.Generator] = None,
    ):
        alpha = np.log(2) / half_life
        if generator is None:
            generator = np.random.default_rng()
        return cls(
            alpha=alpha,
            b=b,
            sigma=sigma,
            step=step,
            value=value,
            step_size=step_size,
            generator=generator,
        )

    def mean(self, steps: int) -> float:
        decay_factor = np.exp(-self.alpha * steps)
        return decay_factor * self.value + self.b * (1 - decay_factor)

    def std(self, steps: int) -> float:
        decay_factor = np.exp(-self.alpha * steps)
        return self.sigma * np.sqrt(1 / (2 * self.alpha) * (1 - decay_factor**2))

    def _advance_multi(self, steps: int) -> OrnsteinUhlenbeckState:
        generator = self.generator
        draw = generator.standard_normal()
        return attr.evolve(self, value=self.mean(steps=steps) + self.std(steps=steps) * draw)

    def advance(self) -> OrnsteinUhlenbeckState:
        return self._advance_multi(steps=self.step_size)

    def conditional_expectation(self, horizon: int) -> Optional[float]:
        return self.mean(steps=horizon)

    def get_value(self) -> Optional[float]:
        return self.value


def ou_transition_function(alpha: float, sigma: float, b: float) -> transition_func:
    """
    Returns a function that can be used to initialize an OU motion of
    fixed parameters at a given value with a given random generator.
    Useful for regime switching
    """

    def transition(value: float, generator: np.random.Generator) -> OrnsteinUhlenbeckState:
        return OrnsteinUhlenbeckState(
            value=value, alpha=alpha, b=b, sigma=sigma, step=1, generator=generator
        )

    return transition


def ou_transition_function_reset_mean(
    alpha: float,
    sigma: float,
) -> transition_func:
    """
    Returns a function that can be used to initialize an OU motion of
    fixed parameters at a given value with a given random generator.
    Resets the mean of the OU process to the current value
    Useful for regime switching
    """

    def transition(value: float, generator: np.random.Generator) -> OrnsteinUhlenbeckState:
        return OrnsteinUhlenbeckState(
            value=value, alpha=alpha, b=value, sigma=sigma, step=1, generator=generator
        )

    return transition
