from __future__ import annotations

from typing import Callable, Optional, Sequence

import attr
import numpy as np
from numpy.random import SeedSequence

from prism_agents.agents.forecast.forecasters.fundamentalist.process.value_state import ValueState

transition_func = Callable[[float, np.random.Generator], ValueState]
transition_funcs = Sequence[transition_func]


@attr.s(frozen=True, auto_attribs=True)
class RegimeSwitchingState(ValueState):
    """
    Class holding the state of a regime switching model

    Should be constructed using the create() method

    transition_matrix: matrix whose (i,j) component is the probability of a transition from regime
                      j to regime i

    regime_transition_functions: Sequence whose ith element is a function for transitioning to state
                                 i given the current value of the process
    current_state: holds the current state of the process corresponding
                    to the currently active regime
    current regime: holds the integer index of the currently active regime
    """

    transition_matrix: np.ndarray = attr.ib()
    regime_transition_functions: transition_funcs = attr.ib()
    current_state: ValueState
    current_regime: int = attr.ib()
    generator: np.random.Generator

    @property
    def num_regimes(self) -> int:
        return len(self.regime_transition_functions)

    @classmethod
    def create(
        cls,
        transition_matrix: np.ndarray,
        regime_transition_functions: transition_funcs,
        starting_regime: int = 0,
        starting_value: float = 1.0,
        seed: Optional[int] = None,
    ):
        (seed1,) = SeedSequence(seed).spawn(1)
        generator = np.random.default_rng(seed1)
        start_state = regime_transition_functions[starting_regime](starting_value, generator)
        return RegimeSwitchingState(
            transition_matrix=transition_matrix,
            regime_transition_functions=regime_transition_functions,
            current_regime=starting_regime,
            current_state=start_state,
            generator=generator,
        )

    @transition_matrix.validator
    def check(self, attribute, value):
        assert self.transition_matrix.ndim == 2
        assert self.transition_matrix.shape == (self.num_regimes, self.num_regimes)
        assert np.all(np.isclose(self.transition_matrix.sum(axis=0), np.ones(self.num_regimes)))

    @current_regime.validator
    def check_current(self, attribute, value):
        assert 0 <= self.current_regime < self.num_regimes

    def advance(self) -> RegimeSwitchingState:
        """
        Check if we have changed regime, and advance in the appropriate regime.
        """
        transition_probs = self.transition_matrix[:, self.current_regime]
        new_regime = self.generator.choice(self.num_regimes, p=transition_probs)
        if new_regime == self.current_regime:
            return attr.evolve(
                self, current_state=self.current_state.advance(), generator=self.generator
            )
        else:
            value = self.get_value()
            assert value is not None
            return attr.evolve(
                self,
                current_state=self.regime_transition_functions[new_regime](
                    value, self.generator
                ).advance(),
                current_regime=new_regime,
                generator=self.generator,
            )

    def conditional_expectation(self, horizon: int) -> Optional[float]:
        """
        Conditional expectation of process conditional on us remaining in current regime
        """
        return self.current_state.conditional_expectation(horizon=horizon)

    def get_value(self) -> Optional[float]:
        return self.current_state.get_value()
