from abc import abstractmethod
from decimal import Decimal
from typing import Any, Callable, List, Optional, cast

import attr
import torch
from pytorch_lightning.trainer import Trainer
from torch.utils.data import DataLoader, Dataset

from prism_agents.agents.forecast.forecast_agent import Forecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import LightningLSTMForecastingModule
from prism_agents.api.schema_utility_functions import MarketData, get_best_ask, get_best_bid


class DataBuffer:
    @abstractmethod
    def add_data(self, datapoint: Any):
        pass

    @abstractmethod
    def get_dataloader(self) -> DataLoader:
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def get_latest_input(self, info: Optional[Any] = None) -> Any:
        pass


class OnlineDataset(Dataset):
    """
    Dataset made to interact with the AdaptiveLSTMForecaster during online learning
    """

    def __init__(self, inputs: torch.Tensor, targets: torch.Tensor):
        assert len(inputs) == len(targets)
        self.inputs = inputs
        self.targets = targets

    def __getitem__(self, idx: int):
        return {"input": self.inputs[idx], "target": self.targets[idx], "name": idx}

    def __len__(self):
        return len(self.inputs)


@attr.s(auto_attribs=True)
class DefaultLSTMDataBuffer(DataBuffer):
    seq_length: int
    target_horizon: int
    predict_mid_point: bool = True
    data_buffer: List[torch.Tensor] = attr.ib(factory=list)
    tick_size = Decimal("0.01")
    last_seen_input: torch.Tensor = torch.Tensor([])
    clear_buffer_after_use: bool = True

    def _get_new_market_input(self, bid: float, ask: float):
        if self.predict_mid_point:
            observation = [(bid + ask) / 2]
        else:
            observation = [bid, ask]
        return torch.Tensor([observation])

    def _prism_market_data_to_forecaster_inputs(self, data: MarketData) -> Optional[torch.Tensor]:
        decimal_bid = get_best_bid(data, tick_size=self.tick_size)
        decimal_ask = get_best_ask(data, tick_size=self.tick_size)
        if decimal_ask is None or decimal_bid is None:
            return None
        new_market_input = self._get_new_market_input(float(decimal_bid), float(decimal_ask))
        return new_market_input

    def add_data(self, datapoint: MarketData):
        torch_datapoint = self._prism_market_data_to_forecaster_inputs(datapoint)
        if torch_datapoint is not None:
            self.data_buffer.append(torch_datapoint)

            last = self.last_seen_input
            self.last_seen_input = torch.concat([last, torch_datapoint])[-self.seq_length :]

    def get_dataloader(self) -> DataLoader:
        # The data buffer are past observed bid/ask in a torch.Tensor
        # we turn this into a dataset of lookahead with horizon n
        sequence = torch.concat(self.data_buffer)
        obs_length = len(sequence)
        start_idx = 0
        target_idx = start_idx + self.seq_length + self.target_horizon - 1
        inputs = []
        targets = []
        while target_idx < obs_length:
            inputs.append(sequence[start_idx : start_idx + self.seq_length])
            targets.append(sequence[target_idx])
            start_idx += 1
            target_idx += 1

        dataset = OnlineDataset(inputs=torch.stack(inputs), targets=torch.stack(targets))
        dataloader = DataLoader(dataset)

        self.modify_data_buffer()
        return dataloader

    def modify_data_buffer(self):
        """
        If specified, clear the data buffer after use in order to not retune on old observations
        """
        if self.clear_buffer_after_use:
            self.data_buffer = []

    def __len__(self):
        return len(self.data_buffer)

    def get_latest_input(self, x: Optional[Any] = None) -> torch.Tensor:
        return self.last_seen_input


@attr.s(auto_attribs=True)
class AdaptiveLSTMState:
    model: LightningLSTMForecastingModule
    data_buffer: DataBuffer


class AdaptiveLSTMForecaster(Forecaster[AdaptiveLSTMState, Optional[float]]):
    def __init__(
        self,
        forecaster: LightningLSTMForecastingModule,
        get_trainer: Callable[[], Trainer],
        data_buffer: Optional[DataBuffer] = None,
        target_horizon: int = 1,
        batch_size_per_updates: int = 1,
        update_condition: Optional[Callable[[AdaptiveLSTMState], bool]] = None,
    ):
        self.forecaster = forecaster
        self.get_trainer = get_trainer
        self.target_horizon = target_horizon
        self.batch_size_per_updates = batch_size_per_updates
        self.seq_length = int(cast(int, self.forecaster.model.seq_length))

        output_size = list(forecaster.parameters())[-1].shape[0]
        self._init_data_buffer = (
            data_buffer
            if data_buffer
            else DefaultLSTMDataBuffer(
                self.seq_length, self.target_horizon, predict_mid_point=output_size == 1
            )
        )
        self.update_condition = update_condition

    def init_state(self, data: MarketData) -> AdaptiveLSTMState:
        self._init_data_buffer.add_data(data)
        self.forecaster = cast(LightningLSTMForecastingModule, self.forecaster)
        state = AdaptiveLSTMState(model=self.forecaster, data_buffer=self._init_data_buffer)
        return state

    def check_update_condition(self, state: AdaptiveLSTMState) -> bool:
        if self.update_condition:
            return self.update_condition(state)

        # default update condition if not supplied
        min_obs = self.batch_size_per_updates - 1
        if len(state.data_buffer) >= min_obs + self.seq_length + self.target_horizon:
            return True
        return False

    def observe(self, state: AdaptiveLSTMState, data: MarketData) -> AdaptiveLSTMState:

        data_buffer = state.data_buffer
        data_buffer.add_data(data)

        # update model
        model = state.model
        if self.check_update_condition(state):
            trainer = self.get_trainer()
            market_dataloader = data_buffer.get_dataloader()
            trainer.fit(model, market_dataloader)

        state = AdaptiveLSTMState(model=model, data_buffer=data_buffer)
        return state

    def forecast(self, state: AdaptiveLSTMState, horizon: Optional[int] = None) -> Optional[float]:
        market_input = state.data_buffer.get_latest_input()

        # not enough history to generate a forecast, return None
        if len(market_input) < self.seq_length:
            return None
        input_ = {"input": market_input.unsqueeze(0)}
        forecast = state.model.predict(input_)["pred"][0]

        # the LSTM forecasts the bid and ask, we take its mid point.
        mid_price_forecast = float(forecast.mean().detach().cpu().numpy())
        return mid_price_forecast
