from decimal import Decimal
from typing import Optional, Union, cast

import torch

from prism_agents.agents.forecast.forecast_agent import Forecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import (
    LightningLSTMForecastingModule,
    LSTMForecastingModule,
)
from prism_agents.api.schema_utility_functions import get_best_ask, get_best_bid
from prism_agents.api.server_interface import MarketData

TorchLSTMForecastingModule = Union[LSTMForecastingModule, LightningLSTMForecastingModule]


class LSTMForecaster(Forecaster[torch.Tensor, Optional[float]]):
    """
    Forecaster module that allows the torch LSTMForecastingModule to interface with a ForecastRunner
    """

    def __init__(self, forecaster: TorchLSTMForecastingModule):
        self.forecaster = forecaster
        self.tick_size = Decimal("0.01")
        self._output_size = list(forecaster.parameters())[-1].shape[0]

    def init_state(self, data: MarketData) -> torch.Tensor:
        new_state = self._prism_market_data_to_forecaster_inputs(data)
        state = new_state if new_state is not None else torch.Tensor([])
        return state

    def observe(self, state: torch.Tensor, data: MarketData) -> torch.Tensor:
        new_market_input = self._prism_market_data_to_forecaster_inputs(data)
        if new_market_input is not None:
            # we cast the sequence length as mypy throws a fit with torch Tensors
            length = int(cast(int, self.forecaster.seq_length))
            state = torch.cat([state, new_market_input])[-length:]
        return state

    def _get_new_market_input(self, bid: float, ask: float):
        if self._output_size == 1:
            observation = [(bid + ask) / 2]
        else:
            observation = [bid, ask]
        return torch.Tensor([observation])

    def _prism_market_data_to_forecaster_inputs(self, data: MarketData) -> Optional[torch.Tensor]:
        decimal_bid = get_best_bid(data, tick_size=self.tick_size)
        decimal_ask = get_best_ask(data, tick_size=self.tick_size)
        if decimal_ask is None or decimal_bid is None:
            return None
        new_market_input = self._get_new_market_input(float(decimal_bid), float(decimal_ask))
        return new_market_input

    def forecast(self, state: torch.Tensor, horizon: Optional[int] = None) -> Optional[float]:
        # not enough history to generate a forecast, return None
        # we cast the sequence length as mypy throws a fit with torch Tensors
        if len(state) < int(cast(int, self.forecaster.seq_length)):
            return None
        input_ = {"input": state.unsqueeze(0)}
        forecast = self.forecaster.predict(input_)["pred"][0]

        # the LSTM on the SPYDataset forecasts the bid and ask, we take its mid point.
        mid_price_forecast = float(forecast.mean().detach().cpu().numpy())
        return mid_price_forecast
