import copy
from typing import Any, Dict, Mapping, Optional

import attr
import pytorch_lightning as pl
import torch
from attrs import asdict
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.nn import LSTM, Linear, MSELoss
from torch.optim import Adam

from prism_agents.agents.utils.serialize import SerializableBaseModel


@attr.s(frozen=True, auto_attribs=True)
class LSTMForecastingModuleInputs:
    input_size: int
    hidden_size: int
    num_layers: int = 1
    bias: bool = True
    batch_first: bool = True
    dropout: float = 0.0
    seq_length: int = 1
    output_size: Optional[int] = None


class ForecastingModule(torch.nn.Module):
    def predict(self, batch: Any) -> Any:
        return self(batch)


class LSTMForecastingModule(SerializableBaseModel, ForecastingModule):
    """
    Pytorch based Forecaster based on an LSTM architecture
    """

    def __init__(self, inputs: LSTMForecastingModuleInputs):
        super().__init__()
        lstm_constructor_inputs = asdict(inputs)
        del lstm_constructor_inputs["seq_length"]
        del lstm_constructor_inputs["output_size"]
        self.lstm = LSTM(**lstm_constructor_inputs, bidirectional=False)
        output_size = inputs.output_size if inputs.output_size is not None else inputs.input_size
        self.head = Linear(in_features=inputs.hidden_size, out_features=output_size)

        self.loss = MSELoss()

        # we keep track of the seq_length in the state of our forecaster to use a fixed
        # window of observation with which to make predictions
        self.register_buffer("seq_length", torch.tensor(inputs.seq_length))
        self._constructor_inputs = inputs

    def predict(self, batch: Mapping[str, torch.Tensor]) -> Dict:
        # (batch_size, seq_length, n_features)
        input_ = batch["input"]
        seq_length = input_.shape[1]
        mean = input_.mean(1)
        std = input_.std(1) + 1e-5
        mean_length = mean.unsqueeze(-1).tile(seq_length).transpose(1, -1)
        std_length = std.unsqueeze(-1).tile(seq_length).transpose(1, -1)
        normed_input = (input_ - mean_length) / std_length

        hidden = self.lstm(normed_input)[0]
        # last output of the LSTM encoder.  (batch_size, hidden_size)
        encoded = hidden[:, -1, :]
        output = self.head(encoded)
        output = output * std + mean

        # output has shape (batch_size, n_features)
        return {"pred": output}

    def forward(self, batch: Mapping[str, torch.Tensor]) -> Dict:
        pred = self.predict(batch)["pred"]
        target = batch["target"]
        loss = self.loss(pred, target)
        return {"loss": loss}

    def get_constructor_inputs(self) -> Dict:
        return asdict(self._constructor_inputs)

    @classmethod
    def from_dict(cls, inputs: Mapping) -> "LSTMForecastingModule":
        return cls(LSTMForecastingModuleInputs(**inputs))


class LightningLSTMForecastingModule(pl.LightningModule, ForecastingModule):
    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_layers: int = 1,
        bias: bool = True,
        batch_first: bool = True,
        dropout: float = 0.0,
        seq_length: int = 1,
        lr: float = 1e-3,
        output_size: Optional[int] = None,
    ):
        """
        For torch lightning to reload checkpointed models, it needs to be able to save
        hyperparameters in .yaml form. These are assumed to be in the __init__ constructor;
        a attrs class to describe these inputs do not allow for this capability,
        hence why we're not a direct child of LSTMForecastingModule
        """
        super().__init__()
        self.save_hyperparameters()
        lstm_params = copy.copy(self.hparams)
        del lstm_params["lr"]
        self.model = LSTMForecastingModule.from_dict(lstm_params)
        self.optimizer_parameters = {"lr": lr}

    def forward(self, batch):
        return self.model(batch)

    def predict(self, batch):
        return self.model.predict(batch)

    def training_step(self, batch: Dict, batch_idx: int) -> STEP_OUTPUT:  # type: ignore
        out = self.model(batch)
        self.log("train_loss", out["loss"])
        return out

    def validation_step(self, batch: Dict, batch_idx: int) -> STEP_OUTPUT:  # type: ignore
        out = self.model(batch)
        self.log("val_loss", out["loss"])
        return out

    def configure_optimizers(self):
        optimizer = Adam(self.model.parameters(), **self.optimizer_parameters)
        return optimizer

    @property
    def seq_length(self):
        return self.model.seq_length
