from decimal import Decimal

import attr
from backports.cached_property import cached_property

from prism_agents.agents.forecast.forecast_agent import Forecaster, ForecastRunner
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.api.agent_schemas import MarketData
from prism_agents.api.schema_utility_functions import get_mid_price


@attr.s(frozen=True, auto_attribs=True)
class MovingAverageState:
    short_average: float
    long_average: float
    mid: float


# mypy does not seem to like attrs standard validators
_validator = attr.validators.and_(attr.validators.gt(0), attr.validators.le(1.0))  # type: ignore


@attr.s(frozen=True, auto_attribs=True)
class MovingAverageForecaster(Forecaster[MovingAverageState, float]):
    """
    A forecaster that forecasts based on the common
    technical analysis rule that compares a short-term moving average
    with a long-term moving average.

    alpha_short is the coefficient in the short-term exponentially weighted moving average.
    alpha_long is the coefficient in the long-term exponentially weighted moving average.
    """

    alpha_long: float = attr.ib(validator=_validator)
    alpha_short: float = attr.ib(validator=_validator)
    tick_size: Decimal = Decimal("0.01")

    def __attrs_post_init__(self):
        assert self.alpha_short > self.alpha_long

    @cached_property
    def center_of_mass_short(self) -> float:
        return -(1 / self.alpha_short - 1)

    @cached_property
    def center_of_mass_long(self) -> float:
        return -(1 / self.alpha_long - 1)

    def init_state(self, data: MarketData) -> MovingAverageState:
        mid = float(get_mid_price(data=data, tick_size=self.tick_size))
        return MovingAverageState(short_average=mid, long_average=mid, mid=mid)

    def observe(self, state: MovingAverageState, data: MarketData) -> MovingAverageState:
        mid = float(get_mid_price(data=data, tick_size=self.tick_size))

        return MovingAverageState(
            short_average=(1 - self.alpha_short) * state.short_average + self.alpha_short * mid,
            long_average=(1 - self.alpha_long) * state.short_average + self.alpha_long * mid,
            mid=mid,
        )

    def forecast(self, state: MovingAverageState, horizon: int) -> float:
        slope = (state.short_average - state.long_average) / (
            self.center_of_mass_short - self.center_of_mass_long
        )
        return slope * horizon + state.mid


def simple_moving_average_agent_runner(
    alpha_short: float,
    alpha_long: float,
    horizon: int,
    volume: int,
    alias: str,
    symbol: str = "PRSM",
    period: float = 1.0,
) -> ForecastRunner:
    forecaster = MovingAverageForecaster(alpha_long=alpha_long, alpha_short=alpha_short)
    trading_rule = SimpleTradingRule(volume=volume)
    return ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=horizon,
        symbol=symbol,
        alias=alias,
        period=period,
    )
