import logging
from decimal import Decimal
from typing import Optional, Sequence

import attr
import pmdarima as pm

from prism_agents.agents.forecast.forecast_agent import Forecaster
from prism_agents.api.agent_schemas import MarketData
from prism_agents.api.schema_utility_functions import get_mid_price
from prism_agents.configuration.prism_agents_logging import prism_agents_configure_logging

logger = logging.getLogger(__name__)
logger = prism_agents_configure_logging(logger)


@attr.s(auto_attribs=True)
class AutoArimaState:
    observations: Sequence[float]
    model: Optional[pm.AutoARIMA]


@attr.s(frozen=True, auto_attribs=True)
class AutoArimaForecaster(Forecaster[AutoArimaState, Optional[float]]):
    """
    Forecaster that fits an ARIMA model to mid price history once it has enough
    observations.  This model is used to make forecasts


    After the model has been fit, it is updated with every observation using (update_maxiter)
    number of iterations of the optimizer.
    """

    tick_size: Decimal = Decimal("0.01")
    min_obs_for_forecast: int = 100
    max_obs: int = 200
    update_maxiter: int = 1

    def init_state(self, data: MarketData) -> AutoArimaState:
        return AutoArimaState(
            observations=[float(get_mid_price(data, tick_size=self.tick_size))], model=None
        )

    def forecast(self, state: AutoArimaState, horizon: int) -> Optional[float]:
        if state.model is None:
            return None
        predictions = state.model.predict(n_periods=horizon)
        return predictions[-1]

    def observe(self, state: AutoArimaState, data: MarketData) -> AutoArimaState:
        new_price = float(get_mid_price(data, tick_size=self.tick_size))
        new_observations = (list(state.observations) + [new_price])[-self.max_obs :]
        if state.model is not None:
            logger.debug("updating arima model")
            new_model = state.model.update([new_price], maxiter=self.update_maxiter)
            logger.debug("done updating arima model")
        elif len(new_observations) < self.min_obs_for_forecast:
            new_model = None
        else:
            logger.info("fitting arima model")
            new_model = pm.AutoARIMA().fit(new_observations)
            logger.info("done fitting arima model")

        return AutoArimaState(observations=new_observations, model=new_model)
