from decimal import ROUND_HALF_DOWN, Decimal
from typing import List, Optional, Tuple, Union

import attr

from prism_agents.agents.forecast.forecast_agent import TradingRule
from prism_agents.api.agent_schemas import ClientData, MarketData, OrderRequest, OrderType
from prism_agents.api.schema_utility_functions import get_mid_price, get_spread


@attr.s(frozen=True, auto_attribs=True)
class SimpleTradingRule(TradingRule[None, Optional[float]]):
    """
    Trading rule that buys or sells if the predicted mid price is expected to
    clear the spread. Forecast horizon is ignored.

    forecasts are expected to be floats
    this trading rule is stateless, so the state is of type None

    omega denotes the required multiple of the spread that the forecast must clear
    before we will make a trade
    """

    volume: int
    tick_size: Decimal = Decimal("0.01")
    omega: Decimal = Decimal("1")
    strict: bool = True
    position_max: Union[int, float] = float("inf")
    position_min: Union[int, float] = -float("inf")

    def generate_trades(
        self,
        forecast: Optional[float],
        state: None,
        horizon: int,
        data: MarketData,
        client_data: ClientData,
    ) -> Tuple[List[OrderRequest], None]:
        """
        Selling before buying or buying before selling will cost us the spread
        in transaction costs.  So we should not trade unless we forecast
        that the price movement will compensate this cost.
        """

        if forecast is None:
            # no trades take place if there is no forecast
            return [], None

        spread = get_spread(data=data, tick_size=self.tick_size)
        mid = get_mid_price(data=data, tick_size=self.tick_size, strict=self.strict)
        d_forecast = Decimal(forecast)

        def _quantize(x: Decimal) -> Decimal:
            return Decimal(max(x, self.tick_size)).quantize(
                self.tick_size, rounding=ROUND_HALF_DOWN
            )

        if (
            forecast > mid + self.omega * spread
            and client_data.position + self.volume <= self.position_max
        ):
            return [
                OrderRequest(
                    symbol=data.symbol,
                    quantity=self.volume,
                    type=OrderType.LIMIT,
                    limit_price=float(_quantize(d_forecast - spread / 2)),
                )
            ], None
        if (
            forecast < mid - self.omega * spread
            and client_data.position - self.volume >= self.position_min
        ):
            return [
                OrderRequest(
                    symbol=data.symbol,
                    quantity=-self.volume,
                    type=OrderType.LIMIT,
                    limit_price=float(_quantize(d_forecast + spread / 2)),
                )
            ], None
        return [], None

    def init_state(self) -> None:
        pass
