from decimal import ROUND_HALF_DOWN, Decimal
from typing import List, Optional, Tuple, Union

import attr

from prism_agents.agents.forecast.forecast_agent import TradingRule
from prism_agents.api.agent_schemas import ClientData, MarketData, OrderRequest, OrderType


@attr.s(frozen=True, auto_attribs=True)
class SpreadTradingRule(TradingRule[None, Optional[float]]):
    """
    Trading rule that submits bid and ask orders around projected price.
    Forecast horizon is ignored

    forecasts are expected to be floats
    this trading rule is stateless, so the state is of type None
    """

    volume: int
    tick_size: Decimal = Decimal("0.01")
    strict: bool = True
    position_max: Union[int, float] = float("inf")
    position_min: Union[int, float] = -float("inf")
    omega: Decimal = Decimal("1")

    def generate_trades(
        self,
        forecast: Optional[float],
        state: None,
        horizon: int,
        data: MarketData,
        client_data: ClientData,
    ) -> Tuple[List[OrderRequest], None]:
        if forecast is None:
            # no trades take place if there is no forecast
            return [], None

        d_forecast = Decimal(forecast)

        def _quantize(x: Decimal) -> Decimal:
            return Decimal(x).quantize(self.tick_size, rounding=ROUND_HALF_DOWN)

        trades = []
        sell_price = _quantize(d_forecast + self.omega * self.tick_size / 2)
        buy_price = _quantize(d_forecast - self.omega * self.tick_size / 2)
        volume_after_sell = client_data.position - self.volume
        volume_after_buy = client_data.position + self.volume
        if sell_price > 0 and volume_after_sell >= self.position_min:
            trades.append(
                OrderRequest(
                    symbol=data.symbol,
                    quantity=-self.volume,
                    type=OrderType.LIMIT,
                    limit_price=float(sell_price),
                )
            )
        if buy_price > 0 and volume_after_buy <= self.position_max:
            trades.append(
                OrderRequest(
                    symbol=data.symbol,
                    quantity=self.volume,
                    type=OrderType.LIMIT,
                    limit_price=float(buy_price),
                )
            )
        return trades, None

    def init_state(self) -> None:
        pass
