from decimal import ROUND_HALF_DOWN, Decimal
from typing import Iterator, List, Optional, Tuple

import attr
import numpy as np

from prism_agents.agents.forecast.forecast_agent import TradingRule
from prism_agents.api.agent_schemas import ClientData, MarketData, OrderRequest, OrderType


@attr.s(frozen=True, auto_attribs=True)
class VolumeProfileTradingRule(TradingRule[None, Optional[float]]):
    """
    Trading rule that submits trades on the bid and ask side of a forecast
    mid price by sampling from given distributions
    """

    ask_profile_dist: Iterator[np.ndarray]
    bid_profile_dist: Iterator[np.ndarray]
    tick_size: Decimal = Decimal("0.01")
    strict: bool = True

    def generate_trades(
        self,
        forecast: Optional[float],
        state: None,
        horizon: int,
        data: MarketData,
        client_data: ClientData,
    ) -> Tuple[List[OrderRequest], None]:
        if forecast is None:
            # no trades take place if there is no forecast
            return [], None

        d_forecast = Decimal(forecast)

        def _quantize(x: Decimal) -> Decimal:
            return Decimal(x).quantize(self.tick_size, rounding=ROUND_HALF_DOWN)

        trades = []
        ask_base = _quantize(d_forecast + self.tick_size / 2)
        bid_base = _quantize(d_forecast - self.tick_size / 2)

        ask_profile = next(self.ask_profile_dist)
        bid_profile = next(self.bid_profile_dist)
        for i in range(len(ask_profile)):
            if ask_profile[i] > 0:
                trades.append(
                    OrderRequest(
                        symbol=data.symbol,
                        quantity=-ask_profile[i],
                        type=OrderType.LIMIT,
                        limit_price=float(ask_base + self.tick_size * i),
                    )
                )
        for i in range(len(bid_profile)):
            if bid_profile[i] > 0:
                trades.append(
                    OrderRequest(
                        symbol=data.symbol,
                        quantity=bid_profile[i],
                        type=OrderType.LIMIT,
                        limit_price=float(bid_base - self.tick_size * i),
                    )
                )

        return trades, None

    def init_state(self) -> None:
        pass


def peaked_poisson_rates(
    expected_vol: int, peak: int = 2, decay_exponent: float = 0.5, sz: int = 25
) -> np.ndarray:
    """
    Poisson rates for a volume distribution
    The rate parameters of these distributions decay according to a power law
    """
    rate_profile = np.ones(dtype=np.int64, shape=(sz,))
    decay_denominator = (abs(peak - np.arange(sz)) + 1) ** decay_exponent
    rate_profile = rate_profile / decay_denominator
    rate_profile = rate_profile / sum(rate_profile) * expected_vol
    return rate_profile


def two_part_peaked_poisson_rates(
    expected_vol_inner: int,
    expected_vol_outer: int,
    peak_inner: int = 2,
    decay_exponent_inner: float = 0.5,
    sz_inner: int = 25,
    peak_outer: int = 2,
    decay_exponent_outer: float = 0.5,
    sz_outer: int = 25,
) -> np.ndarray:
    return np.concatenate(
        [
            peaked_poisson_rates(
                expected_vol=expected_vol_inner,
                peak=peak_inner,
                decay_exponent=decay_exponent_inner,
                sz=sz_inner,
            ),
            peaked_poisson_rates(
                expected_vol=expected_vol_outer,
                peak=peak_outer,
                decay_exponent=decay_exponent_outer,
                sz=sz_outer,
            ),
        ]
    )


def poisson_peaked_profile(
    expected_vol: int, peak: int = 2, decay_exponent: float = 0.5, sz: int = 25, seed: int = 42
) -> Iterator[np.ndarray]:
    """
    Volume distribution in which the volume at any price level is Poisson distributed.
    The rate parameters of these distributions decay according to a power law
    """
    generator = np.random.default_rng(seed=seed)
    rate_profile = peaked_poisson_rates(
        expected_vol=expected_vol, peak=peak, decay_exponent=decay_exponent, sz=sz
    )
    while True:
        yield generator.poisson(lam=rate_profile)
