from typing import Optional, Sequence
from uuid import UUID

import numpy as np

from prism_agents.agents.background.annealing_schedule import AnnealingState, ConstantAnnealingState
from prism_agents.agents.background.poisson_fundamental_trader import (
    PoissonRegion,
    create_multi_part_poisson_fundamentalist,
    create_poisson_fundamentalist,
    create_two_part_poisson_fundamentalist,
)
from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import Fundamentalist
from prism_agents.agents.forecast.forecasters.fundamentalist.process.ornstein_uhlenbeck import (
    OrnsteinUhlenbeckState,
)
from prism_agents.agents.forecast.trading_rule.spread_trading_rule import SpreadTradingRule
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner


def create_basic_ou_fundamental_agent(
    seed: int,
    sigma: float,
    half_life: float,
    period: float,
    mean: float = 10,
) -> ForecastRunner:
    """
    Utility function for ergonomically creating a mean reverting fundamental trader
    """
    state = OrnsteinUhlenbeckState.create_with_half_life(
        half_life=half_life,
        sigma=sigma,
        step=0,
        value=mean,
        b=mean,
        generator=np.random.default_rng(seed),
    )
    rule = SpreadTradingRule(
        volume=10,
    )
    forecast = Fundamentalist(start_state=state)

    ou_runner = ForecastRunner(
        forecaster=forecast,
        trading_rule=rule,
        horizon=1,
        symbol="PRSM",
        alias="fundamentalist",
        period=period,
    )
    return ou_runner


def create_ou_poisson_agent(
    seed: int,
    sigma: float,
    half_life: float,
    fundamental_period: float,
    limit_arrival_rate: int,
    mean: float = 10,
    cancel_factor: float = 0.5,
    alias="fundamental_trader",
    annealer: AnnealingState = ConstantAnnealingState(),
    client_id: Optional[UUID] = None,
    symbol: str = "PRSM",
):
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    state = OrnsteinUhlenbeckState.create_with_half_life(
        half_life=half_life,
        sigma=sigma,
        step=0,
        value=mean,
        b=mean,
        generator=np.random.default_rng(seed_1),
    )
    return create_poisson_fundamentalist(
        initial_state=state,
        limit_arrival_rate=limit_arrival_rate,
        fundamental_period=fundamental_period,
        alias=alias,
        seed=seed_2.generate_state(1)[0],
        cancel_factor=cancel_factor,
        annealer=annealer,
        cliend_id=client_id,
        symbol=symbol,
    )


def create_poisson_ou_fundamental_agent_two_part(
    seed: int,
    inner_anneal_half_life: float,
    sigma: float,
    half_life: float,
    mean: float = 10,
    alias: str = "fundamental",
    fundamental_period: float = 1.0,
    cancel_factor: float = 0.5,
    inner_sz: int = 15,
    outer_sz: int = 10,
    vol_inner: int = 1,
    vol_outer: int = 10,
    inner_limit_arrival_rate=10,
    outer_limit_arrival_rate=3,
) -> StepwiseAgentRunner:
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    process_state = OrnsteinUhlenbeckState.create_with_half_life(
        half_life=half_life,
        sigma=sigma,
        step=0,
        value=mean,
        b=mean,
        generator=np.random.default_rng(seed_1),
    )
    return create_two_part_poisson_fundamentalist(
        fundamental_period=fundamental_period,
        initial_state=process_state,
        alias=alias,
        cancel_factor=cancel_factor,
        inner_sz=inner_sz,
        outer_sz=outer_sz,
        vol_inner=vol_inner,
        vol_outer=vol_outer,
        seed=seed_2.generate_state(1)[0],
        inner_anneal_half_life=inner_anneal_half_life,
        inner_limit_arrival_rate=inner_limit_arrival_rate,
        outer_limit_arrival_rate=outer_limit_arrival_rate,
    )


def create_poisson_ou_fundamental_agent_multi_part(
    seed: int,
    sigma: float,
    half_life: float,
    regions: Sequence[PoissonRegion],
    mean: float = 10,
    alias: str = "fundamental",
    fundamental_period: float = 1.0,
    cancel_factor: float = 0.5,
) -> StepwiseAgentRunner:
    """
    Convenience method to create an OU poisson agent
     whose order book is divided into arbitrarily many regions.
    """
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    process_state = OrnsteinUhlenbeckState.create_with_half_life(
        half_life=half_life,
        sigma=sigma,
        step=0,
        value=mean,
        b=mean,
        generator=np.random.default_rng(seed_1),
    )
    return create_multi_part_poisson_fundamentalist(
        initial_state=process_state,
        alias=alias,
        fundamental_period=fundamental_period,
        cancel_factor=cancel_factor,
        regions=regions,
        seed=seed_2.generate_state(1)[0],
    )
