from typing import Sequence, Tuple

import numpy as np

from prism_agents.agents.background.annealing_schedule import AnnealingState, ConstantAnnealingState
from prism_agents.agents.background.poisson_fundamental_trader import (
    PoissonRegion,
    create_multi_part_poisson_fundamentalist,
    create_poisson_fundamentalist,
    create_two_part_poisson_fundamentalist,
)
from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import (
    FundamentalState,
    create_spread_fundamentalist,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.ornstein_uhlenbeck import (
    ou_transition_function,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.regime_switching import (
    RegimeSwitchingState,
)
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner


def create_simple_ou_regime_switching_agent(
    seed: int,
    sigma: Tuple[float, float, float],
    alpha: Tuple[float, float, float],
    period: float,
    mean: Tuple[float, float, float],
    transition_prob: float = 0.01,
) -> ForecastRunner:
    """
    Function for ergonomically creating a
    fundamental trader that regime switches
    between various mean reversion levels.

    """
    matrix = np.array(
        [
            [1 - 2 * transition_prob, transition_prob, transition_prob],
            [transition_prob, 1 - 2 * transition_prob, transition_prob],
            [transition_prob, transition_prob, 1 - 2 * transition_prob],
        ]
    )
    switching_trader = create_spread_fundamentalist(
        initial_state=RegimeSwitchingState.create(
            transition_matrix=matrix,
            regime_transition_functions=[
                ou_transition_function(sigma=sigma[0], alpha=alpha[0], b=mean[0]),
                ou_transition_function(sigma=sigma[1], alpha=alpha[1], b=mean[1]),
                ou_transition_function(sigma=sigma[2], alpha=alpha[2], b=mean[2]),
            ],
            starting_value=mean[0],
            seed=seed,
        ),
        volume=10,
        period=period,
    )
    return switching_trader


def make_switching_state(
    sigma: Tuple[float, float, float],
    alpha: Tuple[float, float, float],
    mean: Tuple[float, float, float],
    seed: int,
    transition_prob: float = 0.01,
) -> FundamentalState:
    matrix = np.array(
        [
            [1 - 2 * transition_prob, transition_prob, transition_prob],
            [transition_prob, 1 - 2 * transition_prob, transition_prob],
            [transition_prob, transition_prob, 1 - 2 * transition_prob],
        ]
    )
    state = RegimeSwitchingState.create(
        transition_matrix=matrix,
        regime_transition_functions=[
            ou_transition_function(sigma=sigma[0], alpha=alpha[0], b=mean[0]),
            ou_transition_function(sigma=sigma[1], alpha=alpha[1], b=mean[1]),
            ou_transition_function(sigma=sigma[2], alpha=alpha[2], b=mean[2]),
        ],
        seed=seed,
        starting_value=mean[0],
    )
    return state


def create_poisson_ou_regime_switching_agent(
    seed: int,
    sigma: Tuple[float, float, float],
    alpha: Tuple[float, float, float],
    mean: Tuple[float, float, float],
    limit_arrival_rate: int = 10,
    fundamental_period: float = 1.0,
    transition_prob: float = 0.01,
    annealer: AnnealingState = ConstantAnnealingState(),
) -> StepwiseAgentRunner:
    """
    Function for ergonomically creating a poisson
     fundamental trader that regime switches
    between various mean reversion levels.

    """
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    state = make_switching_state(
        sigma=sigma,
        alpha=alpha,
        mean=mean,
        transition_prob=transition_prob,
        seed=seed_1.generate_state(1)[0],
    )
    switching_trader = create_poisson_fundamentalist(
        initial_state=state,
        limit_arrival_rate=limit_arrival_rate,
        fundamental_period=fundamental_period,
        seed=seed_2.generate_state(1)[0],
        annealer=annealer,
    )
    return switching_trader


def create_poisson_ou_regime_switching_fundamental_agent_two_part(
    seed: int,
    inner_anneal_half_life: float,
    sigma: Tuple[float, float, float],
    alpha: Tuple[float, float, float],
    mean: Tuple[float, float, float],
    transition_prob: float = 0.01,
    alias: str = "fundamental",
    fundamental_period: float = 1.0,
    cancel_factor: float = 0.5,
    inner_sz: int = 15,
    outer_sz: int = 10,
    vol_inner: int = 1,
    vol_outer: int = 10,
    inner_limit_arrival_rate=10,
    outer_limit_arrival_rate=3,
) -> StepwiseAgentRunner:
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    state = make_switching_state(
        sigma=sigma,
        alpha=alpha,
        mean=mean,
        transition_prob=transition_prob,
        seed=seed_1.generate_state(1)[0],
    )
    return create_two_part_poisson_fundamentalist(
        fundamental_period=fundamental_period,
        initial_state=state,
        alias=alias,
        cancel_factor=cancel_factor,
        inner_sz=inner_sz,
        outer_sz=outer_sz,
        vol_inner=vol_inner,
        vol_outer=vol_outer,
        seed=seed_2.generate_state(1)[0],
        inner_anneal_half_life=inner_anneal_half_life,
        inner_limit_arrival_rate=inner_limit_arrival_rate,
        outer_limit_arrival_rate=outer_limit_arrival_rate,
    )


def create_poisson_ou_regime_switching_fundamental_agent_multi_part(
    seed: int,
    sigma: Tuple[float, float, float],
    alpha: Tuple[float, float, float],
    mean: Tuple[float, float, float],
    regions: Sequence[PoissonRegion],
    transition_prob: float = 0.01,
    alias: str = "fundamental",
    fundamental_period: float = 1.0,
    cancel_factor: float = 0.5,
) -> StepwiseAgentRunner:
    """
    Convenience method to create an ou regime switching poisson agent
    whose order book is divided into arbitrarily many regions.
    """
    seed_1, seed_2 = np.random.SeedSequence(seed).spawn(2)
    state = make_switching_state(
        sigma=sigma,
        alpha=alpha,
        mean=mean,
        transition_prob=transition_prob,
        seed=seed_1.generate_state(1)[0],
    )
    return create_multi_part_poisson_fundamentalist(
        initial_state=state,
        alias=alias,
        fundamental_period=fundamental_period,
        cancel_factor=cancel_factor,
        regions=regions,
        seed=seed_2.generate_state(1)[0],
    )
