from typing import Optional

import numpy as np
import numpy.random


def vector_arithmetic_brownian_motion(
    num_paths: int,
    mu: np.ndarray,
    sigma: np.ndarray,
    C: np.ndarray,
    step_sizes: np.ndarray,
    start_value: Optional[np.ndarray] = None,
    random_generator: Optional[np.random.Generator] = None,
) -> np.ndarray:
    """
    Simulate a given number of paths of vector arithmetic brownian motion with n assets
    args:
        num_paths: number of paths to generate
        mu:  array of dimension (n,) indicating arithmetic drift
        sigma: array of dimension (n,) indicating arithmetic volatility
        C: array of dimension (n,n) such that CC^T is the correlation between assets
        step_sizes: shape (num_steps,) indicating time step sizes
        start_value: array of dimension (n,) indicating initial value
        random_generator: numpy random number generator
    returns:
        array of dimension (num_paths, num_steps,n) giving simulated values of Brownian Motion
    """
    (num_steps,) = step_sizes.shape
    if random_generator is None:
        random_generator = np.random.default_rng()
    if start_value is None:
        start_value = np.zeros_like(mu)
    n = mu.size
    C = C.reshape(n, n)
    drift_array = mu.reshape((1, 1, n)) * (
        step_sizes.cumsum(axis=0).reshape(1, num_steps, 1)
    )  # should be paths by states by assets
    correlated_normals = (C * random_generator.normal(size=(num_paths, num_steps, 1, n))).sum(
        -1
    )  # dimension paths by steps by n
    shocks = (
        sigma.reshape((1, 1, n)) * np.sqrt(step_sizes.reshape(1, num_steps, 1)) * correlated_normals
    )
    cum_shocks = np.cumsum(shocks, axis=1)
    return start_value + drift_array + cum_shocks


def vector_geometric_brownian_motion(
    num_paths: int,
    start_value: np.ndarray,
    mu: np.ndarray,
    sigma: np.ndarray,
    C: np.ndarray,
    step_sizes: np.ndarray,
    random_generator: np.random.Generator,
) -> np.ndarray:
    """
    Simulate a given number of paths of vector geometric brownian motion with n assets
    args:
        num_paths: number of paths to generate
        start_value : shape (num_paths,n)  indicating starting value for all assets'
        mu:  array of dimension (n,) indicating geometric drift
        sigma: array of dimension (n,) indicating volatility
        C: array of dimension (n,n) such that CC^T is the correlation between assets
        step_sizes: shape (num_steps,) indicating time step sizes
    returns:
        array of dimension (num_paths, num_steps,n) giving simulated values of Brownian Motion
    """
    arithmetic_drift = mu - (sigma**2) / 2
    n = mu.size
    abm = vector_arithmetic_brownian_motion(
        num_paths=num_paths,
        mu=arithmetic_drift,
        sigma=sigma,
        step_sizes=step_sizes,
        random_generator=random_generator,
        C=C,
    )
    return start_value.reshape((num_paths, 1, n)) * np.exp(abm)
