from typing import Callable, Optional

import numpy as np
import numpy.random

from prism_agents.agents.utils.vector_simulation.brownian import vector_arithmetic_brownian_motion

jump_function = Callable[[np.ndarray, np.random.Generator], np.ndarray]


def vector_arithmetic_generic_poisson_jump(
    num_paths: int,
    poisson_rates: np.ndarray,
    jump_func: jump_function,
    step_sizes: np.ndarray,
    generator: np.random.Generator,
) -> np.ndarray:
    """

    Args:
        num_paths:  number of paths to simulate
        poisson_rates: array of dimension (n,) indicating the
                      poisson rate for each process
        jump_func: function accepting an array of dimension
                   (num_paths,num_steps, n) indicating the number of
                   jumps of each process on each path and
                   returning the total jump size of each process on each path
        step_sizes: array of dimension ( num_steps,) indicating the sizes of the time steps
        generator: random generator

    Returns: array of dimension  (num_paths, num_steps, n) indicating
            value of jump process
    """
    (num_steps,) = step_sizes.shape
    (n,) = poisson_rates.shape
    effective_rate = (poisson_rates.reshape(1, n) * step_sizes.reshape(num_steps, 1)).reshape(
        1, num_steps, n
    )
    jump_count = generator.poisson(lam=effective_rate, size=(num_paths, num_steps, n))
    jump_sizes = jump_func(jump_count, generator)
    return jump_sizes.cumsum(axis=1)


def vector_arithmetic_merton_pure_jump(
    num_paths: int,
    poisson_rates: np.ndarray,
    jump_size_mean: np.ndarray,
    jump_size_std: np.ndarray,
    step_sizes: np.ndarray,
    generator: np.random.Generator,
) -> np.ndarray:
    """

    Simulate a jump process  in which jump sizes are Gaussian

    Args:
        num_paths: intenger number of paths to generate
        poisson_rates: (n,) poisson rate for each process
        jump_size_mean: (n,) mean jump size
        jump_size_std: (n,) Standard deviation jump size
        step_sizes: array of dimension ( num_steps,) indicating the sizes of the time steps
        generator: random generator

    Returns:
        array of process values




    """
    n = len(poisson_rates)

    def merton_jump_function(num_jumps: np.ndarray, generator: np.random.Generator) -> np.ndarray:
        means = num_jumps * jump_size_mean.reshape(1, 1, n)
        stds = np.sqrt(num_jumps) * jump_size_std.reshape(1, 1, n)

        return generator.normal(loc=means, scale=stds)

    result = vector_arithmetic_generic_poisson_jump(
        num_paths=num_paths,
        poisson_rates=poisson_rates,
        jump_func=merton_jump_function,
        step_sizes=step_sizes,
        generator=generator,
    )

    return result


def vector_arithmetic_merton_jump(
    num_paths: int,
    mu: np.ndarray,
    sigma: np.ndarray,
    C: np.ndarray,
    step_sizes: np.ndarray,
    poisson_rates: np.ndarray,
    jump_size_mean: np.ndarray,
    jump_size_std: np.ndarray,
    random_generator: np.random.Generator,
    start_value: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Simulate a jump diffusion process process in which jumps are Gaussain
    and diffusion is Brownian motion
    """
    abm = vector_arithmetic_brownian_motion(
        num_paths=num_paths,
        mu=mu,
        sigma=sigma,
        step_sizes=step_sizes,
        random_generator=random_generator,
        C=C,
        start_value=start_value,
    )
    jumps = vector_arithmetic_merton_pure_jump(
        num_paths=num_paths,
        poisson_rates=poisson_rates,
        jump_size_mean=jump_size_mean,
        jump_size_std=jump_size_std,
        step_sizes=step_sizes,
        generator=random_generator,
    )
    return abm + jumps


def vector_geometric_merton_jump(
    num_paths: int,
    mu: np.ndarray,
    sigma: np.ndarray,
    C: np.ndarray,
    step_sizes: np.ndarray,
    poisson_rates: np.ndarray,
    jump_size_mean: np.ndarray,
    jump_size_std: np.ndarray,
    random_generator: np.random.Generator,
    start_value: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Simulate a jump-diffusion process in which jumps are lognormal
    and diffusionn is geometric Brown motion

    """
    arithmetic_drift = mu - (sigma**2) / 2
    n = mu.size
    arith = vector_arithmetic_merton_jump(
        num_paths=num_paths,
        mu=arithmetic_drift,
        sigma=sigma,
        C=C,
        step_sizes=step_sizes,
        poisson_rates=poisson_rates,
        jump_size_std=jump_size_std,
        jump_size_mean=jump_size_mean,
        random_generator=random_generator,
    )
    if start_value is None:
        start_value = np.ones((num_paths, n))
    return start_value.reshape((num_paths, 1, n)) * np.exp(arith)
