from typing import Tuple

import numpy as np

from prism_agents.agents.utils.vector_simulation.brownian import vector_arithmetic_brownian_motion


def vector_arithmetic_kim_omberg(
    num_paths: int,
    start_value: np.ndarray,
    start_drift: np.ndarray,
    sigma: np.ndarray,
    drift_alpha: np.ndarray,
    drift_b: np.ndarray,
    drift_sigma: np.ndarray,
    step_sizes: np.ndarray,
    random_generator: np.random.Generator,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Simulate a given number of paths of a Kim-Omberg process with n assets.
    The Kim-Omberg process is an Ito Process in which the drift is mean-reverting

    Note that we use a simple Euler discretization of this process,
    meaning that using more steps will increase accuracy.
    args:
        num_paths: number of paths to generate
        start_drift:  array of dimension (n,) indicating initial arithmetic drift
        sigma: array of dimension (n,) indicating arithmetic volatility
        drift_alpha: array of dimension(n,) indicating speed of mean reversion
                     of drift
        drift_b: array of dimension (n,) indicating the long term average of the drift
        drift_sigma: array of dimension (n,) indicating the volatility of the drift
        step_sizes: shape (num_steps,) indicating time step sizes
        start_value: array of dimension (n,) indicating initial value
        random_generator: numpy random number generator
    returns:
        array of dimension (num_paths, num_steps,n) giving simulated values of Brownian Motion
        array of dimension (num_paths, num_steps, n) giving values of drift
    """
    num_steps = len(step_sizes)
    n = len(sigma)
    cur_state = start_value.reshape((1, n))
    cur_drift = start_drift.reshape((1, n))
    b = drift_b.reshape((1, n))
    history = []
    drift_history = []
    for i in range(num_steps):
        cur_state = (
            cur_state
            + vector_arithmetic_brownian_motion(
                num_paths=num_paths,
                mu=np.zeros_like(sigma),
                sigma=sigma,
                C=np.zeros((n, n)),
                step_sizes=step_sizes[i].reshape((1,)),
                random_generator=random_generator,
            )[:, 0, :]
            + cur_drift * step_sizes[i]
        )
        history.append(cur_state)
        decay_factor = np.exp(-drift_alpha * step_sizes[i])
        drift_mean = decay_factor * cur_drift + (1 - decay_factor) * b
        drift_std = drift_sigma * np.sqrt(1 / (2 * drift_alpha) * (1 - decay_factor**2))
        cur_drift = drift_mean + drift_std * random_generator.standard_normal(size=(num_paths, n))
        drift_history.append(cur_drift)
    return np.concatenate(
        [item.reshape(num_paths, 1, n) for item in history], axis=1
    ), np.concatenate([item.reshape(num_paths, 1, n) for item in drift_history], axis=1)
