import numpy as np


def ou_process(
    seq_length: int,
    step_size: int,
    half_life: np.ndarray,
    initial_values: np.ndarray,
    b: np.ndarray,
    sigma: np.ndarray,
):
    """
    Vectorized way of creating paths of an Ornstein Uhlenbeck process

    See Section 3.3 in Glasserman "Monte Carlo Methods in Financial Engineering"

    Process evolves according to the SDE


    dr(t) = alpha*(b-r(t))dt + sigma*dW(t)
    where W is Brownian motion
    """
    alpha = np.log(2) / half_life
    n_paths = len(alpha)
    draws = np.random.normal(size=(n_paths, seq_length))
    decay_factor = np.exp(-alpha * step_size)
    # (paths,)
    std = sigma * np.sqrt(1 / (2 * alpha) * (1 - decay_factor**2))
    # (paths, seq_length)
    noise = (draws.T * std).T

    values_seq = [initial_values]
    for i in range(seq_length):
        values = values_seq[-1]
        mean = values * decay_factor + b * (1 - decay_factor)
        values = mean + noise[:, i]
        values_seq.append(values)

    values_seq_np = np.stack(values_seq).T

    return values_seq_np
