import logging
import time
from decimal import Decimal
from typing import List, Optional, Tuple
from uuid import UUID

import numpy as np

from prism_agents.api.server_interface import ServerInterface
from prism_agents.dataset.base import MarketDataset, TimestepInfo
from prism_agents.runner.agent_runner import AgentRunner

logger = logging.getLogger(__name__)


class MarketReplay(AgentRunner):
    """
    Agent that sends bid and ask orders to PRISM based on the dynamics of some underlying
    dataset as described by MarketDataset. This effectively replays the market dynamics
    of the underlying dataset. Note that this does not need to be a dataset coming from a
    Prism episode. Rather this is a generic class that replays some dynamics of the bid/ask
    with strong assumptions on the depth of the order book.

    We use the LOB acronym to describe the Limit Order Book (LOB)

    At period instances in time, the agent posts new orders by
    1) requesting an updated set of bid/ask orders from PRISM
    2) computing the last delta in price movements of some dataset
    3) recording the dataset's spread and trade volume in the dataset's last market state
       (call this variable whose has an integer value "TradeVol")
    3) adding bids/asks with quantity TradeVol at levels
            last_prism_midpoint + delta + i * spread
       with -LOB_depth < i < LOB_depth and i<0 for bid,  i>0 for asks.
    """

    def __init__(
        self,
        dataset: MarketDataset,
        ticker_symbol: str = "PRSM",
        time_init: int = 0,
        LOB_depth: int = 10,
        client_alias: str = "MarketReplay",
        max_run_time: int = 100,
        order_request_delay: int = 10,
        use_last_mid_point: bool = True,
        client_id: Optional[UUID] = None,
    ):
        """
        :param dataset: Dataset that will be used to generate dynamics
        :param ticker_symbol: Ticker symbol to send to the PRISM API
        :param time_init: datasets are ordered by an time integer,
                          this initializes our starting point
        :param LOB_depth: Number of limit orders to send at various prices
        :param client_alias: Alias of the agent displayed to PRISM (max 16 char)
                             This is ignored if a client_id is supplied.
        :param max_run_time: Maximum number of requests sent to PRISM before shutting down
                             Set equal to -1 if infinite.
        :param order_request_delay: Time in second before subsequent new quotes to PRISM
        :param use_last_mid_point: If false, send orders based on the dataset's current stock value
                                   otherwise use the last seen midpoint.
        :param client_id: If given, post prism orders under this client_id.
        """
        self.dataset = dataset
        self._mod = len(dataset)
        self.ticker_symbol = ticker_symbol
        self.LOB_depth = LOB_depth
        self._max_run_time = max_run_time
        self._order_request_delay = order_request_delay
        self.use_last_mid_point = use_last_mid_point
        self.client_alias = client_alias

        self._last_dataset_timestep_info = self._get_dataset_timestep_info(time_init)
        self.client_id = client_id

    def _get_dataset_timestep_info(self, timestep: int) -> TimestepInfo:
        timestep, info = self.dataset.get_next_market_state(timestep)
        return info

    def _get_prism_mid_price(self, server_interface: ServerInterface) -> Optional[float]:
        market_data = server_interface.get_market_data(self.ticker_symbol)

        ask = None
        if len(market_data.asks) > 0:
            ask = np.min([data.price for data in market_data.asks])

        bid = None
        if len(market_data.bids) > 0:
            bid = np.max([data.price for data in market_data.bids])

        if bid and ask:
            return (bid + ask) / 2

        # if only bid or ask exist, return the one existing or return None
        return bid or ask

    def _create_new_replay_orders(
        self, prism_mid_price: Optional[float]
    ) -> Tuple[List[float], List[int]]:
        # Get the mid point movement from historical data, along with its spread
        timestep_info = self._get_dataset_timestep_info(self._last_dataset_timestep_info.timestep)
        next_mid = (timestep_info.ask + timestep_info.bid) / 2
        last_mid = (self._last_dataset_timestep_info.ask + self._last_dataset_timestep_info.bid) / 2
        mid_movement = next_mid - last_mid
        dataset_spread = (timestep_info.ask - timestep_info.bid) / 2

        if dataset_spread < 0:
            # The datasets technically shouldn't have this occurring, but they sometimes do
            # Negative spreads would be refused by PRISM, so we fix it here if it occurs
            logger.warning(
                f"The spread seen in the dataset at timestep "
                f"{timestep_info.timestep} is not positive."
            )
            dataset_spread = abs(dataset_spread)

        # sometimes prism has no mid price (e.g. on restart), so use the dataset's mid price
        prism_mid_price = prism_mid_price or last_mid
        # Update the mid price by the seen market movement
        mid = next_mid
        if self.use_last_mid_point:
            mid = prism_mid_price + mid_movement

        # Add various prices based on the market spread, with size of the market trade volume
        prices = []
        volumes = []
        for depth_of_the_limit_order in range(1, self.LOB_depth + 1):
            bid_price = max(mid - depth_of_the_limit_order * dataset_spread, 0.01)
            prices.append(bid_price)
            volumes.append(timestep_info.volume)

            ask_price = mid + depth_of_the_limit_order * dataset_spread
            prices.append(ask_price)
            volumes.append(-timestep_info.volume)

        self._last_dataset_timestep_info = timestep_info
        return prices, volumes

    def run(
        self,
        server_interface: ServerInterface,
        time_warp_factor: float = 1.0,
        client_id: Optional[UUID] = None,
    ):
        client_id = client_id or self.client_id
        if client_id is None:
            client_id = server_interface.register(self.client_alias)

        while self._max_run_time != 0:

            mid = self._get_prism_mid_price(server_interface)

            prices, volumes = self._create_new_replay_orders(mid)

            for price, volume in zip(prices, volumes):
                server_interface.limit_order(
                    symbol=self.ticker_symbol,
                    client_id=client_id,
                    volume=volume,
                    price=Decimal(price),
                )

            self._max_run_time -= 1
            time.sleep(self._order_request_delay / time_warp_factor)
