from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, List, Optional
from uuid import UUID

import attr

"""
This module contains types for representing requests to and responses from
the prism api.
It mirrors  prism/api.py
"""


@attr.s(auto_attribs=True, frozen=True)
class Client:
    alias: str
    first_name: str = ""
    last_name: str = ""


class OrderType(Enum):
    MARKET = "market"
    LIMIT = "limit"
    STOP = "stop"
    STOPLIMIT = "stoplimit"


class OrderExpiration(Enum):
    FILLORKILL = "fok"
    SHORT = "short"
    DAY = "day"
    GOODTILLCANCEL = "gtc"


class Ticker(Enum):
    PRSM = "PRSM"

    @classmethod
    def valid_ticker(cls, module_name: str) -> bool:
        try:
            Ticker(module_name)
            return True
        except ValueError:
            return False


@attr.s(auto_attribs=True, frozen=True)
class Order:
    id: UUID
    open: datetime
    client: UUID
    symbol: Ticker
    status: str
    price: float
    volume: int


@attr.s(auto_attribs=True, frozen=True)
class Trade:
    id: UUID
    date: datetime
    sender: UUID
    receiver: UUID
    symbol: Ticker
    price: float
    volume: int


@attr.s(frozen=True, auto_attribs=True)
class ClientOrders:
    client_id: UUID
    timestamp: datetime
    orders: List[Order]


@attr.s(frozen=True, auto_attribs=True)
class AggregateOrders:
    price: float
    volume: int


@attr.s(frozen=True, auto_attribs=True)
class MarketData:
    symbol: str
    asks: List[AggregateOrders]
    bids: List[AggregateOrders]
    last_trade_price: float
    last_trade_volume: int
    timestamp: datetime = attr.ib(eq=False)


def _convert_to_int(x) -> int:
    return int(x)


@attr.s(frozen=True, auto_attribs=True)
class OrderRequest:
    symbol: str
    quantity: int = attr.ib(converter=_convert_to_int)
    type: OrderType
    limit_price: Optional[float] = None
    expiration: OrderExpiration = OrderExpiration.DAY
    skip_constraints_key: Optional[str] = None


@attr.s(frozen=True, auto_attribs=True)
class LastTrades:
    timestamp: datetime
    trades: List[Trade]


@attr.s(frozen=True, auto_attribs=True)
class ClientData:
    alias: str
    position: int
    costbasis: Decimal
    realizedpl: Decimal
    timestamp: datetime = attr.ib(eq=False)
    outstanding_bid_volume: int
    outstanding_ask_volume: int
    wealth: Decimal


@attr.s(frozen=True, auto_attribs=True)
class ClientDataTable:
    data: List[ClientData]
    timestamp: Optional[datetime] = None

    def to_dict(self) -> Dict[str, ClientData]:
        return {dat.alias: dat for dat in self.data}
