import uuid
from collections import Counter
from copy import deepcopy
from datetime import datetime
from decimal import Decimal
from math import copysign
from typing import Dict, Tuple
from uuid import UUID

import attr

from prism_agents.api.agent_schemas import (
    AggregateOrders,
    ClientData,
    ClientDataTable,
    ClientOrders,
    MarketData,
    Order,
    OrderRequest,
    OrderType,
    Ticker,
)
from prism_agents.api.local_server.order_store import OrderStore
from prism_agents.api.local_server.simple_priority_queue import SimplePriorityQueue
from prism_agents.api.server_interface import ServerInterface

OPEN = "open"
PARTIAL = "partial"
CANCELLED = "cancelled"
FILLED = "filled"

_prsm = "PRSM"


class ClearingCounter(Counter):
    """
    Counter than deletes keys with value zero.
    Useful to tracking the volume profile of an order book.
    """

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self[key] == 0.0:
            del self[key]


@attr.s(auto_attribs=True, frozen=True)
class ClientRecord:
    """
    Stores client name information together with:
    position: total holdings of asset in question
    cost_basis: book value of position
    realized pl: sum of cash holdings and cost basis
    """

    alias: str
    position: int
    cost_basis: float
    realized_pl: float
    outstanding_bid_volume: int
    outstanding_ask_volume: int
    first: str = ""
    last: str = ""


def _order_queue() -> SimplePriorityQueue[Order]:
    return SimplePriorityQueue(id_func=lambda x: x.id)


@attr.s(auto_attribs=True)
class LocalServerInterface(ServerInterface):
    """
    An in-memory, single-process matching engine
    """

    _live_bids: SimplePriorityQueue[Order] = attr.ib(factory=_order_queue)
    _live_asks: SimplePriorityQueue[Order] = attr.ib(factory=_order_queue)
    _order_store: OrderStore = attr.ib(factory=OrderStore)
    _bid_profile: Dict[float, int] = attr.ib(factory=ClearingCounter)
    _ask_profile: Dict[float, int] = attr.ib(factory=ClearingCounter)
    _clients: Dict[UUID, ClientRecord] = attr.ib(factory=dict)

    def get_market_data(self, symbol: str) -> MarketData:
        return MarketData(
            symbol=_prsm,
            asks=[
                AggregateOrders(price=price, volume=volume)
                for price, volume in self._ask_profile.items()
            ],
            bids=[
                AggregateOrders(price=price, volume=volume)
                for price, volume in self._bid_profile.items()
            ],
            last_trade_price=self._order_store.last_price,
            last_trade_volume=self._order_store.last_volume,
            timestamp=datetime.now(),
        )

    def cancel_order(self, order_id: UUID):
        assert order_id in self._order_store.order_ids(), "Unknown order"
        order = self._order_store.get_by_order_id(order_id)
        assert order.status in {OPEN, PARTIAL}, (
            f"can only cancel open or partially open orders.  order "
            f"{order_id} has status {order.status}"
        )
        new_order = attr.evolve(order, status=CANCELLED)
        self._order_store.update(new_order)
        client = self._clients[order.client]
        if order.volume > 0:
            self._live_bids.remove(order)
            self._bid_profile[order.price] -= order.volume
            self._clients[order.client] = attr.evolve(
                client, outstanding_bid_volume=client.outstanding_bid_volume - order.volume
            )
        elif order.volume < 0:

            self._live_asks.remove(order)
            self._ask_profile[order.price] += order.volume
            self._clients[order.client] = attr.evolve(
                client, outstanding_ask_volume=client.outstanding_ask_volume - abs(order.volume)
            )

    def get_client_orders(self, client_id: UUID) -> ClientOrders:
        return ClientOrders(
            client_id=client_id,
            timestamp=datetime.now(),
            orders=list(self._order_store.get_by_client_id(client_id=client_id)),
        )

    def clear(self):
        self._live_bids.clear()
        self._live_asks.clear()
        self._order_store.clear()
        self._bid_profile = ClearingCounter()
        self._ask_profile = ClearingCounter()
        self._clients = {}

    def register(self, alias: str) -> UUID:
        client_id = uuid.uuid4()
        self._clients[client_id] = ClientRecord(
            alias=alias,
            position=0,
            cost_basis=0,
            realized_pl=0,
            outstanding_bid_volume=0,
            outstanding_ask_volume=0,
        )
        self._order_store.register_client(client_id)
        return client_id

    def submit_order(self, client_id: UUID, order: OrderRequest) -> None:
        if order.type == OrderType.LIMIT:
            price = order.limit_price
            assert price is not None
        elif order.quantity > 0:
            price = float("inf")
        else:
            price = 0
        sender = Order(
            id=uuid.uuid4(),
            open=datetime.now(),
            client=client_id,
            symbol=Ticker(order.symbol),
            status=OPEN,
            price=price,
            volume=order.quantity,
        )
        client = self._clients[client_id]
        if order.quantity > 0:
            same_queue = self._live_bids
            same_profile = self._bid_profile
            opposite_queue = self._live_asks
            opposite_profile = self._ask_profile
            self._clients[client_id] = attr.evolve(
                self._clients[client_id],
                outstanding_bid_volume=client.outstanding_bid_volume + order.quantity,
            )
        else:
            same_queue = self._live_asks
            same_profile = self._ask_profile
            opposite_queue = self._live_bids
            opposite_profile = self._bid_profile
            self._clients[client_id] = attr.evolve(
                self._clients[client_id],
                outstanding_ask_volume=client.outstanding_ask_volume + abs(order.quantity),
            )

        sender = _match_with_queue(
            sender,
            opposite_queue=opposite_queue,
            order_store=self._order_store,
            opposite_profile=opposite_profile,
            clients=self._clients,
        )
        self._order_store.add(sender)
        if sender.status in {OPEN, PARTIAL}:
            _queue_push(sender, same_queue=same_queue)
            same_profile[sender.price] += abs(sender.volume)

    def get_client_data_table(self) -> ClientDataTable:
        timestamp = datetime.now()
        return ClientDataTable(
            data=[
                attr.evolve(self.get_client_data(client_id=client_id), timestamp=timestamp)
                for client_id in self._clients
            ]
        )

    def get_client_data(self, client_id: UUID) -> ClientData:
        record = self._clients[client_id]
        return ClientData(
            alias=record.alias,
            position=record.position,
            costbasis=Decimal(record.cost_basis),
            realizedpl=Decimal(record.realized_pl),
            timestamp=datetime.now(),
            wealth=Decimal(record.realized_pl - record.cost_basis),
            outstanding_bid_volume=record.outstanding_bid_volume,
            outstanding_ask_volume=record.outstanding_ask_volume,
        )

    def spawn_episode(self) -> ServerInterface:
        episode = deepcopy(self)
        episode.clear()
        return episode


def _matches(sender: Order, receiver: Order) -> bool:
    if sender.volume > 0:
        assert receiver.volume < 0
        return True if sender.price >= receiver.price else False
    assert receiver.volume > 0
    return True if sender.price <= receiver.price else False


def _execute(order: Order, volume: int) -> Order:
    assert volume > 0, "Cannot execute zero volume"
    sign = int(order.volume / abs(order.volume))
    new_volume = order.volume - sign * volume
    new_status = PARTIAL if abs(new_volume) > 0 else FILLED
    return Order(
        id=order.id,
        open=order.open,
        client=order.client,
        symbol=order.symbol,
        status=new_status,
        price=order.price,
        volume=new_volume,
    )


def _match_with_order(sender: Order, receiver: Order) -> Tuple[Order, Order, int]:
    trade_volume = min(abs(sender.volume), abs(receiver.volume))
    return _execute(sender, trade_volume), _execute(receiver, trade_volume), trade_volume


def _match_with_queue(
    sender: Order,
    opposite_queue: SimplePriorityQueue,
    opposite_profile: Dict[float, int],
    order_store: OrderStore,
    clients: Dict[UUID, ClientRecord],
) -> Order:
    sender_sign = _sign(sender.volume)
    while True:
        receiver = opposite_queue.peek()
        if receiver is None:
            break
        if not _matches(sender, receiver):
            break
        sender, receiver, trade_volume = _match_with_order(sender, receiver)

        _apply_trade_to_clients(
            clients, sender.client, signed_volume=sender_sign * trade_volume, price=receiver.price
        )
        _apply_trade_to_clients(
            clients,
            receiver.client,
            signed_volume=-1 * sender_sign * trade_volume,
            price=receiver.price,
        )
        order_store.update(receiver)
        opposite_profile[receiver.price] -= trade_volume
        if receiver.status == FILLED:
            opposite_queue.pop()
        else:
            opposite_queue.update_first(receiver)
        order_store.last_volume = trade_volume
        order_store.last_price = receiver.price
        if sender.status == FILLED:
            break
    return sender


def _apply_trade_to_clients(
    clients: Dict[UUID, ClientRecord], client_id: UUID, signed_volume: int, price: float
):
    record = clients[client_id]
    gross = -signed_volume * price
    if record.position == 0 or (_sign(signed_volume) == _sign(record.position)):
        book = -gross
    elif abs(signed_volume) <= (abs(record.position)):
        book = record.cost_basis / record.position * signed_volume
    else:
        _apply_trade_to_clients(clients, client_id, -record.position, price=price)
        _apply_trade_to_clients(clients, client_id, signed_volume + record.position, price=price)
        return
    clients[client_id] = attr.evolve(
        record,
        position=record.position + signed_volume,
        cost_basis=record.cost_basis + book,
        realized_pl=record.realized_pl + gross + book,
        outstanding_bid_volume=clients[client_id].outstanding_bid_volume - max(signed_volume, 0),
        outstanding_ask_volume=clients[client_id].outstanding_ask_volume
        - abs(min(signed_volume, 0)),
    )


def _sign(x: int) -> int:
    return int(copysign(1, x))


def _queue_push(sender: Order, same_queue: SimplePriorityQueue):
    if sender.volume > 0:
        same_queue.push(sender, priority=-sender.price)
    else:
        same_queue.push(sender, priority=sender.price)
