from typing import AbstractSet, Dict, Set
from uuid import UUID

import attr

from prism_agents.api.agent_schemas import Order

_default_last_trade_price = 0.0
_default_last_trade_volume = 0


@attr.s(auto_attribs=True)
class OrderStore:
    """
    Maintains two dictionaries organizing orders by client id and by
    order id. Keeps them consistent.
    """

    _by_order_id: Dict[UUID, Order] = attr.ib(factory=dict)
    _by_client_id: Dict[UUID, Set[Order]] = attr.ib(factory=dict)
    last_price: float = attr.ib(default=_default_last_trade_price)
    last_volume: int = attr.ib(default=_default_last_trade_volume)

    def add(self, order: Order):
        assert order.id not in self._by_order_id
        self._by_order_id[order.id] = order
        assert order.client in self._by_client_id, f"Unknown client {order.client}"
        self._by_client_id[order.client].add(order)

    def get_by_client_id(self, client_id: UUID) -> AbstractSet[Order]:
        return self._by_client_id[client_id]

    def get_by_order_id(self, order_id: UUID) -> Order:
        return self._by_order_id[order_id]

    def update(self, order: Order):
        assert order.id in self._by_order_id
        old_order = self._by_order_id[order.id]
        self._by_order_id[order.id] = order
        self._by_client_id[order.client].remove(old_order)
        self._by_client_id[order.client].add(order)

    def register_client(self, uuid: UUID):
        self._by_client_id[uuid] = set()

    def order_ids(self) -> AbstractSet[UUID]:
        return self._by_order_id.keys()

    def client_ids(self) -> AbstractSet[UUID]:
        return self._by_client_id.keys()

    def clear(self):
        self._by_order_id = {}
        self._by_client_id = {}
        self.last_price = _default_last_trade_price
        self.last_volume = _default_last_trade_volume
