from heapq import heappop, heappush
from typing import Callable, Dict, Generic, List, Optional, TypeVar, Union
from uuid import UUID

import attr


class _RemovedType:
    pass


_removed = _RemovedType()


T = TypeVar("T")


@attr.s(auto_attribs=True)
class _Record(Generic[T]):
    priority: float
    contents: Union[T, _RemovedType] = attr.ib(order=False)


@attr.s(auto_attribs=True)
class SimplePriorityQueue(Generic[T]):
    """
    A simple heap-based priority queue that maintain a _record_finder dictionary
    to allow fast cancellation of orders.
    Based on example from: https://docs.python.org/3/library/heapq.html

    """

    id_func: Callable[[T], UUID]

    _heap: List[_Record] = attr.ib(factory=list)
    _record_finder: Dict[UUID, _Record] = attr.ib(factory=dict)

    def _clear_removed(self):
        while True:
            if len(self._heap) == 0:
                return
            if self._heap[0].contents != _removed:
                return
            heappop(self._heap)

    def push(self, order: T, priority: float):
        entry = _Record(priority=priority, contents=order)
        heappush(self._heap, entry)
        self._record_finder[self.id_func(order)] = entry

    def pop(self) -> Optional[T]:
        self._clear_removed()
        if len(self._heap) == 0:
            return None
        record = heappop(self._heap)
        assert not isinstance(record.contents, _RemovedType)
        del self._record_finder[self.id_func(record.contents)]
        return record.contents

    def remove(self, order: T):
        record = self._record_finder.pop(self.id_func(order))
        record.contents = _removed

    def update_first(self, new_order: T):
        self._heap[0].contents = new_order

    def peek(self) -> Optional[T]:
        self._clear_removed()
        if len(self._heap) == 0:
            return None
        item = self._heap[0].contents
        assert not isinstance(item, _RemovedType)
        return item

    def clear(self):
        self._heap = []
        self._record_finder = {}

    def __len__(self):
        return len(self._record_finder)
