import datetime
import logging
from decimal import Decimal
from json import JSONDecodeError
from typing import List, Dict, Optional
from uuid import UUID

import attr
import cattr
import requests
from requests import Response

from prism_agents.api.agent_schemas import (
    ClientData,
    ClientDataTable,
    ClientOrders,
    MarketData,
    OrderRequest,
)
from prism_agents.api.server_interface import ServerInterface
from prism_agents.configuration.prism_agents_logging import prism_agents_configure_logging

logger = logging.getLogger(__name__)
logger = prism_agents_configure_logging(logger, level=logging.INFO)

converter = cattr.Converter()
converter.register_structure_hook(
    datetime.datetime, lambda x, _: datetime.datetime.fromisoformat(x)
)
converter.register_structure_hook(UUID, lambda x, _: UUID(x))
converter.register_structure_hook(Decimal, lambda x, _: Decimal(x))

"""
This module contains functions wrapping requests to then prism api.
"""


def _strip_backslash(url: str) -> str:
    return url.rstrip("/")


@attr.s(frozen=True, auto_attribs=True)
class RemoteServerInterface(ServerInterface):
    """
    An interface to a remote server accessed through rest calls
    Enabling robust=True will prevent crashes when non-200 status codes are returned.
    Note that crashes are still possible if the returned data cannot be
    parsed using the appropriate cattrs converter.
    """

    server_url: str = attr.ib(converter=_strip_backslash)
    access_token: str = attr.ib()
    robust: bool = True
    skip_constraints_key_override: Optional[str] = None    

    def __attrs_post_init__(self):
        header = {
            "Authorization": f"Bearer {self.access_token}",
            "accept": "application/json",
            "Content-Type": "application/json",
        }
        object.__setattr__(self,'HEADER', header)

    def get_market_data(self, symbol: str) -> MarketData:
        logger.debug(f"getting market data for symbol {symbol}")
        response: requests.Response = requests.get(f"{self.server_url}/api/v1/market/{symbol}/data", headers=self.HEADER)
        data = converter.structure(response.json(), MarketData)
        self._check_status(response)
        return data

    def cancel_order(self, order_id: UUID):
        logger.debug(f"canceling order {order_id}")
        response = requests.delete(f"{self.server_url}/api/v1/market/{order_id}", headers=self.HEADER)
        if response.status_code != 200:
            if not self.robust:
                # don't throw an error if it was a valid order but another agent
                # matched with it before we could cancel
                assert (
                    response.status_code == 422
                    and "Cannot cancel an order that is not open" in response.json()["detail"]
                ), response.json()
            else:
                logger.warning(
                    f"Received status code {response.status_code} in response {response.json()} "
                )

    def get_client_orders(self, client_id: UUID) -> ClientOrders:
        logger.debug(f"Getting client orders for client {client_id}")
        response: requests.Response = requests.get(
            f"{self.server_url}/api/v1/accounts/{client_id}/orders", headers=self.HEADER
        )
        self._check_status(response)
        return converter.structure(response.json(), ClientOrders)

    def clear(self):
        logger.debug("clearing market")
        response = requests.delete(f"{self.server_url}/api/v1/market")
        assert response.status_code == 200, response.json()

    def register(self, alias: str) -> UUID:        
        logger.debug(f"Registering client {alias}")
        response = requests.post(f"{self.server_url}/api/v1/accounts", json={"alias": alias}, 
            headers=self.HEADER)
        self._check_status(response)
        logger.info(response)  # just to have the UUID available in the console
        return response.json()

    def submit_order(self, client_id: UUID, order: OrderRequest) -> None:
        logger.debug(
            f"client {client_id} is submitting a {order.type} for"
            f" symbol {order.symbol} with volume {order.quantity}"
        )
        if self.skip_constraints_key_override is not None:
            order = attr.evolve(order, skip_constraints_key=self.skip_constraints_key_override)
        response = requests.post(
            f"{self.server_url}/api/v1/accounts/{client_id}/orders",
            json=converter.unstructure(order), headers=self.HEADER
        )
        self._check_status(response)
        logger.debug(response.json())

    def get_client_data_table(self) -> ClientDataTable:
        response = requests.get(url=f"{self.server_url}/api/v1/accounts/data", headers=self.HEADER)
        self._check_status(response)
        return converter.structure(response.json(), ClientDataTable)

    def get_client_data(self, client_id: UUID) -> ClientData:
        response = requests.get(url=f"{self.server_url}/api/v1/accounts/{client_id}/data", headers=self.HEADER)
        self._check_status(response)
        return converter.structure(response.json(), ClientData)

    def spawn_episode(self) -> ServerInterface:
        raise NotImplementedError

    def get_market_data_history(self, symbol: str) -> List[MarketData]:
        response = requests.get(url=f"{self.server_url}/api/v1/market/{symbol}/data_history", headers=self.HEADER)
        self._check_status(response)
        return converter.structure(response.json(), List[MarketData])

    def get_clients_data_history(self) -> List[ClientDataTable]:
        response = requests.get(url=f"{self.server_url}/api/v1/accounts/data_history", headers=self.HEADER)
        self._check_status(response)
        return converter.structure(response.json(), List[ClientDataTable])

    def _check_status(self, response: Response):
        if response.status_code != 200:
            try:
                msg = f"Received status code {response.status_code} in response {response.json()}"
            except JSONDecodeError:
                msg = (
                    f"Received status code {response.status_code}. Could not decode JSON response."
                )
            if self.robust:
                logger.warning(msg)
            else:
                raise AssertionError(msg)
