import logging
from decimal import ROUND_HALF_DOWN, Decimal
from typing import Dict, List, Optional
from uuid import UUID

from attr import asdict

from prism_agents.api.agent_schemas import ClientData, MarketData
from prism_agents.api.server_interface import ServerInterface
from prism_agents.configuration.prism_agents_logging import prism_agents_configure_logging

logger = logging.getLogger(__name__)
logger = prism_agents_configure_logging(logger)


def get_best_bid(data: MarketData, tick_size: Decimal) -> Optional[Decimal]:
    """
    Given MarketData describing the current state of the order book, get the best bid price
    """
    if len(data.bids) == 0:
        return None
    return Decimal(max([item.price for item in data.bids])).quantize(tick_size, ROUND_HALF_DOWN)


def get_best_ask(data: MarketData, tick_size: Decimal) -> Optional[Decimal]:
    """
    Given MarketData describing the current state of the order book, get the best ask price
    """
    if len(data.asks) == 0:
        return None
    return Decimal(min([item.price for item in data.asks])).quantize(tick_size, ROUND_HALF_DOWN)


def get_mid_price(data: MarketData, tick_size: Decimal = Decimal("0.01"), strict=False) -> Decimal:
    """
    Given MarketData received from prism, get the mid price
    If the data has no best bid/ask, fall back on a default
    Note that the increment of the mid price can potentially be half of a tick
    """
    best_bid = get_best_bid(data, tick_size=tick_size)
    best_ask = get_best_ask(data, tick_size=tick_size)
    if best_bid is None or best_ask is None:
        if best_bid is not None:
            return best_bid
        if best_ask is not None:
            return best_ask
        return Decimal(data.last_trade_price).quantize(tick_size)
    if best_ask < best_bid:
        logger.warning(
            f"Invalid market state when computing . " f"Ask {best_ask} is less than bid {best_bid}"
        )
    if strict:
        assert (
            best_ask >= best_bid
        ), f"Invalid market state.  Ask {best_ask} is less than bid {best_bid}"
    return (best_bid + best_ask) / Decimal("2")


def get_spread(data: MarketData, tick_size: Decimal) -> Decimal:
    best_bid = get_best_bid(data, tick_size=tick_size)
    best_ask = get_best_ask(data, tick_size=tick_size)
    if best_bid is None or best_ask is None:
        return tick_size
    return best_ask - best_bid


def cancel_all_client_orders(client_id: UUID, server: ServerInterface):
    orders = server.get_client_orders(client_id=client_id)
    for order in orders.orders:
        if order.status in {"open", "partial"}:
            server.cancel_order(order_id=order.id)


def parse_decimal_from_client_data(
    client_data: ClientData,
    keys_to_include: Optional[List[str]] = None,
    keys_to_exclude: Optional[List[str]] = None,
) -> Dict:
    """
    Helper function that turns the ClientData into a dictionary and parse Decimal values to floats
    """

    keys_to_exclude_set = set([]) if keys_to_exclude is None else set(keys_to_exclude)
    client_data_dict = asdict(client_data)
    client_data_dict_returned = {}

    if keys_to_include is None:
        keys_to_include = list(client_data_dict.keys())

    for key in keys_to_include:
        if key in keys_to_exclude_set:
            continue

        client_data_dict_returned[key] = client_data_dict[key]
        if isinstance(client_data_dict[key], Decimal):
            client_data_dict_returned[key] = float(client_data_dict[key])

    return client_data_dict_returned
