from __future__ import annotations

from abc import ABC, abstractmethod
from decimal import Decimal
from uuid import UUID

from prism_agents.api.agent_schemas import (
    ClientData,
    ClientDataTable,
    ClientOrders,
    MarketData,
    OrderExpiration,
    OrderRequest,
    OrderType,
)


class ServerInterface(ABC):
    """
    Encapsulates a mode of communication with with a prism server
    """

    @abstractmethod
    def get_market_data(self, symbol: str) -> MarketData:
        pass

    @abstractmethod
    def cancel_order(self, order_id: UUID):
        pass

    @abstractmethod
    def get_client_orders(self, client_id: UUID) -> ClientOrders:
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def register(self, alias: str) -> UUID:
        pass

    @abstractmethod
    def submit_order(self, client_id: UUID, order: OrderRequest) -> None:
        pass

    @abstractmethod
    def get_client_data_table(self) -> ClientDataTable:
        pass

    @abstractmethod
    def get_client_data(self, client_id: UUID) -> ClientData:
        pass

    @abstractmethod
    def spawn_episode(self) -> ServerInterface:
        pass

    def market_order(self, symbol: str, client_id: UUID, volume: int) -> None:
        """
        Convention: positive-> buy, negative -> sell
        """
        order_request = OrderRequest(
            symbol=symbol,
            type=OrderType.MARKET,
            quantity=volume,
            limit_price=-1,
            expiration=OrderExpiration.DAY,
        )
        return self.submit_order(client_id=client_id, order=order_request)

    def limit_order(self, symbol: str, client_id: UUID, volume: int, price: Decimal) -> None:
        """
        Convention: positive-> buy, negative -> sell
        """
        order_request = OrderRequest(
            symbol=symbol,
            type=OrderType.LIMIT,
            quantity=volume,
            limit_price=float(price),
            expiration=OrderExpiration.DAY,
        )
        return self.submit_order(client_id=client_id, order=order_request)
