from decimal import ROUND_HALF_DOWN, Decimal
from statistics import mean
from typing import List, Optional, Sequence, Set, Tuple

import numpy as np
import xarray as xr
from statsmodels.tsa.stattools import acf

from prism_agents.api.agent_schemas import AggregateOrders, ClientData, ClientDataTable, MarketData
from prism_agents.api.schema_utility_functions import (
    get_best_ask,
    get_best_bid,
    get_mid_price,
    parse_decimal_from_client_data,
)
from prism_agents.runner.experiment_result import ExperimentResult, TimedObservation

VOLUME = "volume"
ALIAS = "alias"
POSITION = "position"
COSTBASIS = "costbasis"
REALIZEDPL = "realizedpl"
WEALTH = "wealth"
PNL_RANKING = "pnl_ranking"
BID_PROFILE = "bid_profile_volume"
ASK_PROFILE = "ask_profile_volume"
PRICE_DIMENSION = "price"
TIMESTAMP = "timestamp"
TIMESTEP = "timestep"
SIMULATION_TIME = "simulation_time (s)"
OUTSTANDING_BID_VOLUME = "outstanding_bid_volume"
OUTSTANDING_ASK_VOLUME = "outstanding_ask_volume"


BID_PREFIX = "bid_"
ASK_PREFIX = "ask_"
BID_PRICE_DIMENSION = BID_PREFIX + PRICE_DIMENSION
ASK_PRICE_DIMENSION = ASK_PREFIX + PRICE_DIMENSION
BID_VOLUME = BID_PREFIX + VOLUME
ASK_VOLUME = ASK_PREFIX + VOLUME

ASK_DEPTH_DIMENSION = "ask_depth"
BID_DEPTH_DIMENSION = "bid_depth"

BEST_BID = "best_bid"
BEST_ASK = "best_ask"
MID = "mid"
MID_INCREMENT = "mid_increment"

LAG_DIMENSION = "lag"
ACF = "acf"
ACF_LOWER = "acf_lower"
ACF_UPPER = "acf_upper"

ORDER_BOOK_SNAPSHOT = "order_book_snapshot"
SNAPSHOT_PRICE_DIM = "snapshot_price"

BATCHED_RUN_DIM = "batched_run"


def client_data_to_xr(client_data: ClientData) -> xr.Dataset:
    dict_client_data = parse_decimal_from_client_data(client_data)
    return xr.Dataset(dict_client_data).assign_coords({ALIAS: client_data.alias})


def client_data_table_to_xr(client_data_table: ClientDataTable) -> xr.Dataset:
    datasets = [client_data_to_xr(item) for item in client_data_table.data]
    if len(datasets) == 0:
        empty = xr.DataArray([], dims=[ALIAS])
        return xr.Dataset(
            {
                POSITION: empty,
                COSTBASIS: empty,
                REALIZEDPL: empty,
                OUTSTANDING_BID_VOLUME: empty,
                OUTSTANDING_ASK_VOLUME: empty,
            }
        ).assign_coords({ALIAS: []})

    return xr.concat(datasets, dim=ALIAS)


def aggregate_orders_to_xr(aggregate_orders: AggregateOrders) -> xr.Dataset:
    return xr.Dataset({VOLUME: aggregate_orders.volume}).assign_coords(
        {PRICE_DIMENSION: aggregate_orders.price}
    )


def _empty_orders_xr() -> xr.Dataset:
    return xr.Dataset(
        {
            VOLUME: xr.DataArray([], dims=(PRICE_DIMENSION,)),
        }
    )


def _combined_aggregate_orders_xr(orders: List[AggregateOrders]) -> xr.Dataset:
    return (
        xr.concat([aggregate_orders_to_xr(item) for item in orders], dim=PRICE_DIMENSION)
        if len(orders) > 0
        else _empty_orders_xr()
    )


def market_data_to_xr(market_data: MarketData) -> xr.Dataset:
    bid_dataset = _combined_aggregate_orders_xr(market_data.bids).rename(
        {PRICE_DIMENSION: BID_PRICE_DIMENSION, VOLUME: BID_VOLUME}
    )
    ask_dataset = _combined_aggregate_orders_xr(market_data.asks).rename(
        {PRICE_DIMENSION: ASK_PRICE_DIMENSION, VOLUME: ASK_VOLUME}
    )
    return bid_dataset.merge(ask_dataset).assign_coords({TIMESTAMP: market_data.timestamp})


def market_data_summary(
    market_data: MarketData, tick_size: Decimal = Decimal("0.01")
) -> xr.Dataset:
    """
    Dataset describing state of order book at one time
    """
    best_bid = get_best_bid(market_data, tick_size=tick_size)
    best_ask = get_best_ask(market_data, tick_size=tick_size)
    mid = get_mid_price(market_data, tick_size=tick_size)
    return xr.Dataset(
        {
            BEST_BID: float(best_bid) if best_bid is not None else float("nan"),
            BEST_ASK: float(best_ask) if best_ask is not None else float("nan"),
            MID: float(mid),
            TIMESTAMP: market_data.timestamp,
        }
    )


def market_history_summary_dataset(
    market_history: Sequence[TimedObservation],
    tick_size: Decimal = Decimal("0.01"),
    ticker: str = "PRSM",
) -> xr.Dataset:
    """
    Dataset summarizing the state of the order book over time
    """
    assert len(market_history) > 0
    ds_combined = xr.concat(
        [
            market_data_summary(item.observation.market[ticker], tick_size=tick_size)
            for item in market_history
        ],
        dim=TIMESTEP,
    )
    ds_combined[SIMULATION_TIME] = xr.DataArray(
        [item.time for item in market_history], dims=(TIMESTEP,)
    )
    ds_combined = _summary_dataset_with_mid_increments(ds_combined)
    ds_combined = _summary_dataset_with_mid_increment_autocorr(ds_combined)
    return ds_combined


def _summary_dataset_with_mid_increments(ds: xr.Dataset) -> xr.Dataset:
    ds = ds.copy()
    ds[MID_INCREMENT] = xr.DataArray(
        ds[MID].values - ds[MID].shift({TIMESTEP: 1}).values, dims=ds[MID].dims
    )
    return ds


def _summary_dataset_with_mid_increment_autocorr(ds: xr.Dataset) -> xr.Dataset:
    ds = ds.copy()
    if len(ds[MID_INCREMENT].dropna(TIMESTEP)) <= 1:
        return ds

    autocorr, conf = acf(ds[MID_INCREMENT].values, alpha=0.05, missing="drop")
    ds[ACF] = xr.DataArray(autocorr, dims=(LAG_DIMENSION,))
    ds[ACF_LOWER] = xr.DataArray(conf[:, 0], dims=(LAG_DIMENSION,))
    ds[ACF_UPPER] = xr.DataArray(conf[:, 1], dims=(LAG_DIMENSION,))
    return ds


def _filter_ds_by_keys(ds: xr.Dataset, keys: Set[str]):
    return xr.Dataset({key: value for key, value in ds.items() if key in keys})


def client_data_dataset(obs: Sequence[TimedObservation]) -> xr.Dataset:
    """
    Dataset showing client attributes like position and PnL over time
    """
    client_data_list = [
        _filter_ds_by_keys(
            client_data_table_to_xr(observation.observation.client_data),
            {POSITION, COSTBASIS, REALIZEDPL, OUTSTANDING_BID_VOLUME, OUTSTANDING_ASK_VOLUME},
        )
        for observation in obs
    ]
    ds_combined = xr.concat(
        client_data_list,
        dim=TIMESTEP,
    )
    ds_combined[SIMULATION_TIME] = xr.DataArray([item.time for item in obs], dims=(TIMESTEP,))
    ds_combined = client_dataset_with_wealth(ds_combined)
    ds_combined = client_dataset_with_realized_pnl_ranking(ds_combined)
    return ds_combined


def multi_run_client_data_dataset(results: Sequence[ExperimentResult]) -> xr.Dataset:
    return xr.concat(
        [client_data_dataset(result.observations) for result in results], dim=BATCHED_RUN_DIM
    )


def volume_profile_dataset(
    market_data: MarketData, max_depth: int = 25, tick_size: Decimal = Decimal("0.01")
) -> xr.Dataset:
    """
    Dataset showing the volume profile of the order book
    """
    bid_array = xr.DataArray([0.0] * max_depth, dims=(BID_DEPTH_DIMENSION,))
    ask_array = xr.DataArray([0.0] * max_depth, dims=(ASK_DEPTH_DIMENSION,))
    mid = get_mid_price(market_data, tick_size=tick_size)
    for bid in market_data.bids:
        steps_from_mid = int(
            abs((Decimal(bid.price) - mid) / tick_size).quantize(
                tick_size, rounding=ROUND_HALF_DOWN
            )
        )
        if steps_from_mid < max_depth:
            bid_array[steps_from_mid] += bid.volume

    for ask in market_data.asks:
        steps_from_mid = int(
            abs((Decimal(ask.price) - mid) / tick_size).quantize(
                tick_size, rounding=ROUND_HALF_DOWN
            )
        )
        if steps_from_mid < max_depth:
            ask_array[steps_from_mid] += ask.volume

    return xr.Dataset({BID_PROFILE: bid_array, ASK_PROFILE: ask_array})


def volume_profile_over_time_dataset(
    observations: Sequence[TimedObservation], max_depth: int = 25, ticker: str = "PRSM"
) -> xr.Dataset:
    """
    Dataset showing the volume profile of the order book over time
    """
    ds_combined = xr.concat(
        [
            volume_profile_dataset(market_data=obs.observation.market[ticker], max_depth=max_depth)
            for obs in observations
        ],
        dim=TIMESTEP,
    )

    ds_combined[SIMULATION_TIME] = xr.DataArray(
        [item.time for item in observations], dims=(TIMESTEP,)
    )
    return ds_combined


def client_dataset_with_wealth(ds: xr.Dataset) -> xr.Dataset:
    ds = ds.copy()
    ds[WEALTH] = ds[REALIZEDPL] - ds[COSTBASIS]
    return ds


def client_dataset_with_realized_pnl_ranking(ds: xr.Dataset) -> xr.Dataset:
    ds = ds.copy()
    pnl_array = ds[REALIZEDPL]
    ds[PNL_RANKING] = pnl_array.sizes[ALIAS] - pnl_array.argsort(axis=pnl_array.dims.index(ALIAS))
    return ds


def make_order_book_snapshot(
    market_data: MarketData,
    min_price: Decimal,
    max_price: Decimal,
    tick_size: Decimal = Decimal("0.01"),
) -> xr.DataArray:
    """
    Return a 1d DataArray containing the
    volume of orders at each tick in a range
    Positive corresponds to bids, negative corresponds to asks
    """
    n_bins = int((max_price - min_price) / tick_size) + 1
    if n_bins > 30000:
        raise RuntimeError(f"Got {n_bins} bins, which is greater than 30000, the maximum allowed")
    prices = [min_price + tick_size * i for i in range(n_bins)]
    array: xr.DataArray = xr.DataArray(np.zeros(n_bins), dims=(SNAPSHOT_PRICE_DIM,)).assign_coords(
        {SNAPSHOT_PRICE_DIM: prices}
    )
    price_to_index = {price: i for i, price in enumerate(prices)}
    for bid in market_data.bids:
        price = Decimal(bid.price).quantize(tick_size)
        if price in price_to_index:
            array[price_to_index[price]] += bid.volume
    for ask in market_data.asks:
        price = Decimal(ask.price).quantize(tick_size)
        if price in price_to_index:
            array[price_to_index[price]] -= ask.volume
    return array


def make_order_book_snapshot_over_time(
    observations: Sequence[TimedObservation],
    ticker: str = "PRSM",
    bounds: Optional[Tuple[Decimal, Decimal]] = None,
    tick_size: Decimal = Decimal("0.01"),
) -> xr.Dataset:
    """
    return a Dataset containing a 2d array
    whose slices show the volume of orders
    at each tick in a range at a given timestep
    """
    if bounds is None:
        mean_mid = Decimal(
            mean([float(get_mid_price(obs.observation.market[ticker])) for obs in observations])
        ).quantize(tick_size)
        bounds = mean_mid - tick_size * 50, mean_mid + tick_size * 50

    ds = xr.Dataset(
        {
            ORDER_BOOK_SNAPSHOT: xr.concat(
                [
                    make_order_book_snapshot(
                        obs.observation.market[ticker],
                        min_price=bounds[0],
                        max_price=bounds[1],
                        tick_size=tick_size,
                    )
                    for obs in observations
                ],
                dim=TIMESTEP,
            )
        }
    )
    ds[SIMULATION_TIME] = xr.DataArray([item.time for item in observations], dims=(TIMESTEP,))
    return ds
