from typing import Any, Callable, Dict, List

import attr
import numpy as np
import torch
from torch.utils.data import IterableDataset


@attr.s(auto_attribs=True)
class InfoRange:
    options: List
    mode: str = attr.field()

    @mode.validator
    def check_mode(self, attribute, value):
        assert value in ["uniform", "choice"]

    def validate_options_length_if_uniform_sampling(self):
        if self.mode == "uniform" and len(self.options) != 2:
            raise ValueError(f"Length of options should be 2, found {len(self.options)}!")

    def __attrs_post_init__(self) -> None:
        self.validate_options_length_if_uniform_sampling()


def draw_from_input_range(input_range: Dict[str, InfoRange], batch_size: int) -> Dict[str, Any]:
    input_drawn = {}
    for key, info in input_range.items():
        if info.mode == "uniform":
            input_drawn[key] = np.random.uniform(info.options[0], info.options[1], size=batch_size)
        else:
            input_drawn[key] = np.random.choice(info.options, size=batch_size)
    return input_drawn


@attr.s(auto_attribs=True)
class FundamentalProcessDataset(IterableDataset):
    """
    This class allows us to create datasets of Torch Tensors based on a stochastic process
    generator.

    The input range is a dictionary keys values represent the range which can be taken by the inputs
    of the stochastic process generator. This allows the generation of paths coming from processes
    having different characteristics and hence perform domain randomization.
    """

    stochastic_process_generator: Callable[..., np.ndarray]
    input_range: Dict[str, InfoRange]
    batch_size: int
    seq_length: int
    target_horizon: int
    cuda: bool = False
    batches_per_epoch: int = 100

    def __iter__(self):
        batch_counter = 0
        while batch_counter < self.batches_per_epoch:
            batch_counter += 1
            drawn_inputs = draw_from_input_range(self.input_range, self.batch_size)
            input_dict = {"seq_length": self.seq_length + self.target_horizon, **drawn_inputs}
            paths = self.stochastic_process_generator(**input_dict)
            x = torch.Tensor(paths[:, : self.seq_length]).unsqueeze(-1)
            target_loc = self.seq_length + self.target_horizon
            y = torch.Tensor(paths[:, target_loc - 1 : target_loc])
            if self.cuda:
                x.cuda()
                y.cuda()
            yield {"input": x, "target": y}
