from decimal import Decimal
from typing import List, Tuple

import attr
import pandas as pd
import torch
from attr import asdict

from prism_agents.api.remote_request_funcs import ClientDataTable
from prism_agents.api.schema_utility_functions import (
    MarketData,
    get_best_ask,
    get_best_bid,
    get_mid_price,
    parse_decimal_from_client_data,
)
from prism_agents.dataset.base import MarketDataset, TimestepInfo


def client_data_table_to_df(client_data_table: ClientDataTable) -> pd.DataFrame:
    records = [
        parse_decimal_from_client_data(client_data) for client_data in client_data_table.data
    ]
    df_client_data = pd.DataFrame.from_records(records)
    return df_client_data


def client_data_history_to_df(client_data_history: List[ClientDataTable]) -> pd.DataFrame:
    history_frames = []
    for ii, table in enumerate(client_data_history):
        df = client_data_table_to_df(table)
        df["snapshot"] = ii
        history_frames.append(df)
    history_df = pd.concat(history_frames)
    return history_df


def market_history_to_df(market_history: List[MarketData], tick_size: Decimal = Decimal("0.01")):
    history_frames = pd.DataFrame.from_records([asdict(data) for data in market_history])
    history_frames["best_bid"] = [
        get_best_bid(data, tick_size=tick_size) for data in market_history
    ]
    history_frames["best_ask"] = [
        get_best_ask(data, tick_size=tick_size) for data in market_history
    ]
    history_frames["mid"] = [get_mid_price(data, tick_size=tick_size) for data in market_history]

    for col_name in ["best_bid", "best_ask", "mid"]:
        history_frames[col_name] = history_frames[col_name].apply(lambda x: float(x) if x else None)

    return history_frames


@attr.s(auto_attribs=True)
class PRISMMarketDataset(MarketDataset):
    market_history: List[MarketData]
    seq_length: int
    target_horizon: int
    use_mid: bool = True
    cuda: bool = False

    def __attrs_post_init__(self):
        self.data = market_history_to_df(self.market_history)
        if self.use_mid:
            self.features = ["mid"]
        else:
            self.features = ["best_bid", "best_ask"]

        self.data = self.data[self.features].dropna()
        self._torch_data = torch.from_numpy(self.data[self.features].values).type(torch.float32)
        assert len(self._torch_data) >= self.seq_length + self.target_horizon

    def get_features(self):
        return self.features

    def __len__(self):
        return len(self._torch_data[: -self.seq_length - self.target_horizon + 1])

    def __getitem__(self, idx: int):
        return {
            "input": self._torch_data[idx : idx + self.seq_length],
            "target": self._torch_data[idx + self.seq_length + self.target_horizon - 1],
            "name": f"{idx}",
        }

    def get_next_market_state(self, time: int) -> Tuple[int, TimestepInfo]:
        volume = self.data["last_trade_volume"]
        return time % len(self), TimestepInfo(
            ask=self.data[time]["best_ask"],
            bid=self.data[time]["best_bid"],
            volume=volume,
            timestep=time,
        )
