from decimal import Decimal
from typing import Dict, Optional, Sequence, Tuple

import matplotlib.figure
import numpy as np
import pandas as pd
import plotly.express as px
import scipy.stats
import statsmodels.graphics.tsaplots
import xarray as xr
from plotly.graph_objs import Figure

from prism_agents.api.xr_util import (
    ALIAS,
    ASK_PROFILE,
    BATCHED_RUN_DIM,
    BEST_ASK,
    BEST_BID,
    BID_PROFILE,
    MID,
    MID_INCREMENT,
    ORDER_BOOK_SNAPSHOT,
    PNL_RANKING,
    POSITION,
    REALIZEDPL,
    SIMULATION_TIME,
    SNAPSHOT_PRICE_DIM,
    TIMESTEP,
    WEALTH,
    client_data_dataset,
    make_order_book_snapshot_over_time,
    market_history_summary_dataset,
    multi_run_client_data_dataset,
    volume_profile_over_time_dataset,
)
from prism_agents.runner.experiment_result import ExperimentResult, TimedObservation

MARKET_SUMMARY_PLOT = "market_summary"
POSITION_HISTORY_PLOT = "position_history"
WEALTH_HISTORY_PLOT = "wealth_history"
REALIZED_PL_PLOT = "realized_pl"
BID_MEAN_VOLUME_PROFILE_PLOT = "bid_mean_volume_profile"
ASK_MEAN_VOLUME_PROFILE_PLOT = "ask_mean_volume_profile"
SNAPSHOT_OVER_TIME_PLOT = "order_book_snapshot_evolution"

FINAL_WEALTH_BOXPLOT = "final_wealth_boxplot"
FINAL_REALIZED_PNL_BOXPLOT = "final_realized_pnl_boxplot"

FINAL_PNL_RANKING_TABLE = "final_pnl_ranking_table"
FINAL_WEALTH_TABLE = "final_wealth_table"
FINAL_PNL_TABLE = "final_pnl_table"


def _xr_market_history_summary_plot(ds: xr.Dataset) -> Figure:
    df: pd.DataFrame = ds.to_dataframe().reset_index()
    df = df.dropna(subset=[BEST_BID, BEST_ASK, MID])
    df = df.sort_values(by=SIMULATION_TIME)
    return px.line(df, x=SIMULATION_TIME, y=[BEST_BID, BEST_ASK, MID])


def _xr_market_history_autocorr_plot(ds: xr.Dataset) -> matplotlib.figure.Figure:
    return statsmodels.graphics.tsaplots.plot_acf(ds[MID_INCREMENT].values, alpha=0.05)


def market_history_summary_plot(
    market_history: Sequence[TimedObservation],
    tick_size: Decimal = Decimal("0.01"),
) -> Figure:
    ds = market_history_summary_dataset(market_history=market_history, tick_size=tick_size)
    return _xr_market_history_summary_plot(ds)


def position_history_plot(df: pd.DataFrame) -> Figure:
    return px.line(df, x=SIMULATION_TIME, y=POSITION, color=ALIAS)


def wealth_history_plot(df: pd.DataFrame) -> Figure:
    return px.line(df, x=SIMULATION_TIME, y=WEALTH, color=ALIAS)


def realized_pl_plot(df: pd.DataFrame) -> Figure:
    return px.line(df, x=SIMULATION_TIME, y=REALIZEDPL, color=ALIAS)


def bid_ask_volume_profile_plots(volume_profile: xr.Dataset) -> Dict[str, Figure]:
    """
    Plot the mean volume profile of the bid and ask sides of the dataset
    """
    x_label = "tick from mid price"
    bid_mean = (volume_profile[BID_PROFILE]).mean(dim=TIMESTEP).to_dataframe()
    bid_mean[x_label] = list(range(len(bid_mean)))

    ask_mean = (volume_profile[ASK_PROFILE]).mean(dim=TIMESTEP).to_dataframe()
    ask_mean[x_label] = list(range(len(ask_mean)))

    ask_plot = px.line(
        data_frame=ask_mean, x=x_label, y=ASK_PROFILE, labels={ASK_PROFILE: "mean ask volume"}
    )
    bid_plot = px.line(
        data_frame=bid_mean, x=x_label, y=BID_PROFILE, labels={BID_PROFILE: "mean bid volume"}
    )
    return {BID_MEAN_VOLUME_PROFILE_PLOT: bid_plot, ASK_MEAN_VOLUME_PROFILE_PLOT: ask_plot}


def basic_summary_plots(
    result: ExperimentResult, tick_size: Decimal = Decimal("0.01")
) -> Dict[str, Figure]:
    """
    Collection of plots describing path of market and order book
    """
    market_dataset = market_history_summary_dataset(
        market_history=result.observations, tick_size=tick_size
    )
    client_dataset = client_data_dataset(result.observations)
    client_dataframe = client_dataset.to_dataframe().reset_index()
    plots = {
        MARKET_SUMMARY_PLOT: _xr_market_history_summary_plot(market_dataset),
        POSITION_HISTORY_PLOT: position_history_plot(client_dataframe),
        WEALTH_HISTORY_PLOT: wealth_history_plot(client_dataframe),
        REALIZED_PL_PLOT: realized_pl_plot(client_dataframe),
    }
    volume_dataset = volume_profile_over_time_dataset(result.observations)
    plots.update(bid_ask_volume_profile_plots(volume_dataset))
    return plots


def snapshot_plot(
    result: ExperimentResult,
    bounds: Optional[Tuple[Decimal, Decimal]] = None,
) -> Dict[str, Figure]:
    """
    Create a plot showing the evolution of the volume of
    orders at each tick in a range
    over time
    """
    ds = make_order_book_snapshot_over_time(result.observations, bounds=bounds)
    da: xr.DataArray = ds[ORDER_BOOK_SNAPSHOT].assign_coords({TIMESTEP: ds[SIMULATION_TIME]})
    da = da.transpose(SNAPSHOT_PRICE_DIM, TIMESTEP)
    da = da.rename("volume")
    return {
        SNAPSHOT_OVER_TIME_PLOT: px.imshow(
            da, origin="lower", color_continuous_scale="tropic", color_continuous_midpoint=0.0
        )
    }


def basic_multi_experiment_plots(results: Sequence[ExperimentResult]) -> Dict[str, Figure]:
    client_ds = multi_run_client_data_dataset(results)
    return multi_run_client_boxplots(client_ds)


def basic_multi_experiment_tables(results: Sequence[ExperimentResult]) -> Dict[str, pd.DataFrame]:
    client_ds = multi_run_client_data_dataset(results)
    return multi_run_client_tables(client_ds)


def _make_multi_run_boxplot(da: xr.DataArray) -> Figure:
    df = da.to_dataframe().reset_index()
    return px.box(df, x=ALIAS, y=da.name)


def _make_multi_run_table(da: xr.DataArray, confidence_level=0.95) -> pd.DataFrame:
    mean = da.mean(dim=BATCHED_RUN_DIM)
    da = da.transpose(ALIAS, BATCHED_RUN_DIM)
    lower_mean_ce, upper_mean_ce = scipy.stats.bootstrap(
        data=(da,),
        axis=da.dims.index(BATCHED_RUN_DIM),
        statistic=np.mean,
        confidence_level=confidence_level,
        method="basic",
    ).confidence_interval
    lower_mean_ce = xr.DataArray(lower_mean_ce, dims=(ALIAS,))
    upper_mean_ce = xr.DataArray(upper_mean_ce, dims=(ALIAS,))
    statistic_dim = "statistic"
    da = xr.concat([mean, lower_mean_ce, upper_mean_ce], dim=statistic_dim).transpose(
        ALIAS, statistic_dim
    )
    df: pd.DataFrame = da.to_pandas()
    df.columns = [
        str(da.name) + f"_{nm}"
        for nm in ["mean", f"mean_lower_{confidence_level}", f"mean_upper_{confidence_level}"]
    ]
    df.rows = da.coords[ALIAS].values.tolist()
    df = df.reset_index()
    return df


def multi_run_client_tables(multi_run_client_dataset: xr.Dataset) -> Dict[str, pd.DataFrame]:
    """
    Create a table recording some statistics across a
    batch of runs, together with associated confidence
    intervals
    """
    final_pnl_ranking = multi_run_client_dataset[PNL_RANKING][{TIMESTEP: -1}]
    final_wealth = multi_run_client_dataset[WEALTH][{TIMESTEP: -1}]
    final_pnl = multi_run_client_dataset[REALIZEDPL][{TIMESTEP: -1}]
    return {
        FINAL_PNL_RANKING_TABLE: _make_multi_run_table(final_pnl_ranking),
        FINAL_PNL_TABLE: _make_multi_run_table(final_pnl),
        FINAL_WEALTH_TABLE: _make_multi_run_table(final_wealth),
    }


def multi_run_client_boxplots(multi_run_client_dataset: xr.Dataset) -> Dict[str, Figure]:
    final_wealth = multi_run_client_dataset[WEALTH][{TIMESTEP: -1}]
    final_pnl = multi_run_client_dataset[REALIZEDPL][{TIMESTEP: -1}]
    return {
        FINAL_WEALTH_BOXPLOT: _make_multi_run_boxplot(final_wealth),
        FINAL_REALIZED_PNL_BOXPLOT: _make_multi_run_boxplot(final_pnl),
    }
