import time
from typing import Dict, Optional, Sequence
from uuid import UUID

import attr
import ray
from tqdm import tqdm

from prism_agents.api.server_interface import ServerInterface
from prism_agents.runner.agent_runner import AgentRunner
from prism_agents.runner.experiment_result import ExperimentResult, Observation, TimedObservation


@attr.s(frozen=True, auto_attribs=True)
class ExperimentSpec:
    duration: float
    time_warp_factor: float = 1.0
    num_observations: int = 10


@ray.remote
def _run_agent(agent: AgentRunner, kwargs: Dict):
    agent.run(**kwargs)


def run_experiment(
    agents: Sequence[AgentRunner],
    experiment_spec: ExperimentSpec,
    server_interface: ServerInterface,
    client_ids: Optional[Sequence[Optional[UUID]]] = None,
) -> ExperimentResult:
    kwargs = {
        "server_interface": server_interface,
        "time_warp_factor": experiment_spec.time_warp_factor,
    }
    scaled_duration = experiment_spec.duration / experiment_spec.time_warp_factor

    client_ids = client_ids if client_ids else [None] * len(agents)
    assert len(client_ids) == len(agents)
    for agent, client_id in zip(agents, client_ids):
        kwargs["client_id"] = client_id
        _run_agent.remote(agent, kwargs)

    result = sleep_and_observe(
        scaled_duration,
        num_observations=experiment_spec.num_observations,
        server_interface=server_interface,
    )
    return result


def sleep_and_observe(
    duration: float, num_observations: int, server_interface: ServerInterface
) -> ExperimentResult:
    client_data = []
    order_book_data = []
    wait_time = duration / num_observations
    for _ in tqdm(range(num_observations)):
        time.sleep(wait_time)
        client_data.append(server_interface.get_client_data_table())
        order_book_data.append({"PRSM": server_interface.get_market_data("PRSM")})
    return ExperimentResult(
        observations=[
            TimedObservation(i * wait_time, Observation(market=order_item, client_data=client_item))
            for i, order_item, client_item in zip(
                range(1, num_observations + 1), order_book_data, client_data
            )
        ]
    )
