import heapq
from typing import Iterator, List, Optional, Sequence, Tuple
from uuid import UUID

import attr
from tqdm import tqdm

from prism_agents.api.server_interface import ServerInterface
from prism_agents.runner.experiment_result import ExperimentResult, Observation, TimedObservation
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner


@attr.s(frozen=True, auto_attribs=True)
class SimpleExperimentSpec:
    """
    Configuration for a single-process simulation
    """

    duration: float
    observation_times: Sequence[float]


@attr.s(auto_attribs=True)
class _Observer(StepwiseAgentRunner):
    """
    An artificial agent used to observe the simulation and record results
    """

    observation_times: Sequence[float]
    observations: List[TimedObservation] = attr.ib(factory=list)
    symbol: str = "PRSM"

    def run_step(
        self, server_interface: ServerInterface, client_id: Optional[UUID] = None
    ) -> Iterator[float]:
        o_times = sorted(self.observation_times)
        sleep_times = [o_times[0]] + [o_times[i] - o_times[i - 1] for i in range(1, len(o_times))]
        for sleep_time, obs_time in zip(sleep_times, o_times):
            yield sleep_time
            self.observations.append(
                TimedObservation(
                    obs_time,
                    Observation(
                        market={self.symbol: server_interface.get_market_data(self.symbol)},
                        client_data=server_interface.get_client_data_table(),
                    ),
                )
            )
        while True:
            yield 100


def run_experiment_simple(
    agents: Sequence[StepwiseAgentRunner],
    experiment_spec: SimpleExperimentSpec,
    server_interface: ServerInterface,
) -> ExperimentResult:
    """
    Run an experiment in single process, using a priority queue to keep track of which agent will
    act next.
    """
    queue: List[Tuple[float, int, Iterator[float]]] = []
    observer = _Observer(observation_times=experiment_spec.observation_times)
    queue.append((0.0, -1, observer.run_step(server_interface)))
    for i, agent in enumerate(agents):
        heapq.heappush(queue, (0.0, i, agent.run_step(server_interface=server_interface)))
    time = 0.0
    with tqdm(total=experiment_spec.duration) as pbar:
        while time <= experiment_spec.duration:
            new_time, agent_no, current_runner = heapq.heappop(queue)
            pbar.update(new_time - time)
            time = new_time
            sleep = next(current_runner)
            heapq.heappush(queue, (time + sleep, agent_no, current_runner))
    return ExperimentResult(observations=observer.observations)
