from pathlib import Path

import attr
import pandas as pd
import plotly
import plotly.express as px
import torch
import xarray as xr

from prism_agents.agents.cst_agent_torch.cst_consts import PAPER_CST_PARAMS
from prism_agents.agents.cst_agent_torch.cst_core import CSTState
from prism_agents.agents.cst_agent_torch.cst_plot import (
    EVENT_DIMENSION,
    MID_TICK,
    QUANTITY,
    TICK_DIMENSION,
    TIME,
    n_steps_dataset,
)
from prism_agents.plotting.plot_util import save_plotly


def make_dataset() -> xr.Dataset:
    quantities = torch.Tensor([0] * 25)
    initial_width = 5
    for i in range(initial_width):
        quantities[12 - i] = -1
    for i in range(initial_width):
        quantities[13 + i] = 1
    generator = torch.Generator().manual_seed(123456789)
    params = attr.evolve(PAPER_CST_PARAMS, max_width=25)
    ds = n_steps_dataset(
        params=params,
        initial_state=CSTState(quantities=quantities),
        generator=generator,
        num_steps=6000,
    )
    return ds


def mid_price_plot(df: pd.DataFrame) -> plotly.graph_objs.Figure:
    df = df.groupby(TIME).mean().reset_index()
    fig = px.line(df, x=TIME, y=MID_TICK)
    return fig


def final_order_book(df: pd.DataFrame) -> plotly.graph_objs.Figure:
    events = df[EVENT_DIMENSION]
    final_df = df[events == max(events)]
    fig = px.bar(final_df, x=TICK_DIMENSION, y=QUANTITY)
    return fig


def book_heatmap(ds: xr.Dataset) -> plotly.graph_objs.Figure:
    da = ds[QUANTITY]
    da = da.assign_coords({TIME: (EVENT_DIMENSION, ds[TIME])})
    return px.imshow(da, origin="lower")


if __name__ == "__main__":
    ds = make_dataset()
    output_dir = Path("/data/user/q4xwsdt/projects/prisme/experiments/v20210516")
    mid_path = output_dir.joinpath("mid_tick.html")
    final_book_path = output_dir.joinpath("final_order_book.html")
    book_heatmap_path = output_dir.joinpath("book_heatmap.html")

    df = ds.to_dataframe().reset_index()
    mid_plot = mid_price_plot(df)
    final_book_plot = final_order_book(df)
    book_heatmap_plot = book_heatmap(ds)
    save_plotly(mid_plot, mid_path)
    save_plotly(final_book_plot, final_book_path)
    save_plotly(book_heatmap_plot, book_heatmap_path)
