import logging
from decimal import Decimal
from pathlib import Path
from typing import Optional

import numpy as np
import pytorch_lightning as pl
import wandb
from attrs import asdict
from wandb.wandb_run import Run

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import Fundamentalist
from prism_agents.agents.forecast.forecasters.fundamentalist.process.ornstein_uhlenbeck import (
    OrnsteinUhlenbeckState,
)
from prism_agents.agents.forecast.forecasters.lstm.adaptive_lstm import AdaptiveLSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.lstm import LSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import (
    LightningLSTMForecastingModule,
    LSTMForecastingModuleInputs,
)
from prism_agents.agents.forecast.forecasters.moving_average import MovingAverageForecaster
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.agents.forecast.trading_rule.spread_trading_rule import SpreadTradingRule
from prism_agents.api.local_server.local_server_interface import LocalServerInterface
from prism_agents.api.xr_util import client_data_dataset
from prism_agents.plotting.market_history_plot import basic_summary_plots
from prism_agents.plotting.plot_util import save_plots_to_dir
from prism_agents.runner.simple_experiment_runner import SimpleExperimentSpec, run_experiment_simple
from prism_agents.tracking.loading import load_from_wandb

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_arima(offset: float, period: float):
    limit = 1
    auto_forecater = AutoArimaForecaster(min_obs_for_forecast=20, max_obs=200)
    arima_trader = ForecastRunner(
        forecaster=auto_forecater,
        trading_rule=SimpleTradingRule(volume=1, position_min=-limit, position_max=limit),
        horizon=2,
        symbol="PRSM",
        alias="arima",
        period=period,
        offset=offset,
    )
    return [arima_trader]


def get_lstms(offset_init: float, period: float, wandb_run: Run):
    limit = 1
    offset = offset_init
    models = {"lstm_train": "model-2t4rszac:v0", "lstm_domain_random": "model-1eln69pz:v0"}
    agents = []
    for name, artifact_name in models.items():
        torch_forecaster = load_from_wandb(
            cls=LightningLSTMForecastingModule,
            run=wandb_run,
            wandb_path=f"prism-agents/prism-agents/{artifact_name}",
            type="model",
        )
        forecaster = LSTMForecaster(forecaster=torch_forecaster)
        trading_rule = SimpleTradingRule(
            volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
        )
        lstm_forecast = ForecastRunner(
            forecaster=forecaster,
            trading_rule=trading_rule,
            horizon=1,
            alias=f"{name}",
            symbol="PRSM",
            offset=offset,
            period=period,
        )
        agents.append(lstm_forecast)
        offset += 0.005 / len(models)

    rdm_lstm = LightningLSTMForecastingModule(
        input_size=1, hidden_size=32, num_layers=3, seq_length=20
    )
    forecaster = LSTMForecaster(forecaster=rdm_lstm)
    trading_rule = SimpleTradingRule(
        volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
    )
    lstm_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias="UntrainedLSTM",
        symbol="PRSM",
        offset=offset,
        period=period,
    )
    agents.append(lstm_forecast)
    return agents


def get_adaptive_lstm(
    offset_init: float,
    period: float,
    alias: str,
    lr: float = 1e-4,
    seq_length: int = 20,
    wandb_run: Optional[Run] = None,
    model_name: Optional[str] = None,
):

    if model_name is not None:
        assert wandb_run is not None
        torch_forecaster = load_from_wandb(
            cls=LightningLSTMForecastingModule,
            run=wandb_run,
            wandb_path=f"prism-agents/prism-agents/{model_name}",
            type="model",
        )
    else:
        seq_length = 20
        _inputs = LSTMForecastingModuleInputs(
            input_size=1,
            hidden_size=32,
            num_layers=3,
            seq_length=seq_length,
        )
        torch_forecaster = LightningLSTMForecastingModule(lr=lr, **asdict(_inputs))

    def get_trainer():
        trainer = pl.Trainer(max_epochs=3, enable_checkpointing=False, logger=False)
        return trainer

    limit = 1
    trading_rule = SimpleTradingRule(
        volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
    )

    forecaster = AdaptiveLSTMForecaster(forecaster=torch_forecaster, get_trainer=get_trainer)
    lstm_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias=alias,
        symbol="PRSM",
        offset=offset_init,
        period=period,
    )
    return lstm_forecast


def get_adaptive_lstms(offset_init: float, period: float, wandb_run: Run):

    pretrained_lstm_forecast = get_adaptive_lstm(
        offset_init,
        period,
        alias="lr=1e-4_pre_trained_adaptive_lstm",
        wandb_run=wandb_run,
        model_name="model-1eln69pz:v0",
        lr=1e-4,
    )

    rdm_lstm_forecast = get_adaptive_lstm(
        offset_init + 0.001, period, alias="lr=1e-3 untrained_adaptive_lstm", lr=1e-3
    )

    return [pretrained_lstm_forecast, rdm_lstm_forecast]


def get_moving_average(offset_init: float, period: float):
    forecaster = MovingAverageForecaster(alpha_long=0.5, alpha_short=0.9)
    limit = 1
    trading_rule = SimpleTradingRule(
        volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
    )
    ma_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias="moving_average",
        symbol="PRSM",
        offset=offset_init,
        period=period,
    )
    return [ma_forecast]


if __name__ == "__main__":
    out_dir = Path("sandbox/lstm_test_outputs")
    run = wandb.init(
        name="lstm_on_prism_tests",
        project="prism-agents",
        job_type="PRISM_run_test",
    )

    alias_ou = "ou"
    alias_lstm = "lstm"
    period = 0.02
    duration = 60
    ou_forecast = Fundamentalist(
        start_state=OrnsteinUhlenbeckState.create_with_half_life(
            half_life=3,
            sigma=5,
            step=0,
            value=90,
            b=50,
        ),
    )

    rule = SpreadTradingRule(
        volume=2,
    )
    ou_trader = ForecastRunner(
        forecaster=ou_forecast,
        trading_rule=rule,
        horizon=1,
        symbol="PRSM",
        alias=alias_ou,
        period=period,
    )
    agents = [ou_trader]
    offset = 0.001

    agents.extend(get_adaptive_lstms(offset, period, run))
    offset += 0.003

    agents.extend(get_moving_average(offset, period))
    offset += 0.001

    agents.extend(get_arima(offset, period))
    offset += 0.001

    agents.extend(get_lstms(offset, period, run))
    offset += 0.005

    result = run_experiment_simple(
        agents,
        experiment_spec=SimpleExperimentSpec(
            duration=duration,
            observation_times=list(np.arange(0, duration, 0.01)),
        ),
        server_interface=LocalServerInterface(),
    )
    if out_dir is not None:
        plots = basic_summary_plots(result)
        client_data_dataset(result.observations)
        save_plots_to_dir(plots, folder=out_dir)
        for name, plot in plots.items():
            wandb.log({name: plot})
