import time
from decimal import Decimal
from pathlib import Path
from typing import Optional

import wandb
from plato.bear.engine.bear_engine import BearEngine, BearEngineState
from torch.optim import Adam

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.lstm.dataset import SPYBidAskVolumeLSTMDataset
from prism_agents.agents.forecast.forecasters.lstm.lstm import LSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import (
    LSTMForecastingModule,
    LSTMForecastingModuleInputs,
)
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import (
    SimpleTradingRule,
    TradingRule,
)
from prism_agents.api.remote_request_funcs import RemoteServerInterface
from prism_agents.runner.experiment_runner import ExperimentSpec, run_experiment
from prism_agents.tracking.checkpoint import ModelReloadCheckpoint, WANDBBearCheckpoint
from prism_agents.tracking.loading import load_from_checkpoint


def get_lstm_forecast_runner(
    symbol: str,
    alias: str,
    torch_forecasting_module: LSTMForecastingModule,
    trading_rule: Optional[TradingRule] = None,
    period: float = 5.0,
):
    """
    Helper function to make a ForecastRunner out of a trading rule and a forecasting module
    """
    if trading_rule is None:
        trading_rule = SimpleTradingRule(volume=10, omega=Decimal(3))
    forecaster = LSTMForecaster(forecaster=torch_forecasting_module)
    runner = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias=alias,
        symbol=symbol,
        period=period,
    )
    return runner


def train_lstm(path: Path, wandb_sync: bool = False):
    dataset = SPYBidAskVolumeLSTMDataset(tiny=True, batch_size=200, target_shift=5)

    seq_length = dataset[0]["input"].shape[0]
    _inputs = LSTMForecastingModuleInputs(
        input_size=2, hidden_size=5, num_layers=2, seq_length=seq_length
    )
    f = LSTMForecastingModule(_inputs)

    # check that fitting the LSTM changes its weights/predictions
    state = BearEngineState(max_epochs=2)
    if wandb_sync:
        run = wandb.init(
            project="prism-agents", job_type="test", entity="prism-agents", name="bear-test"
        )
        chkpt = WANDBBearCheckpoint(str(path), run=run)
    else:
        chkpt = ModelReloadCheckpoint(str(path))

    engine = BearEngine(checkpoint=chkpt, state=state)
    bear_dataset = {"train": dataset, "eval": dataset}
    engine.dataset = bear_dataset
    engine.model = f
    engine.optimizer = Adam(f.parameters(), lr=1e-3)
    engine.train()


if __name__ == "__main__":
    time_tag = str(time.time())[-3:]
    path = Path("sandbox/lstm_tests/example")
    if not path.exists():
        print("training the LSTM")
        train_lstm(path)

    url = "http://francis-test-server-route-a5w0-photon.apps.borealis-rnd-scc.fg.rbc.com"
    server_interface = RemoteServerInterface(url)

    trading_rule = SimpleTradingRule(volume=100, omega=Decimal("-100"), strict=False)

    torch_forecasting_module = load_from_checkpoint(
        LSTMForecastingModule, checkpoint_path=path, checkpoint_name="last"
    )
    agents = []
    for i in range(2):
        lstm_runner = get_lstm_forecast_runner(
            symbol="PRSM",
            alias=f"{time_tag}_t{i}_lstm",
            torch_forecasting_module=torch_forecasting_module,
            trading_rule=trading_rule,
            period=0.5,
        )
        agents.append(lstm_runner)

    specs = ExperimentSpec(duration=10, time_warp_factor=1, num_observations=10)
    run_experiment(agents=agents, experiment_spec=specs, server_interface=server_interface)
