import cProfile
import pstats
from pathlib import Path
from typing import Optional

import numpy as np

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import Fundamentalist
from prism_agents.agents.forecast.forecasters.fundamentalist.process.ornstein_uhlenbeck import (
    OrnsteinUhlenbeckState,
)
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.agents.forecast.trading_rule.spread_trading_rule import SpreadTradingRule
from prism_agents.api.local_server.local_server_interface import LocalServerInterface
from prism_agents.api.xr_util import client_data_dataset
from prism_agents.plotting.market_history_plot import basic_summary_plots
from prism_agents.plotting.plot_util import save_plots_to_dir
from prism_agents.runner.simple_experiment_runner import SimpleExperimentSpec, run_experiment_simple


def arima_experiment(out_dir: Optional[Path] = None):
    alias_ou = "ou"
    alias_arima = "arima"
    period = 0.02
    duration = 14
    ou_forecast = Fundamentalist(
        start_state=OrnsteinUhlenbeckState.create_with_half_life(
            half_life=2, sigma=5, step=0, value=50, b=50
        ),
    )
    rule = SpreadTradingRule(
        volume=10,
    )
    ou_trader = ForecastRunner(
        forecaster=ou_forecast,
        trading_rule=rule,
        horizon=1,
        symbol="PRSM",
        alias=alias_ou,
        period=period,
    )
    limit = 1
    auto_forecater = AutoArimaForecaster(min_obs_for_forecast=200, max_obs=200)
    arima_trader = ForecastRunner(
        forecaster=auto_forecater,
        trading_rule=SimpleTradingRule(volume=1, position_min=-limit, position_max=limit),
        horizon=2,
        symbol="PRSM",
        alias=alias_arima,
        period=period,
        offset=0.01,
    )
    result = run_experiment_simple(
        [ou_trader, arima_trader],
        experiment_spec=SimpleExperimentSpec(
            duration=duration,
            observation_times=list(np.arange(0, duration, 0.01)),
        ),
        server_interface=LocalServerInterface(),
    )
    if out_dir is not None:
        plots = basic_summary_plots(result)
        client_data_dataset(result.observations)
        save_plots_to_dir(plots, folder=out_dir)


prof_stats_name = "prof_stats_base.prof"


def run_with_profile():
    cProfile.run("arima_experiment()", prof_stats_name)


def load_prof():
    p = pstats.Stats(prof_stats_name)
    p.sort_stats("cumulative").print_stats(40)


if __name__ == "__main__":
    run_with_profile()
