from pathlib import Path
from typing import Dict, Optional

import numpy as np
import pandas as pd

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import (
    create_spread_fundamentalist,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.dataset_process import (
    create_dataset_process_from_series,
)
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.api.local_server.local_server_interface import LocalServerInterface
from prism_agents.plotting.market_history_plot import basic_summary_plots
from prism_agents.plotting.plot_util import save_plots_to_dir
from prism_agents.runner.simple_experiment_runner import SimpleExperimentSpec, run_experiment_simple


def load_m4_series(pth: Path) -> Dict[str, pd.Series]:
    """
    In M4 datasets, time series are stored in the rows of the dataframe
    """
    df = pd.read_parquet(pth)
    return {name: ser.dropna() for name, ser in df.iterrows()}


def m4_experiment(
    m4_path: Path,
    series_name: str,
    out_dir: Optional[Path] = None,
):
    duration = 12
    alias_arima = "arima"
    period = 0.02
    m4_dict = load_m4_series(m4_path)
    m4_series = m4_dict[series_name]
    m4_trader = create_spread_fundamentalist(
        initial_state=create_dataset_process_from_series(m4_series)
    )
    limit = 1
    auto_forecater = AutoArimaForecaster(min_obs_for_forecast=200, max_obs=200)
    arima_trader = ForecastRunner(
        forecaster=auto_forecater,
        trading_rule=SimpleTradingRule(volume=1, position_min=-limit, position_max=limit),
        horizon=2,
        symbol="PRSM",
        alias=alias_arima,
        period=period,
        offset=0.01,
    )
    result = run_experiment_simple(
        [m4_trader, arima_trader],
        experiment_spec=SimpleExperimentSpec(
            duration=duration,
            observation_times=list(np.arange(0, duration, 0.01)),
        ),
        server_interface=LocalServerInterface(),
    )
    if out_dir is not None:
        plots = basic_summary_plots(result)
        save_plots_to_dir(plots, folder=out_dir)


if __name__ == "__main__":
    m4_path = Path(
        "/data/user/q4xwsdt/src/m4/M4-methods-master/Dataset/Train/Daily-train.parquet.gzip"
    )
    out_path = Path("/data/user/q4xwsdt/prism_test_results/v20220620/m4_plots")
    m4_experiment(m4_path, series_name="D2", out_dir=out_path)
