import logging
from decimal import Decimal
from pathlib import Path
from typing import List, Optional

import numpy as np
import pytorch_lightning as pl
import wandb
from attrs import asdict
from wandb.wandb_run import Run

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.fundamentalist.fund_forecaster import Fundamentalist
from prism_agents.agents.forecast.forecasters.fundamentalist.process.ornstein_uhlenbeck import (
    ou_transition_function,
)
from prism_agents.agents.forecast.forecasters.fundamentalist.process.regime_switching import (
    RegimeSwitchingState,
)
from prism_agents.agents.forecast.forecasters.lstm.adaptive_lstm import AdaptiveLSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import (
    LightningLSTMForecastingModule,
    LSTMForecastingModuleInputs,
)
from prism_agents.agents.forecast.forecasters.moving_average import MovingAverageForecaster
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.agents.forecast.trading_rule.spread_trading_rule import SpreadTradingRule
from prism_agents.api.local_server.local_server_interface import LocalServerInterface
from prism_agents.plotting.market_history_plot import basic_multi_experiment_tables
from prism_agents.runner.simple_batch_runner import run_batch
from prism_agents.runner.simple_experiment_runner import SimpleExperimentSpec
from prism_agents.runner.stepwise_agent_runner import StepwiseAgentRunner
from prism_agents.tracking.loading import load_from_wandb

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def get_arima(offset: float, period: float):
    limit = 1
    auto_forecater = AutoArimaForecaster(min_obs_for_forecast=20, max_obs=200)
    arima_trader = ForecastRunner(
        forecaster=auto_forecater,
        trading_rule=SimpleTradingRule(volume=1, position_min=-limit, position_max=limit),
        horizon=2,
        symbol="PRSM",
        alias="arima",
        period=period,
        offset=offset,
    )
    return [arima_trader]


def get_adaptive_lstm(
    offset_init: float,
    period: float,
    alias: str,
    lr: float = 1e-4,
    wandb_run: Optional[Run] = None,
    model_name: Optional[str] = None,
):

    if model_name is not None:
        assert wandb_run is not None
        torch_forecaster = load_from_wandb(
            cls=LightningLSTMForecastingModule,
            run=wandb_run,
            wandb_path=f"prism-agents/prism-agents/{model_name}",
            type="model",
        )
    else:
        seq_length = 20
        _inputs = LSTMForecastingModuleInputs(
            input_size=1,
            hidden_size=32,
            num_layers=3,
            seq_length=seq_length,
        )
        torch_forecaster = LightningLSTMForecastingModule(lr=lr, **asdict(_inputs))

    def get_trainer():
        trainer = pl.Trainer(max_epochs=3, enable_checkpointing=False, logger=False)
        return trainer

    limit = 1
    trading_rule = SimpleTradingRule(
        volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
    )

    forecaster = AdaptiveLSTMForecaster(forecaster=torch_forecaster, get_trainer=get_trainer)
    lstm_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias=alias,
        symbol="PRSM",
        offset=offset_init,
        period=period,
    )
    return lstm_forecast


def get_adaptive_lstms(offset_init: float, period: float):

    rdm_lstm_forecast = get_adaptive_lstm(
        offset_init + 0.001, period, alias="lr=1e-3 untrained_adaptive_lstm", lr=1e-3
    )

    return [rdm_lstm_forecast]


def get_moving_average(offset_init: float, period: float):
    forecaster = MovingAverageForecaster(alpha_long=0.5, alpha_short=0.9)
    limit = 1
    trading_rule = SimpleTradingRule(
        volume=1, omega=Decimal(1), strict=False, position_min=-limit, position_max=limit
    )
    ma_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias="moving_average",
        symbol="PRSM",
        offset=offset_init,
        period=period,
    )
    return [ma_forecast]


def make_agents() -> List[StepwiseAgentRunner]:
    alias_fundamental = "fundamental"
    period = 0.02
    matrix = np.array([[0.98, 0.01, 0.01], [0.01, 0.98, 0.01], [0.01, 0.01, 0.98]])
    ou_forecast = Fundamentalist(
        start_state=RegimeSwitchingState.create(
            transition_matrix=matrix,
            regime_transition_functions=[
                ou_transition_function(alpha=1 / 2, sigma=0.05, b=1),
                ou_transition_function(alpha=1 / 2, sigma=0.05, b=1.1),
                ou_transition_function(alpha=1 / 2, sigma=0.05, b=1.4),
            ],
            starting_value=1,
        )
    )

    rule = SpreadTradingRule(
        volume=10,
    )
    ou_trader = ForecastRunner(
        forecaster=ou_forecast,
        trading_rule=rule,
        horizon=1,
        symbol="PRSM",
        alias=alias_fundamental,
        period=period,
    )
    agents = [ou_trader]
    offset = 0.001

    agents.extend(get_adaptive_lstms(offset, period))
    offset += 0.003
    agents.extend(get_arima(offset, period))
    return agents


if __name__ == "__main__":
    out_dir = Path("sandbox/lstm_test_regime_switch")
    run = wandb.init(
        name="simple_lstm_on_prism_tests",
        project="prism-agents",
        job_type="PRISM_run_test",
    )

    duration = 50
    num_runs = 20
    experiment_spec = SimpleExperimentSpec(
        duration=duration,
        observation_times=[duration],
    )
    results = run_batch(
        spec=experiment_spec,
        agent_sequences=[make_agents() for _ in range(num_runs)],
        server_interface=LocalServerInterface(),
        mode="parallel",
        parallel_workers=10,
    )
    if out_dir is not None:
        tables = basic_multi_experiment_tables(results)
        for name, table in tables.items():
            wandb.log({name: wandb.Table(dataframe=table)})
