import json
import logging
import os
from pathlib import Path
from typing import Type, Union

import pytorch_lightning as pl
import torch
from wandb.sdk.wandb_run import Run

from prism_agents.agents.utils.serialize import SerializableBaseModel

serial_type = Union[Type[SerializableBaseModel], Type[pl.LightningModule]]

logger = logging.getLogger(__name__)

LIGHTNING_CHKPT_EXT = ".ckpt"
CONSTRUCTOR_INPUTS_CHKPT_NAME = "constructor_inputs.json"


def load_from_checkpoint(cls: serial_type, checkpoint_path: Path, checkpoint_name: str = ""):
    """
    Load weights from a checkpoint that used a SerializableBaseModel or from a model
    that was checkpointed by pytorch lightning
    """
    if issubclass(cls, pl.LightningModule):
        # allow us to use this function with a pytorch lightning module class
        checkpoints_found = [f for f in checkpoint_path.glob("*" + LIGHTNING_CHKPT_EXT)]
        assert len(checkpoints_found) == 1, (
            "More than one checkpoint found in the WANDB artifact dir, "
            "refine the wandb_path to be more specific."
        )

        logger.info(f"Loading Lightning Model from {checkpoints_found[0]}.")
        return cls.load_from_checkpoint(str(checkpoints_found[0]))

    assert hasattr(cls, "from_dict")

    state = torch.load(checkpoint_path.absolute() / checkpoint_name)
    with open(str(checkpoint_path.absolute() / CONSTRUCTOR_INPUTS_CHKPT_NAME), "r") as file:
        inputs_dict = json.load(file)
    torch_module = cls.from_dict(inputs_dict)
    torch_module.load_state_dict(state["model"], strict=False)
    return torch_module


def download_wandb_artifact(wandb_path: str, run: Run, type: str = "model"):
    artifact = run.use_artifact(wandb_path, type=type)
    artifact_dir_path = Path(os.environ.get("WANDB_DIR", "artifacts")) / artifact.name
    artifact_dir = artifact.download(str(artifact_dir_path))
    return artifact_dir


def load_from_wandb(cls: serial_type, wandb_path: str, run: Run, type: str = "model"):
    artifact_dir = download_wandb_artifact(wandb_path, run, type)
    return load_from_checkpoint(cls, Path(artifact_dir), "")
