import argparse
import copy
from pathlib import Path

import pytorch_lightning as pl
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from torch.utils.data import DataLoader

from prism_agents.agents.forecast.forecasters.lstm.nn import LightningLSTMForecastingModule
from prism_agents.agents.utils.vector_simulation.processes import ou_process
from prism_agents.dataset.fundamental_process import FundamentalProcessDataset, InfoRange

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--out_dir",
        type=str,
        default="sandbox/lstm_training_example",
        help="Directory to store results such as our trained model.",
    )
    args = parser.parse_args()

    out_dir = Path(args.out_dir)
    # First Generate a Dataset Based on the Ornstein Uhlenbeck Stochastic Process
    input_range = {
        "step_size": InfoRange(mode="choice", options=[1]),
        "initial_values": InfoRange(mode="uniform", options=[30, 100]),
        "b": InfoRange(mode="uniform", options=[30, 100]),
        "sigma": InfoRange(mode="uniform", options=[1, 10]),
        "half_life": InfoRange(mode="uniform", options=[0.1, 10]),
    }

    observation_history = 20
    train_dataset = FundamentalProcessDataset(
        stochastic_process_generator=ou_process,
        input_range=input_range,
        batch_size=128,
        seq_length=observation_history,
        target_horizon=1,
        batches_per_epoch=10,
    )
    valid_dataset = copy.deepcopy(train_dataset)
    valid_dataset.batches_per_epoch = 2

    # Generate our dataloaders used during training
    train_dataloader = DataLoader(train_dataset, batch_size=None, sampler=None)
    valid_dataloader = DataLoader(valid_dataset, batch_size=None, sampler=None)

    # Create the LSTM model
    f = LightningLSTMForecastingModule(
        input_size=1, hidden_size=16, num_layers=3, seq_length=observation_history, lr=1e-4
    )

    chkpt = pl.callbacks.ModelCheckpoint()
    early_stopping = EarlyStopping(monitor="val_loss", mode="min", patience=5)
    trainer = pl.Trainer(
        max_epochs=3,
        callbacks=[chkpt, early_stopping],
        default_root_dir=str(out_dir),
    )
    trainer.fit(f, train_dataloaders=train_dataloader, val_dataloaders=valid_dataloader)

    # Once fitted, the models should be saved at the checkpointed location and can be reloaded in
    # other scripts
    f_reloaded = LightningLSTMForecastingModule.load_from_checkpoint(chkpt.best_model_path)
