"""
This script builds a single forecasting agent and runs it on a prism server.
"""
import argparse
import logging
import os
import time
from pathlib import Path

from dotenv import load_dotenv

from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.api.remote_request_funcs import RemoteServerInterface
from prism_agents.tutorials.utils import (
    DEFAULT_PRISM_AGENTS_ID_FOLDER_LOCATION,
    PRISM_DEFAULT_TEST_SERVER,
    get_client_id,
    load_client_id,
    save_client_id,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

PRISM_SYMBOL = "PRSM"

load_dotenv(override=True)

PRISM_DEFAULT_TEST_SERVER = os.environ["PRISM_DEFAULT_TEST_SERVER"]
PRISM_ACCESS_TOKEN = os.environ["PRISM_ACCESS_TOKEN"]


def register_and_manage_id(alias: str, server_interface: RemoteServerInterface):
    # When one registers a new alias, the client_id should be stored somewhere safe as
    # it is the participant's secret and unique identifier to their account
    id_location = Path(
        os.environ.get("PRISM_AGENTS_ID_FOLDER_LOCATION", DEFAULT_PRISM_AGENTS_ID_FOLDER_LOCATION)
    )
    try:
        # This would fail if the alias is already registered on a server
        client_id = server_interface.register(alias)
        # If we were able to register, we save the id to a file
        # WARNING: this overwrites any old client id files that may have been registered
        # with the same alias. This means one loses the client ID of these old agents.
        save_client_id(alias, client_id, location=id_location)
    except AssertionError:
        # If already registered on prism, you may have saved the id already, simply load it.
        client_id = load_client_id(alias, location=id_location)

    """
    Note: a helper function that does the above (registers, saves, loads) exists
    We can replace the entirety of this `register_and_manage_id` function by the single line below.

    This is the recommend way to register and load an alias as it offers very detailed
    error messages. If many users run the same tutorial on the same prism server, alias collisions
    are notoriously common and can cause confusing errors.

    Here we explicitly pass the overwrite=True flag, otherwise it will throw an error if a
    client ID file for a given alias already exists.
    """
    client_id = get_client_id(alias, server_interface=server_interface, overwrite=True)
    return client_id


def get_arima(
    period: float,
    server_interface: RemoteServerInterface,
    position_limit: int = 1,
):
    """
    Load an AutoArimaForecaster with a Simple trading rule
    """
    # Note, prism-agents runners will automatically register themselves after a call to .run()
    # if no client_id is passed during their construction and/or when called .run()
    # Although useful during quick development, this is generally *not* the desired behavior
    # as it will create a new entity under which to trade. Here we explicitly register the agent.
    alias = "arima" + str(time.time())[-6:]

    client_id = register_and_manage_id(alias, server_interface)
    # Once the above function is understood, replace the above line by the commented one below
    # client_id = get_client_id(alias, server_interface=server_interface, overwrite=True)

    auto_forecaster = AutoArimaForecaster(min_obs_for_forecast=20, max_obs=200)
    trading_rule = SimpleTradingRule(
        volume=1, position_min=-position_limit, position_max=position_limit, strict=False
    )
    arima_trader = ForecastRunner(
        forecaster=auto_forecaster,
        trading_rule=trading_rule,
        alias=alias,
        symbol=PRISM_SYMBOL,
        period=period,
        client_id=client_id,
    )
    return arima_trader


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--url", type=str, default=PRISM_DEFAULT_TEST_SERVER, help="URL of the Prism server"
    )
    parser.add_argument(
        "--access-token", type=str, default=PRISM_ACCESS_TOKEN, help="Access token for Prism server"
    )
    args = parser.parse_args()
    print(args)
    server = RemoteServerInterface(server_url=args.url, access_token=args.access_token, robust=True)
    arima_forecaster_agent = get_arima(period=2.0, position_limit=5, server_interface=server)
    # we use a robust server interface here so that our run
    # does not crash due to networking or async issues.
    # We recommend a non-robust server interface for testing
    # and a robust server interface
    # for the competition.
    client_id = arima_forecaster_agent.client_id
    # Warning: client_id is an optional parameter to .run() if we initialized the agent with a
    # client_id, however we explicitly pass it in this tutorial as a reminder that
    # this is an important piece of information for Prism to know who is sending orders!
    logging.info(
        "The arima forecaster is running! Visit the PRISM server"
        " dashboard to see any trading activity!"
    )
    arima_forecaster_agent.run(server_interface=server, client_id=client_id)
