"""
This script builds 3 Forecast Agent Models and 1 Fundamental Agent model and make them run
together on a deployed Prism server.

"""

import argparse
import logging
import os
import time
from decimal import Decimal
from pathlib import Path

import pytorch_lightning as pl
import ray
from dotenv import load_dotenv

from prism_agents.agents.background.annealing_schedule import ExponentialAnnealingState
from prism_agents.agents.forecast.forecast_agent import ForecastRunner
from prism_agents.agents.forecast.forecasters.lstm.adaptive_lstm import AdaptiveLSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.lstm import LSTMForecaster
from prism_agents.agents.forecast.forecasters.lstm.nn import LightningLSTMForecastingModule
from prism_agents.agents.forecast.forecasters.refitting.auto_arima import AutoArimaForecaster
from prism_agents.agents.forecast.trading_rule.simple_trading_rule import SimpleTradingRule
from prism_agents.agents.utils.agent_creation.ou import create_ou_poisson_agent
from prism_agents.api.remote_request_funcs import RemoteServerInterface
from prism_agents.api.xr_util import client_data_dataset
from prism_agents.plotting.market_history_plot import basic_summary_plots
from prism_agents.plotting.plot_util import save_plots_to_dir
from prism_agents.runner.experiment_runner import ExperimentSpec, run_experiment
from prism_agents.tutorials.utils import PRISM_DEFAULT_TEST_SERVER, get_client_id

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

PRISM_SYMBOL = "PRSM"

load_dotenv(override=True)

PRISM_DEFAULT_TEST_SERVER = os.environ["PRISM_DEFAULT_TEST_SERVER"]
PRISM_ACCESS_TOKEN = os.environ["PRISM_ACCESS_TOKEN"]

# Change this identifier if the client IDs are not found and the alias are already registered
tutorial_alias_identifier = str(time.time())[-5:]


def get_arima(period: float, server_interface: RemoteServerInterface, position_limit: int = 1):
    """
    Load an AutoArimaForecaster with a Simple trading rule
    """
    alias = "arima " + tutorial_alias_identifier
    client_id = get_client_id(alias, server_interface)

    auto_forecater = AutoArimaForecaster(min_obs_for_forecast=20, max_obs=200)
    trading_rule = SimpleTradingRule(
        volume=1, position_min=-position_limit, position_max=position_limit, strict=False
    )
    arima_trader = ForecastRunner(
        forecaster=auto_forecater,
        trading_rule=trading_rule,
        symbol=PRISM_SYMBOL,
        alias=alias,
        period=period,
        client_id=client_id,
    )
    return arima_trader


def get_lstm(period: float, server_interface: RemoteServerInterface, position_limit: int = 1):
    alias = "Untrained LSTM " + tutorial_alias_identifier
    client_id = get_client_id(alias, server_interface)
    untrained_torch_forecaster = LightningLSTMForecastingModule(
        input_size=1, hidden_size=32, num_layers=3, seq_length=20
    )
    forecaster = LSTMForecaster(forecaster=untrained_torch_forecaster)
    trading_rule = SimpleTradingRule(
        volume=1,
        omega=Decimal(1),
        strict=False,
        position_min=-position_limit,
        position_max=position_limit,
    )
    lstm_forecaster = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias=alias,
        symbol=PRISM_SYMBOL,
        period=period,
        client_id=client_id,
    )
    return lstm_forecaster


def get_adaptive_lstm(
    period: float,
    server_interface: RemoteServerInterface,
    position_limit: int = 1,
    lr: float = 1e-3,
):
    alias = "Adaptive LSTM " + tutorial_alias_identifier
    client_id = get_client_id(alias, server_interface)

    untrained_torch_forecaster = LightningLSTMForecastingModule(
        input_size=1,
        hidden_size=32,
        num_layers=3,
        seq_length=20,
        lr=lr,
    )

    # For the Adaptive LSTM based on torch lightning
    # we need to supply a get_trainer function that returns the Trainer to be used for an update
    def get_trainer():
        trainer = pl.Trainer(max_epochs=3, enable_checkpointing=False, logger=False)
        return trainer

    trading_rule = SimpleTradingRule(
        volume=1,
        omega=Decimal(1),
        strict=False,
        position_min=-position_limit,
        position_max=position_limit,
    )

    forecaster = AdaptiveLSTMForecaster(
        forecaster=untrained_torch_forecaster, get_trainer=get_trainer
    )
    adaptive_lstm_forecast = ForecastRunner(
        forecaster=forecaster,
        trading_rule=trading_rule,
        horizon=1,
        alias=alias,
        symbol=PRISM_SYMBOL,
        period=period,
        client_id=client_id,
    )
    return adaptive_lstm_forecast


def get_fundamentalist(
    period: float, server_interface: RemoteServerInterface, annealer_half_life: float
):
    alias = "OU FT " + tutorial_alias_identifier
    client_id = get_client_id(alias, server_interface)
    ou_runner = create_ou_poisson_agent(
        seed=71,
        fundamental_period=period,
        limit_arrival_rate=10,
        half_life=3,
        sigma=0.1,
        mean=10,
        client_id=client_id,
        symbol=PRISM_SYMBOL,
        annealer=ExponentialAnnealingState.create_with_half_life(annealer_half_life),
    )
    return ou_runner


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--out_dir",
        type=str,
        default="sandbox/test_outputs",
        help="Directory to store plots and experiment results.",
    )
    parser.add_argument(
        "--url", type=str, default=PRISM_DEFAULT_TEST_SERVER, help="URL of the Prism server"
    )
    parser.add_argument(
        "--access-token", type=str, default=PRISM_ACCESS_TOKEN, help="Access token for Prism server"
    )
    parser.add_argument(
        "--duration",
        type=int,
        default=60,
        help="Duration in seconds at which the agents posts on the PRISM server",
    )

    args = parser.parse_args()
    out_dir = Path(args.out_dir)

    server_interface = RemoteServerInterface(server_url=args.url, access_token=args.access_token)

    agents = []
    # Make every agent act every 0.2 second with a position limit of 5
    agents.append(get_arima(0.2, position_limit=5, server_interface=server_interface))
    agents.append(get_lstm(0.2, position_limit=5, server_interface=server_interface))
    agents.append(get_adaptive_lstm(0.2, position_limit=5, server_interface=server_interface))

    # We make the fundamentalist post more frequently than our agents to the prism server

    # It also does not have a position limit.
    period = 0.2
    duration = args.duration  # number of seconds for which to run the experiment
    # set the fundamental agent to decay half its volume
    # over every quarter of the competition.
    annealer_half_life = (duration / period) / 4
    agents.append(
        get_fundamentalist(
            period, server_interface=server_interface, annealer_half_life=annealer_half_life
        )
    )

    experiment_spec = ExperimentSpec(
        duration=duration,
        num_observations=args.duration
        * 2,  # Number of snapshot to be taken during the experiment being ran.
        time_warp_factor=1,
    )

    # The run_experiment uses ray to spawn a process for each agents.
    # The following call to ray.init() is optional. But we prefer to initialize ray
    # with specific arguments for the tutorial's purpose.
    # Note that ray logs to `/tmp/ray` by default, if this directory is protected on your machine
    # you can change the location at which these are saved by uncommenting the _temp_dir=... line.
    ray.init(
        include_dashboard=False,
        # _temp_dir=/your/path/to/ray/logs
    )

    result = run_experiment(
        agents,
        experiment_spec=experiment_spec,
        server_interface=server_interface,
    )

    # Create and save basic out of the box plot summaries
    if out_dir is not None:
        plots = basic_summary_plots(result)
        client_data_dataset(result.observations)
        save_plots_to_dir(plots, folder=out_dir)
