import json
import logging
import os
from pathlib import Path
from typing import Optional
from uuid import UUID

from prism_agents.api.server_interface import ServerInterface

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

PRISM_DEFAULT_TEST_SERVER = "https://prism.borealisai.com/ac70cf62-e1dc-4e6f-b082-d0a63d8ad979"

PRISM_DEFAULT_TEST_SERVER_UI = "https://prism.borealisai.com"

PRISM_AGENTS_TUTORIAL_LOCATION = Path(os.path.realpath(__file__)).parents[0]
IDS_FOLDER = "client_ids"
DEFAULT_PRISM_AGENTS_ID_FOLDER_LOCATION = PRISM_AGENTS_TUTORIAL_LOCATION / IDS_FOLDER


class AliasRegisteredButClientIDNotRecognized(Exception):
    """
    Raised if a client ID file is found and the alias is registered on Prism but the loaded
     client ID is not registered on Prism.
    """


def save_client_id(alias: str, client_id: UUID, location: Path):
    os.makedirs(location, exist_ok=True)
    logging.info(
        f"Saving the client ID for alias {alias} at {str(location / alias)}. The client ID is "
        f"the unique identifier of the agent {alias} allowing authentication in Prism. "
        f"Do not share or lose this file! If you lose your client ID, contact the Prism team."
    )
    json.dump(str(client_id), open(location / alias, "w"))


def load_client_id(alias: str, location: Optional[Path] = None) -> UUID:
    if location is None:
        location = Path(
            os.environ.get(
                "PRISM_AGENTS_ID_FOLDER_LOCATION", DEFAULT_PRISM_AGENTS_ID_FOLDER_LOCATION
            )
        )
    if not os.path.exists(location / alias):
        raise FileNotFoundError(
            f"The alias {alias} is registered on Prism but no client ID file "
            f"{str(location / alias)} is found. If you are the owner of this alias "
            f"but have lost your client ID, please contact the Prism team."
        )
    logger.info(f"Loading client ID for {alias} from {str(location / alias)}.")
    return UUID(json.load(open(location / alias, "r")))


def register_and_save_alias(
    alias: str, server_interface: ServerInterface, location: Path, overwrite: bool = False
) -> UUID:
    if os.path.exists(location / alias) and not overwrite:
        raise FileExistsError(
            f"The alias {alias} already exists at {str(location / alias)} yet does not "
            f"exists on the Prism server. This means that this alias was likely registered on "
            f"another server. If you don't want to save this client ID, pass overwrite=True to "
            f"overwrite this alias' client ID."
        )

    logger.info(f"Client ID for {alias} not found in PRISM, registering the alias on PRISM.")
    client_id = server_interface.register(alias)
    save_client_id(alias, client_id, location=location)
    return client_id


def get_client_id(alias: str, server_interface: ServerInterface, overwrite: bool = False) -> UUID:
    """
    Helper function to save/load a client id if (not) found in the client_ids folder or not
    registered on Prism.

    If overwrite is not True and a local file containing the client ID associated with the alias is
    found but the same alias is not already registered, an error will be thrown.
    If overwrite is False, the current file containing the client ID will be overwritten.
    """

    location = Path(
        os.environ.get("PRISM_AGENTS_ID_FOLDER_LOCATION", DEFAULT_PRISM_AGENTS_ID_FOLDER_LOCATION)
    )

    client_data_table = server_interface.get_client_data_table().data
    aliases = [client.alias for client in client_data_table]
    if alias not in aliases:
        client_id = register_and_save_alias(
            alias, server_interface, location=location, overwrite=overwrite
        )
    else:
        client_id = load_client_id(alias, location=location)

    try:
        client_data = server_interface.get_client_data(client_id)
        if client_data.alias != alias:
            logger.warning(
                f"The loaded client ID is registered to {client_data.alias} and not {alias}! "
                f"Therefore trades with this ID will be under the alias {client_data.alias}."
                f"Something odd is occuring but if you got here, you must know what you're doing."
            )
    except Exception as e:  # This error is generated by attrs
        # failing to parse the returned json in RemoteServerInterface
        raise AliasRegisteredButClientIDNotRecognized(
            f"It seems {alias} is already registered on Prism and a client ID file for "
            f"the alias was found at {str(location / alias)}. However the Prism server "
            f"does not recognize this client ID. Someone likely has already registered "
            f"the alias before you, in which case you should pick a different alias!"
            f"If you may have overwritten the alias' client ID file by mistake and you "
            f"believe you are the owner of this alias, please contact the Prism team."
        ) from e

    return client_id
